#pragma once
#include <LibCore/Proxy.h>
#include <LibHTTP/HeaderMap.h>
#include <LibURL/URL.h>
#include <RequestServer/CacheLevel.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::RequestServer {

enum class MessageID : i32 {
    InitTransport = 1,
    InitTransportResponse = 2,
    ConnectNewClient = 3,
    ConnectNewClientResponse = 4,
    SetDnsServer = 5,
    IsSupportedProtocol = 6,
    IsSupportedProtocolResponse = 7,
    StartRequest = 8,
    StopRequest = 9,
    StopRequestResponse = 10,
    SetCertificate = 11,
    SetCertificateResponse = 12,
    EnsureConnection = 13,
    WebsocketConnect = 14,
    WebsocketSend = 15,
    WebsocketClose = 16,
    WebsocketSetCertificate = 17,
    WebsocketSetCertificateResponse = 18,
};

class InitTransportResponse final : public IPC::Message {
public:

    InitTransportResponse(InitTransportResponse const&) = default;
    InitTransportResponse(InitTransportResponse&&) = default;
    InitTransportResponse& operator=(InitTransportResponse const&) = default;
    InitTransportResponse(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransportResponse(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransportResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransportResponse; }
    static i32 static_message_id() { return (int)MessageID::InitTransportResponse; }
    virtual const char* message_name() const override { return "RequestServer::InitTransportResponse"; }

    static ErrorOr<NonnullOwnPtr<InitTransportResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto peer_pid = TRY((decoder.decode<int>()));

        return make<InitTransportResponse>(move(peer_pid));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InitTransportResponse));

        TRY(stream.encode(m_peer_pid));

        return buffer;
    }

    const int& peer_pid() const { return m_peer_pid; }
    int take_peer_pid() { return move(m_peer_pid); }

private:

    int m_peer_pid;

};

class InitTransport final : public IPC::Message {
public:

   typedef class InitTransportResponse ResponseType;

    InitTransport(InitTransport const&) = default;
    InitTransport(InitTransport&&) = default;
    InitTransport& operator=(InitTransport const&) = default;
    InitTransport(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransport(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransport() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransport; }
    static i32 static_message_id() { return (int)MessageID::InitTransport; }
    virtual const char* message_name() const override { return "RequestServer::InitTransport"; }

    static ErrorOr<NonnullOwnPtr<InitTransport>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto peer_pid = TRY((decoder.decode<int>()));

        return make<InitTransport>(move(peer_pid));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InitTransport));

        TRY(stream.encode(m_peer_pid));

        return buffer;
    }

    const int& peer_pid() const { return m_peer_pid; }
    int take_peer_pid() { return move(m_peer_pid); }

private:

    int m_peer_pid;

};

class ConnectNewClientResponse final : public IPC::Message {
public:

    ConnectNewClientResponse(ConnectNewClientResponse const&) = default;
    ConnectNewClientResponse(ConnectNewClientResponse&&) = default;
    ConnectNewClientResponse& operator=(ConnectNewClientResponse const&) = default;
    ConnectNewClientResponse(IPC::File client_socket) : m_client_socket(move(client_socket)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, IPC::File>)
    ConnectNewClientResponse(WrappedReturnType&& value)
        : m_client_socket(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ConnectNewClientResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectNewClientResponse; }
    static i32 static_message_id() { return (int)MessageID::ConnectNewClientResponse; }
    virtual const char* message_name() const override { return "RequestServer::ConnectNewClientResponse"; }

    static ErrorOr<NonnullOwnPtr<ConnectNewClientResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto client_socket = TRY((decoder.decode<IPC::File>()));

        return make<ConnectNewClientResponse>(move(client_socket));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ConnectNewClientResponse));

        TRY(stream.encode(m_client_socket));

        return buffer;
    }

    const IPC::File& client_socket() const { return m_client_socket; }
    IPC::File take_client_socket() { return move(m_client_socket); }

private:

    IPC::File m_client_socket;

};

class ConnectNewClient final : public IPC::Message {
public:

   typedef class ConnectNewClientResponse ResponseType;

    ConnectNewClient(ConnectNewClient const&) = default;
    ConnectNewClient(ConnectNewClient&&) = default;
    ConnectNewClient& operator=(ConnectNewClient const&) = default;
    ConnectNewClient() {}

    virtual ~ConnectNewClient() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectNewClient; }
    static i32 static_message_id() { return (int)MessageID::ConnectNewClient; }
    virtual const char* message_name() const override { return "RequestServer::ConnectNewClient"; }

    static ErrorOr<NonnullOwnPtr<ConnectNewClient>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        return make<ConnectNewClient>();
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ConnectNewClient));

        return buffer;
    }

private:

};

class SetDnsServer final : public IPC::Message {
public:

    SetDnsServer(SetDnsServer const&) = default;
    SetDnsServer(SetDnsServer&&) = default;
    SetDnsServer& operator=(SetDnsServer const&) = default;
    SetDnsServer(ByteString host_or_address, u16 port, bool use_tls) : m_host_or_address(move(host_or_address)), m_port(move(port)), m_use_tls(move(use_tls)) {}

    virtual ~SetDnsServer() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::SetDnsServer; }
    static i32 static_message_id() { return (int)MessageID::SetDnsServer; }
    virtual const char* message_name() const override { return "RequestServer::SetDnsServer"; }

    static ErrorOr<NonnullOwnPtr<SetDnsServer>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto host_or_address = TRY((decoder.decode<ByteString>()));

        auto port = TRY((decoder.decode<u16>()));

        auto use_tls = TRY((decoder.decode<bool>()));

        return make<SetDnsServer>(move(host_or_address), move(port), move(use_tls));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetDnsServer));

        TRY(stream.encode(m_host_or_address));

        TRY(stream.encode(m_port));

        TRY(stream.encode(m_use_tls));

        return buffer;
    }

    const ByteString& host_or_address() const { return m_host_or_address; }
    ByteString take_host_or_address() { return move(m_host_or_address); }

    const u16& port() const { return m_port; }
    u16 take_port() { return move(m_port); }

    const bool& use_tls() const { return m_use_tls; }
    bool take_use_tls() { return move(m_use_tls); }

private:

    ByteString m_host_or_address;

    u16 m_port;

    bool m_use_tls;

};

class IsSupportedProtocolResponse final : public IPC::Message {
public:

    IsSupportedProtocolResponse(IsSupportedProtocolResponse const&) = default;
    IsSupportedProtocolResponse(IsSupportedProtocolResponse&&) = default;
    IsSupportedProtocolResponse& operator=(IsSupportedProtocolResponse const&) = default;
    IsSupportedProtocolResponse(bool supported) : m_supported(move(supported)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, bool>)
    IsSupportedProtocolResponse(WrappedReturnType&& value)
        : m_supported(forward<WrappedReturnType>(value))
    {
    }

    virtual ~IsSupportedProtocolResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::IsSupportedProtocolResponse; }
    static i32 static_message_id() { return (int)MessageID::IsSupportedProtocolResponse; }
    virtual const char* message_name() const override { return "RequestServer::IsSupportedProtocolResponse"; }

    static ErrorOr<NonnullOwnPtr<IsSupportedProtocolResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto supported = TRY((decoder.decode<bool>()));

        return make<IsSupportedProtocolResponse>(move(supported));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::IsSupportedProtocolResponse));

        TRY(stream.encode(m_supported));

        return buffer;
    }

    const bool& supported() const { return m_supported; }
    bool take_supported() { return move(m_supported); }

private:

    bool m_supported;

};

class IsSupportedProtocol final : public IPC::Message {
public:

   typedef class IsSupportedProtocolResponse ResponseType;

    IsSupportedProtocol(IsSupportedProtocol const&) = default;
    IsSupportedProtocol(IsSupportedProtocol&&) = default;
    IsSupportedProtocol& operator=(IsSupportedProtocol const&) = default;
    IsSupportedProtocol(ByteString protocol) : m_protocol(move(protocol)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, ByteString>)
    IsSupportedProtocol(WrappedReturnType&& value)
        : m_protocol(forward<WrappedReturnType>(value))
    {
    }

    virtual ~IsSupportedProtocol() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::IsSupportedProtocol; }
    static i32 static_message_id() { return (int)MessageID::IsSupportedProtocol; }
    virtual const char* message_name() const override { return "RequestServer::IsSupportedProtocol"; }

    static ErrorOr<NonnullOwnPtr<IsSupportedProtocol>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto protocol = TRY((decoder.decode<ByteString>()));

        return make<IsSupportedProtocol>(move(protocol));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::IsSupportedProtocol));

        TRY(stream.encode(m_protocol));

        return buffer;
    }

    const ByteString& protocol() const { return m_protocol; }
    ByteString take_protocol() { return move(m_protocol); }

private:

    ByteString m_protocol;

};

class StartRequest final : public IPC::Message {
public:

    StartRequest(StartRequest const&) = default;
    StartRequest(StartRequest&&) = default;
    StartRequest& operator=(StartRequest const&) = default;
    StartRequest(i32 request_id, ByteString method, URL::URL url, HTTP::HeaderMap request_headers, ByteBuffer request_body, Core::ProxyData proxy_data) : m_request_id(move(request_id)), m_method(move(method)), m_url(move(url)), m_request_headers(move(request_headers)), m_request_body(move(request_body)), m_proxy_data(move(proxy_data)) {}

    virtual ~StartRequest() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::StartRequest; }
    static i32 static_message_id() { return (int)MessageID::StartRequest; }
    virtual const char* message_name() const override { return "RequestServer::StartRequest"; }

    static ErrorOr<NonnullOwnPtr<StartRequest>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto request_id = TRY((decoder.decode<i32>()));

        auto method = TRY((decoder.decode<ByteString>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto request_headers = TRY((decoder.decode<HTTP::HeaderMap>()));

        auto request_body = TRY((decoder.decode<ByteBuffer>()));

        auto proxy_data = TRY((decoder.decode<Core::ProxyData>()));

        return make<StartRequest>(move(request_id), move(method), move(url), move(request_headers), move(request_body), move(proxy_data));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::StartRequest));

        TRY(stream.encode(m_request_id));

        TRY(stream.encode(m_method));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_request_headers));

        TRY(stream.encode(m_request_body));

        TRY(stream.encode(m_proxy_data));

        return buffer;
    }

    const i32& request_id() const { return m_request_id; }
    i32 take_request_id() { return move(m_request_id); }

    const ByteString& method() const { return m_method; }
    ByteString take_method() { return move(m_method); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const HTTP::HeaderMap& request_headers() const { return m_request_headers; }
    HTTP::HeaderMap take_request_headers() { return move(m_request_headers); }

    const ByteBuffer& request_body() const { return m_request_body; }
    ByteBuffer take_request_body() { return move(m_request_body); }

    const Core::ProxyData& proxy_data() const { return m_proxy_data; }
    Core::ProxyData take_proxy_data() { return move(m_proxy_data); }

private:

    i32 m_request_id;

    ByteString m_method;

    URL::URL m_url;

    HTTP::HeaderMap m_request_headers;

    ByteBuffer m_request_body;

    Core::ProxyData m_proxy_data;

};

class StopRequestResponse final : public IPC::Message {
public:

    StopRequestResponse(StopRequestResponse const&) = default;
    StopRequestResponse(StopRequestResponse&&) = default;
    StopRequestResponse& operator=(StopRequestResponse const&) = default;
    StopRequestResponse(bool success) : m_success(move(success)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, bool>)
    StopRequestResponse(WrappedReturnType&& value)
        : m_success(forward<WrappedReturnType>(value))
    {
    }

    virtual ~StopRequestResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::StopRequestResponse; }
    static i32 static_message_id() { return (int)MessageID::StopRequestResponse; }
    virtual const char* message_name() const override { return "RequestServer::StopRequestResponse"; }

    static ErrorOr<NonnullOwnPtr<StopRequestResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto success = TRY((decoder.decode<bool>()));

        return make<StopRequestResponse>(move(success));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::StopRequestResponse));

        TRY(stream.encode(m_success));

        return buffer;
    }

    const bool& success() const { return m_success; }
    bool take_success() { return move(m_success); }

private:

    bool m_success;

};

class StopRequest final : public IPC::Message {
public:

   typedef class StopRequestResponse ResponseType;

    StopRequest(StopRequest const&) = default;
    StopRequest(StopRequest&&) = default;
    StopRequest& operator=(StopRequest const&) = default;
    StopRequest(i32 request_id) : m_request_id(move(request_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, i32>)
    StopRequest(WrappedReturnType&& value)
        : m_request_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~StopRequest() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::StopRequest; }
    static i32 static_message_id() { return (int)MessageID::StopRequest; }
    virtual const char* message_name() const override { return "RequestServer::StopRequest"; }

    static ErrorOr<NonnullOwnPtr<StopRequest>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto request_id = TRY((decoder.decode<i32>()));

        return make<StopRequest>(move(request_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::StopRequest));

        TRY(stream.encode(m_request_id));

        return buffer;
    }

    const i32& request_id() const { return m_request_id; }
    i32 take_request_id() { return move(m_request_id); }

private:

    i32 m_request_id;

};

class SetCertificateResponse final : public IPC::Message {
public:

    SetCertificateResponse(SetCertificateResponse const&) = default;
    SetCertificateResponse(SetCertificateResponse&&) = default;
    SetCertificateResponse& operator=(SetCertificateResponse const&) = default;
    SetCertificateResponse(bool success) : m_success(move(success)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, bool>)
    SetCertificateResponse(WrappedReturnType&& value)
        : m_success(forward<WrappedReturnType>(value))
    {
    }

    virtual ~SetCertificateResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::SetCertificateResponse; }
    static i32 static_message_id() { return (int)MessageID::SetCertificateResponse; }
    virtual const char* message_name() const override { return "RequestServer::SetCertificateResponse"; }

    static ErrorOr<NonnullOwnPtr<SetCertificateResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto success = TRY((decoder.decode<bool>()));

        return make<SetCertificateResponse>(move(success));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetCertificateResponse));

        TRY(stream.encode(m_success));

        return buffer;
    }

    const bool& success() const { return m_success; }
    bool take_success() { return move(m_success); }

private:

    bool m_success;

};

class SetCertificate final : public IPC::Message {
public:

   typedef class SetCertificateResponse ResponseType;

    SetCertificate(SetCertificate const&) = default;
    SetCertificate(SetCertificate&&) = default;
    SetCertificate& operator=(SetCertificate const&) = default;
    SetCertificate(i32 request_id, ByteString certificate, ByteString key) : m_request_id(move(request_id)), m_certificate(move(certificate)), m_key(move(key)) {}

    virtual ~SetCertificate() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::SetCertificate; }
    static i32 static_message_id() { return (int)MessageID::SetCertificate; }
    virtual const char* message_name() const override { return "RequestServer::SetCertificate"; }

    static ErrorOr<NonnullOwnPtr<SetCertificate>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto request_id = TRY((decoder.decode<i32>()));

        auto certificate = TRY((decoder.decode<ByteString>()));

        auto key = TRY((decoder.decode<ByteString>()));

        return make<SetCertificate>(move(request_id), move(certificate), move(key));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetCertificate));

        TRY(stream.encode(m_request_id));

        TRY(stream.encode(m_certificate));

        TRY(stream.encode(m_key));

        return buffer;
    }

    const i32& request_id() const { return m_request_id; }
    i32 take_request_id() { return move(m_request_id); }

    const ByteString& certificate() const { return m_certificate; }
    ByteString take_certificate() { return move(m_certificate); }

    const ByteString& key() const { return m_key; }
    ByteString take_key() { return move(m_key); }

private:

    i32 m_request_id;

    ByteString m_certificate;

    ByteString m_key;

};

class EnsureConnection final : public IPC::Message {
public:

    EnsureConnection(EnsureConnection const&) = default;
    EnsureConnection(EnsureConnection&&) = default;
    EnsureConnection& operator=(EnsureConnection const&) = default;
    EnsureConnection(URL::URL url, ::RequestServer::CacheLevel cache_level) : m_url(move(url)), m_cache_level(move(cache_level)) {}

    virtual ~EnsureConnection() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::EnsureConnection; }
    static i32 static_message_id() { return (int)MessageID::EnsureConnection; }
    virtual const char* message_name() const override { return "RequestServer::EnsureConnection"; }

    static ErrorOr<NonnullOwnPtr<EnsureConnection>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto url = TRY((decoder.decode<URL::URL>()));

        auto cache_level = TRY((decoder.decode<::RequestServer::CacheLevel>()));

        return make<EnsureConnection>(move(url), move(cache_level));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::EnsureConnection));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_cache_level));

        return buffer;
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ::RequestServer::CacheLevel& cache_level() const { return m_cache_level; }
    ::RequestServer::CacheLevel take_cache_level() { return move(m_cache_level); }

private:

    URL::URL m_url;

    ::RequestServer::CacheLevel m_cache_level;

};

class WebsocketConnect final : public IPC::Message {
public:

    WebsocketConnect(WebsocketConnect const&) = default;
    WebsocketConnect(WebsocketConnect&&) = default;
    WebsocketConnect& operator=(WebsocketConnect const&) = default;
    WebsocketConnect(i64 websocket_id, URL::URL url, ByteString origin, Vector<ByteString> protocols, Vector<ByteString> extensions, HTTP::HeaderMap additional_request_headers) : m_websocket_id(move(websocket_id)), m_url(move(url)), m_origin(move(origin)), m_protocols(move(protocols)), m_extensions(move(extensions)), m_additional_request_headers(move(additional_request_headers)) {}

    virtual ~WebsocketConnect() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketConnect; }
    static i32 static_message_id() { return (int)MessageID::WebsocketConnect; }
    virtual const char* message_name() const override { return "RequestServer::WebsocketConnect"; }

    static ErrorOr<NonnullOwnPtr<WebsocketConnect>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto websocket_id = TRY((decoder.decode<i64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto origin = TRY((decoder.decode<ByteString>()));

        auto protocols = TRY((decoder.decode<Vector<ByteString>>()));

        auto extensions = TRY((decoder.decode<Vector<ByteString>>()));

        auto additional_request_headers = TRY((decoder.decode<HTTP::HeaderMap>()));

        return make<WebsocketConnect>(move(websocket_id), move(url), move(origin), move(protocols), move(extensions), move(additional_request_headers));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::WebsocketConnect));

        TRY(stream.encode(m_websocket_id));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_origin));

        TRY(stream.encode(m_protocols));

        TRY(stream.encode(m_extensions));

        TRY(stream.encode(m_additional_request_headers));

        return buffer;
    }

    const i64& websocket_id() const { return m_websocket_id; }
    i64 take_websocket_id() { return move(m_websocket_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& origin() const { return m_origin; }
    ByteString take_origin() { return move(m_origin); }

    const Vector<ByteString>& protocols() const { return m_protocols; }
    Vector<ByteString> take_protocols() { return move(m_protocols); }

    const Vector<ByteString>& extensions() const { return m_extensions; }
    Vector<ByteString> take_extensions() { return move(m_extensions); }

    const HTTP::HeaderMap& additional_request_headers() const { return m_additional_request_headers; }
    HTTP::HeaderMap take_additional_request_headers() { return move(m_additional_request_headers); }

private:

    i64 m_websocket_id;

    URL::URL m_url;

    ByteString m_origin;

    Vector<ByteString> m_protocols;

    Vector<ByteString> m_extensions;

    HTTP::HeaderMap m_additional_request_headers;

};

class WebsocketSend final : public IPC::Message {
public:

    WebsocketSend(WebsocketSend const&) = default;
    WebsocketSend(WebsocketSend&&) = default;
    WebsocketSend& operator=(WebsocketSend const&) = default;
    WebsocketSend(i64 websocket_id, bool is_text, ByteBuffer data) : m_websocket_id(move(websocket_id)), m_is_text(move(is_text)), m_data(move(data)) {}

    virtual ~WebsocketSend() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketSend; }
    static i32 static_message_id() { return (int)MessageID::WebsocketSend; }
    virtual const char* message_name() const override { return "RequestServer::WebsocketSend"; }

    static ErrorOr<NonnullOwnPtr<WebsocketSend>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto websocket_id = TRY((decoder.decode<i64>()));

        auto is_text = TRY((decoder.decode<bool>()));

        auto data = TRY((decoder.decode<ByteBuffer>()));

        return make<WebsocketSend>(move(websocket_id), move(is_text), move(data));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::WebsocketSend));

        TRY(stream.encode(m_websocket_id));

        TRY(stream.encode(m_is_text));

        TRY(stream.encode(m_data));

        return buffer;
    }

    const i64& websocket_id() const { return m_websocket_id; }
    i64 take_websocket_id() { return move(m_websocket_id); }

    const bool& is_text() const { return m_is_text; }
    bool take_is_text() { return move(m_is_text); }

    const ByteBuffer& data() const { return m_data; }
    ByteBuffer take_data() { return move(m_data); }

private:

    i64 m_websocket_id;

    bool m_is_text;

    ByteBuffer m_data;

};

class WebsocketClose final : public IPC::Message {
public:

    WebsocketClose(WebsocketClose const&) = default;
    WebsocketClose(WebsocketClose&&) = default;
    WebsocketClose& operator=(WebsocketClose const&) = default;
    WebsocketClose(i64 websocket_id, u16 code, ByteString reason) : m_websocket_id(move(websocket_id)), m_code(move(code)), m_reason(move(reason)) {}

    virtual ~WebsocketClose() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketClose; }
    static i32 static_message_id() { return (int)MessageID::WebsocketClose; }
    virtual const char* message_name() const override { return "RequestServer::WebsocketClose"; }

    static ErrorOr<NonnullOwnPtr<WebsocketClose>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto websocket_id = TRY((decoder.decode<i64>()));

        auto code = TRY((decoder.decode<u16>()));

        auto reason = TRY((decoder.decode<ByteString>()));

        return make<WebsocketClose>(move(websocket_id), move(code), move(reason));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::WebsocketClose));

        TRY(stream.encode(m_websocket_id));

        TRY(stream.encode(m_code));

        TRY(stream.encode(m_reason));

        return buffer;
    }

    const i64& websocket_id() const { return m_websocket_id; }
    i64 take_websocket_id() { return move(m_websocket_id); }

    const u16& code() const { return m_code; }
    u16 take_code() { return move(m_code); }

    const ByteString& reason() const { return m_reason; }
    ByteString take_reason() { return move(m_reason); }

private:

    i64 m_websocket_id;

    u16 m_code;

    ByteString m_reason;

};

class WebsocketSetCertificateResponse final : public IPC::Message {
public:

    WebsocketSetCertificateResponse(WebsocketSetCertificateResponse const&) = default;
    WebsocketSetCertificateResponse(WebsocketSetCertificateResponse&&) = default;
    WebsocketSetCertificateResponse& operator=(WebsocketSetCertificateResponse const&) = default;
    WebsocketSetCertificateResponse(bool success) : m_success(move(success)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, bool>)
    WebsocketSetCertificateResponse(WrappedReturnType&& value)
        : m_success(forward<WrappedReturnType>(value))
    {
    }

    virtual ~WebsocketSetCertificateResponse() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketSetCertificateResponse; }
    static i32 static_message_id() { return (int)MessageID::WebsocketSetCertificateResponse; }
    virtual const char* message_name() const override { return "RequestServer::WebsocketSetCertificateResponse"; }

    static ErrorOr<NonnullOwnPtr<WebsocketSetCertificateResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto success = TRY((decoder.decode<bool>()));

        return make<WebsocketSetCertificateResponse>(move(success));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::WebsocketSetCertificateResponse));

        TRY(stream.encode(m_success));

        return buffer;
    }

    const bool& success() const { return m_success; }
    bool take_success() { return move(m_success); }

private:

    bool m_success;

};

class WebsocketSetCertificate final : public IPC::Message {
public:

   typedef class WebsocketSetCertificateResponse ResponseType;

    WebsocketSetCertificate(WebsocketSetCertificate const&) = default;
    WebsocketSetCertificate(WebsocketSetCertificate&&) = default;
    WebsocketSetCertificate& operator=(WebsocketSetCertificate const&) = default;
    WebsocketSetCertificate(i64 request_id, ByteString certificate, ByteString key) : m_request_id(move(request_id)), m_certificate(move(certificate)), m_key(move(key)) {}

    virtual ~WebsocketSetCertificate() override {}

    virtual u32 endpoint_magic() const override { return 1270977860; }
    virtual i32 message_id() const override { return (int)MessageID::WebsocketSetCertificate; }
    static i32 static_message_id() { return (int)MessageID::WebsocketSetCertificate; }
    virtual const char* message_name() const override { return "RequestServer::WebsocketSetCertificate"; }

    static ErrorOr<NonnullOwnPtr<WebsocketSetCertificate>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto request_id = TRY((decoder.decode<i64>()));

        auto certificate = TRY((decoder.decode<ByteString>()));

        auto key = TRY((decoder.decode<ByteString>()));

        return make<WebsocketSetCertificate>(move(request_id), move(certificate), move(key));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::WebsocketSetCertificate));

        TRY(stream.encode(m_request_id));

        TRY(stream.encode(m_certificate));

        TRY(stream.encode(m_key));

        return buffer;
    }

    const i64& request_id() const { return m_request_id; }
    i64 take_request_id() { return move(m_request_id); }

    const ByteString& certificate() const { return m_certificate; }
    ByteString take_certificate() { return move(m_certificate); }

    const ByteString& key() const { return m_key; }
    ByteString take_key() { return move(m_key); }

private:

    i64 m_request_id;

    ByteString m_certificate;

    ByteString m_key;

};

} // namespace Messages::RequestServer

template<typename LocalEndpoint, typename PeerEndpoint>
class RequestServerProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    RequestServerProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    int init_transport(
int peer_pid) {
        return m_connection.template send_sync<Messages::RequestServer::InitTransport>(peer_pid)->take_peer_pid();
    }

    void async_init_transport(
int peer_pid) {
        MUST(m_connection.post_message(Messages::RequestServer::InitTransport { peer_pid }));

    }

    IPC::IPCErrorOr<int> try_init_transport(
int peer_pid) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::InitTransport>(peer_pid);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    IPC::File connect_new_client(
) {
        return m_connection.template send_sync<Messages::RequestServer::ConnectNewClient>()->take_client_socket();
    }

    void async_connect_new_client(
) {
        MUST(m_connection.post_message(Messages::RequestServer::ConnectNewClient {  }));

    }

    IPC::IPCErrorOr<IPC::File> try_connect_new_client(
) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::ConnectNewClient>();
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_set_dns_server(
ByteString host_or_address, u16 port, bool use_tls) {
        MUST(m_connection.post_message(Messages::RequestServer::SetDnsServer { move(host_or_address), port, use_tls }));

    }

    bool is_supported_protocol(
ByteString protocol) {
        return m_connection.template send_sync<Messages::RequestServer::IsSupportedProtocol>(move(protocol))->take_supported();
    }

    void async_is_supported_protocol(
ByteString protocol) {
        MUST(m_connection.post_message(Messages::RequestServer::IsSupportedProtocol { move(protocol) }));

    }

    IPC::IPCErrorOr<bool> try_is_supported_protocol(
ByteString protocol) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::IsSupportedProtocol>(move(protocol));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_start_request(
i32 request_id, ByteString method, URL::URL url, HTTP::HeaderMap request_headers, ByteBuffer request_body, Core::ProxyData proxy_data) {
        MUST(m_connection.post_message(Messages::RequestServer::StartRequest { request_id, move(method), move(url), move(request_headers), move(request_body), move(proxy_data) }));

    }

    bool stop_request(
i32 request_id) {
        return m_connection.template send_sync<Messages::RequestServer::StopRequest>(request_id)->take_success();
    }

    void async_stop_request(
i32 request_id) {
        MUST(m_connection.post_message(Messages::RequestServer::StopRequest { request_id }));

    }

    IPC::IPCErrorOr<bool> try_stop_request(
i32 request_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::StopRequest>(request_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    bool set_certificate(
i32 request_id, ByteString certificate, ByteString key) {
        return m_connection.template send_sync<Messages::RequestServer::SetCertificate>(request_id, move(certificate), move(key))->take_success();
    }

    void async_set_certificate(
i32 request_id, ByteString certificate, ByteString key) {
        MUST(m_connection.post_message(Messages::RequestServer::SetCertificate { request_id, move(certificate), move(key) }));

    }

    IPC::IPCErrorOr<bool> try_set_certificate(
i32 request_id, ByteString certificate, ByteString key) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::SetCertificate>(request_id, move(certificate), move(key));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_ensure_connection(
URL::URL url, ::RequestServer::CacheLevel cache_level) {
        MUST(m_connection.post_message(Messages::RequestServer::EnsureConnection { move(url), move(cache_level) }));

    }

    void async_websocket_connect(
i64 websocket_id, URL::URL url, ByteString origin, Vector<ByteString> protocols, Vector<ByteString> extensions, HTTP::HeaderMap additional_request_headers) {
        MUST(m_connection.post_message(Messages::RequestServer::WebsocketConnect { websocket_id, move(url), move(origin), move(protocols), move(extensions), move(additional_request_headers) }));

    }

    void async_websocket_send(
i64 websocket_id, bool is_text, ByteBuffer data) {
        MUST(m_connection.post_message(Messages::RequestServer::WebsocketSend { websocket_id, is_text, move(data) }));

    }

    void async_websocket_close(
i64 websocket_id, u16 code, ByteString reason) {
        MUST(m_connection.post_message(Messages::RequestServer::WebsocketClose { websocket_id, code, move(reason) }));

    }

    bool websocket_set_certificate(
i64 request_id, ByteString certificate, ByteString key) {
        return m_connection.template send_sync<Messages::RequestServer::WebsocketSetCertificate>(request_id, move(certificate), move(key))->take_success();
    }

    void async_websocket_set_certificate(
i64 request_id, ByteString certificate, ByteString key) {
        MUST(m_connection.post_message(Messages::RequestServer::WebsocketSetCertificate { request_id, move(certificate), move(key) }));

    }

    IPC::IPCErrorOr<bool> try_websocket_set_certificate(
i64 request_id, ByteString certificate, ByteString key) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::RequestServer::WebsocketSetCertificate>(request_id, move(certificate), move(key));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class RequestServerProxy;
class RequestServerStub;

class RequestServerEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = RequestServerProxy<LocalEndpoint, RequestServerEndpoint>;
    using Stub = RequestServerStub;

    static u32 static_magic() { return 1270977860; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 1270977860) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::RequestServer::MessageID::InitTransport:
            return TRY(Messages::RequestServer::InitTransport::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::InitTransportResponse:
            return TRY(Messages::RequestServer::InitTransportResponse::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::ConnectNewClient:
            return TRY(Messages::RequestServer::ConnectNewClient::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::ConnectNewClientResponse:
            return TRY(Messages::RequestServer::ConnectNewClientResponse::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::SetDnsServer:
            return TRY(Messages::RequestServer::SetDnsServer::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::IsSupportedProtocol:
            return TRY(Messages::RequestServer::IsSupportedProtocol::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::IsSupportedProtocolResponse:
            return TRY(Messages::RequestServer::IsSupportedProtocolResponse::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::StartRequest:
            return TRY(Messages::RequestServer::StartRequest::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::StopRequest:
            return TRY(Messages::RequestServer::StopRequest::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::StopRequestResponse:
            return TRY(Messages::RequestServer::StopRequestResponse::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::SetCertificate:
            return TRY(Messages::RequestServer::SetCertificate::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::SetCertificateResponse:
            return TRY(Messages::RequestServer::SetCertificateResponse::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::EnsureConnection:
            return TRY(Messages::RequestServer::EnsureConnection::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::WebsocketConnect:
            return TRY(Messages::RequestServer::WebsocketConnect::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::WebsocketSend:
            return TRY(Messages::RequestServer::WebsocketSend::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::WebsocketClose:
            return TRY(Messages::RequestServer::WebsocketClose::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::WebsocketSetCertificate:
            return TRY(Messages::RequestServer::WebsocketSetCertificate::decode(stream, files));
        case (int)Messages::RequestServer::MessageID::WebsocketSetCertificateResponse:
            return TRY(Messages::RequestServer::WebsocketSetCertificateResponse::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode RequestServer message");
        }

        VERIFY_NOT_REACHED();
    }

};

class RequestServerStub : public IPC::Stub {
public:
    RequestServerStub() { }
    virtual ~RequestServerStub() override { }

    virtual u32 magic() const override { return 1270977860; }
    virtual ByteString name() const override { return "RequestServer"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(const IPC::Message& message) override
    {
        switch (message.message_id()) {

        case (int)Messages::RequestServer::MessageID::InitTransport: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::InitTransport&>(message);
            auto response = init_transport(request.peer_pid());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::RequestServer::MessageID::ConnectNewClient: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::ConnectNewClient&>(message);
            auto response = connect_new_client();
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::RequestServer::MessageID::SetDnsServer: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::SetDnsServer&>(message);
            set_dns_server(request.host_or_address(), request.port(), request.use_tls());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::IsSupportedProtocol: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::IsSupportedProtocol&>(message);
            auto response = is_supported_protocol(request.protocol());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::RequestServer::MessageID::StartRequest: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::StartRequest&>(message);
            start_request(request.request_id(), request.method(), request.url(), request.request_headers(), request.request_body(), request.proxy_data());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::StopRequest: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::StopRequest&>(message);
            auto response = stop_request(request.request_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::RequestServer::MessageID::SetCertificate: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::SetCertificate&>(message);
            auto response = set_certificate(request.request_id(), request.certificate(), request.key());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::RequestServer::MessageID::EnsureConnection: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::EnsureConnection&>(message);
            ensure_connection(request.url(), request.cache_level());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::WebsocketConnect: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::WebsocketConnect&>(message);
            websocket_connect(request.websocket_id(), request.url(), request.origin(), request.protocols(), request.extensions(), request.additional_request_headers());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::WebsocketSend: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::WebsocketSend&>(message);
            websocket_send(request.websocket_id(), request.is_text(), request.data());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::WebsocketClose: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::WebsocketClose&>(message);
            websocket_close(request.websocket_id(), request.code(), request.reason());
            return nullptr;

        }

        case (int)Messages::RequestServer::MessageID::WebsocketSetCertificate: {

            [[maybe_unused]] auto& request = static_cast<const Messages::RequestServer::WebsocketSetCertificate&>(message);
            auto response = websocket_set_certificate(request.request_id(), request.certificate(), request.key());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        default:
            return Error::from_string_literal("Unknown message ID for RequestServer endpoint");
        }
    }

    virtual Messages::RequestServer::InitTransportResponse init_transport(
[[maybe_unused]] int peer_pid) = 0;

    virtual Messages::RequestServer::ConnectNewClientResponse connect_new_client(
) = 0;

    virtual void set_dns_server(
[[maybe_unused]] ByteString const& host_or_address, [[maybe_unused]] u16 port, [[maybe_unused]] bool use_tls) = 0;

    virtual Messages::RequestServer::IsSupportedProtocolResponse is_supported_protocol(
[[maybe_unused]] ByteString const& protocol) = 0;

    virtual void start_request(
[[maybe_unused]] i32 request_id, [[maybe_unused]] ByteString const& method, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] HTTP::HeaderMap const& request_headers, [[maybe_unused]] ByteBuffer const& request_body, [[maybe_unused]] Core::ProxyData const& proxy_data) = 0;

    virtual Messages::RequestServer::StopRequestResponse stop_request(
[[maybe_unused]] i32 request_id) = 0;

    virtual Messages::RequestServer::SetCertificateResponse set_certificate(
[[maybe_unused]] i32 request_id, [[maybe_unused]] ByteString const& certificate, [[maybe_unused]] ByteString const& key) = 0;

    virtual void ensure_connection(
[[maybe_unused]] URL::URL const& url, [[maybe_unused]] ::RequestServer::CacheLevel const& cache_level) = 0;

    virtual void websocket_connect(
[[maybe_unused]] i64 websocket_id, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] ByteString const& origin, [[maybe_unused]] Vector<ByteString> const& protocols, [[maybe_unused]] Vector<ByteString> const& extensions, [[maybe_unused]] HTTP::HeaderMap const& additional_request_headers) = 0;

    virtual void websocket_send(
[[maybe_unused]] i64 websocket_id, [[maybe_unused]] bool is_text, [[maybe_unused]] ByteBuffer const& data) = 0;

    virtual void websocket_close(
[[maybe_unused]] i64 websocket_id, [[maybe_unused]] u16 code, [[maybe_unused]] ByteString const& reason) = 0;

    virtual Messages::RequestServer::WebsocketSetCertificateResponse websocket_set_certificate(
[[maybe_unused]] i64 request_id, [[maybe_unused]] ByteString const& certificate, [[maybe_unused]] ByteString const& key) = 0;

private:
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
