#pragma once
#include <LibCore/AnonymousBuffer.h>
#include <LibGfx/Color.h>
#include <LibGfx/ShareableBitmap.h>
#include <LibURL/URL.h>
#include <LibWeb/Cookie/Cookie.h>
#include <LibWeb/Cookie/ParsedCookie.h>
#include <LibWeb/CSS/Selector.h>
#include <LibWeb/CSS/StyleSheetIdentifier.h>
#include <LibWeb/HTML/ActivateTab.h>
#include <LibWeb/HTML/AudioPlayState.h>
#include <LibWeb/HTML/FileFilter.h>
#include <LibWeb/HTML/SelectedFile.h>
#include <LibWeb/HTML/SelectItem.h>
#include <LibWeb/HTML/WebViewHints.h>
#include <LibWeb/Page/EventResult.h>
#include <LibWeb/Page/Page.h>
#include <LibWebView/Attribute.h>
#include <LibWebView/ProcessHandle.h>
#include <LibWebView/PageInfo.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::WebContentClient {

enum class MessageID : i32 {
    DidStartLoading = 1,
    DidFinishLoading = 2,
    DidRequestRefresh = 3,
    DidPaint = 4,
    DidRequestCursorChange = 5,
    DidChangeTitle = 6,
    DidChangeUrl = 7,
    DidRequestTooltipOverride = 8,
    DidStopTooltipOverride = 9,
    DidEnterTooltipArea = 10,
    DidLeaveTooltipArea = 11,
    DidHoverLink = 12,
    DidUnhoverLink = 13,
    DidClickLink = 14,
    DidMiddleClickLink = 15,
    DidRequestContextMenu = 16,
    DidRequestLinkContextMenu = 17,
    DidRequestImageContextMenu = 18,
    DidRequestMediaContextMenu = 19,
    DidRequestAlert = 20,
    DidRequestConfirm = 21,
    DidRequestPrompt = 22,
    DidRequestSetPromptText = 23,
    DidRequestAcceptDialog = 24,
    DidRequestDismissDialog = 25,
    DidGetSource = 26,
    DidInspectDomTree = 27,
    DidInspectDomNode = 28,
    DidInspectAccessibilityTree = 29,
    DidGetHoveredNodeId = 30,
    DidFinishEditingDomNode = 31,
    DidGetDomNodeHtml = 32,
    InspectorDidListStyleSheets = 33,
    InspectorDidRequestStyleSheetSource = 34,
    DidGetStyleSheetSource = 35,
    DidTakeScreenshot = 36,
    DidGetInternalPageInfo = 37,
    DidChangeFavicon = 38,
    DidRequestAllCookies = 39,
    DidRequestAllCookiesResponse = 40,
    DidRequestNamedCookie = 41,
    DidRequestNamedCookieResponse = 42,
    DidRequestCookie = 43,
    DidRequestCookieResponse = 44,
    DidSetCookie = 45,
    DidSetCookieResponse = 46,
    DidUpdateCookie = 47,
    DidExpireCookiesWithTimeOffset = 48,
    DidUpdateResourceCount = 49,
    DidRequestNewWebView = 50,
    DidRequestNewWebViewResponse = 51,
    DidRequestActivateTab = 52,
    DidCloseBrowsingContext = 53,
    DidRequestRestoreWindow = 54,
    DidRequestRepositionWindow = 55,
    DidRequestResizeWindow = 56,
    DidRequestMaximizeWindow = 57,
    DidRequestMinimizeWindow = 58,
    DidRequestFullscreenWindow = 59,
    DidRequestFile = 60,
    DidRequestColorPicker = 61,
    DidRequestFilePicker = 62,
    DidRequestSelectDropdown = 63,
    DidFinishHandlingInputEvent = 64,
    DidChangeThemeColor = 65,
    DidInsertClipboardEntry = 66,
    DidUpdateNavigationButtonsState = 67,
    DidAllocateBackingStores = 68,
    DidChangeAudioPlayState = 69,
    DidOutputJsConsoleMessage = 70,
    DidGetJsConsoleMessages = 71,
    DidFinishTextTest = 72,
    DidSetTestTimeout = 73,
    DidSetBrowserZoom = 74,
    DidFindInPage = 75,
    RequestWorkerAgent = 76,
    RequestWorkerAgentResponse = 77,
    InspectorDidLoad = 78,
    InspectorDidSelectDomNode = 79,
    InspectorDidSetDomNodeText = 80,
    InspectorDidSetDomNodeTag = 81,
    InspectorDidAddDomNodeAttributes = 82,
    InspectorDidReplaceDomNodeAttribute = 83,
    InspectorDidRequestDomTreeContextMenu = 84,
    InspectorDidRequestCookieContextMenu = 85,
    InspectorDidExecuteConsoleScript = 86,
    InspectorDidExportInspectorHtml = 87,
};

class DidStartLoading final : public IPC::Message {
public:

    DidStartLoading(DidStartLoading const&) = default;
    DidStartLoading(DidStartLoading&&) = default;
    DidStartLoading& operator=(DidStartLoading const&) = default;
    DidStartLoading(u64 page_id, URL::URL url, bool is_redirect) : m_page_id(move(page_id)), m_url(move(url)), m_is_redirect(move(is_redirect)) {}

    virtual ~DidStartLoading() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidStartLoading; }
    static i32 static_message_id() { return (int)MessageID::DidStartLoading; }
    virtual const char* message_name() const override { return "WebContentClient::DidStartLoading"; }

    static ErrorOr<NonnullOwnPtr<DidStartLoading>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto is_redirect = TRY((decoder.decode<bool>()));

        return make<DidStartLoading>(move(page_id), move(url), move(is_redirect));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidStartLoading));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_is_redirect));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const bool& is_redirect() const { return m_is_redirect; }
    bool take_is_redirect() { return move(m_is_redirect); }

private:

    u64 m_page_id;

    URL::URL m_url;

    bool m_is_redirect;

};

class DidFinishLoading final : public IPC::Message {
public:

    DidFinishLoading(DidFinishLoading const&) = default;
    DidFinishLoading(DidFinishLoading&&) = default;
    DidFinishLoading& operator=(DidFinishLoading const&) = default;
    DidFinishLoading(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidFinishLoading() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishLoading; }
    static i32 static_message_id() { return (int)MessageID::DidFinishLoading; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishLoading"; }

    static ErrorOr<NonnullOwnPtr<DidFinishLoading>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        return make<DidFinishLoading>(move(page_id), move(url));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidFinishLoading));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:

    u64 m_page_id;

    URL::URL m_url;

};

class DidRequestRefresh final : public IPC::Message {
public:

    DidRequestRefresh(DidRequestRefresh const&) = default;
    DidRequestRefresh(DidRequestRefresh&&) = default;
    DidRequestRefresh& operator=(DidRequestRefresh const&) = default;
    DidRequestRefresh(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestRefresh(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestRefresh() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRefresh; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRefresh; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRefresh"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRefresh>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestRefresh>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestRefresh));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidPaint final : public IPC::Message {
public:

    DidPaint(DidPaint const&) = default;
    DidPaint(DidPaint&&) = default;
    DidPaint& operator=(DidPaint const&) = default;
    DidPaint(u64 page_id, Gfx::IntRect content_rect, i32 bitmap_id) : m_page_id(move(page_id)), m_content_rect(move(content_rect)), m_bitmap_id(move(bitmap_id)) {}

    virtual ~DidPaint() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidPaint; }
    static i32 static_message_id() { return (int)MessageID::DidPaint; }
    virtual const char* message_name() const override { return "WebContentClient::DidPaint"; }

    static ErrorOr<NonnullOwnPtr<DidPaint>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_rect = TRY((decoder.decode<Gfx::IntRect>()));

        auto bitmap_id = TRY((decoder.decode<i32>()));

        return make<DidPaint>(move(page_id), move(content_rect), move(bitmap_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidPaint));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_rect));

        TRY(stream.encode(m_bitmap_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntRect& content_rect() const { return m_content_rect; }
    Gfx::IntRect take_content_rect() { return move(m_content_rect); }

    const i32& bitmap_id() const { return m_bitmap_id; }
    i32 take_bitmap_id() { return move(m_bitmap_id); }

private:

    u64 m_page_id;

    Gfx::IntRect m_content_rect;

    i32 m_bitmap_id;

};

class DidRequestCursorChange final : public IPC::Message {
public:

    DidRequestCursorChange(DidRequestCursorChange const&) = default;
    DidRequestCursorChange(DidRequestCursorChange&&) = default;
    DidRequestCursorChange& operator=(DidRequestCursorChange const&) = default;
    DidRequestCursorChange(u64 page_id, i32 cursor_type) : m_page_id(move(page_id)), m_cursor_type(move(cursor_type)) {}

    virtual ~DidRequestCursorChange() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCursorChange; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCursorChange; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCursorChange"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCursorChange>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto cursor_type = TRY((decoder.decode<i32>()));

        return make<DidRequestCursorChange>(move(page_id), move(cursor_type));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestCursorChange));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_cursor_type));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& cursor_type() const { return m_cursor_type; }
    i32 take_cursor_type() { return move(m_cursor_type); }

private:

    u64 m_page_id;

    i32 m_cursor_type;

};

class DidChangeTitle final : public IPC::Message {
public:

    DidChangeTitle(DidChangeTitle const&) = default;
    DidChangeTitle(DidChangeTitle&&) = default;
    DidChangeTitle& operator=(DidChangeTitle const&) = default;
    DidChangeTitle(u64 page_id, ByteString title) : m_page_id(move(page_id)), m_title(move(title)) {}

    virtual ~DidChangeTitle() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeTitle; }
    static i32 static_message_id() { return (int)MessageID::DidChangeTitle; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeTitle"; }

    static ErrorOr<NonnullOwnPtr<DidChangeTitle>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto title = TRY((decoder.decode<ByteString>()));

        return make<DidChangeTitle>(move(page_id), move(title));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidChangeTitle));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_title));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:

    u64 m_page_id;

    ByteString m_title;

};

class DidChangeUrl final : public IPC::Message {
public:

    DidChangeUrl(DidChangeUrl const&) = default;
    DidChangeUrl(DidChangeUrl&&) = default;
    DidChangeUrl& operator=(DidChangeUrl const&) = default;
    DidChangeUrl(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidChangeUrl() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeUrl; }
    static i32 static_message_id() { return (int)MessageID::DidChangeUrl; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeUrl"; }

    static ErrorOr<NonnullOwnPtr<DidChangeUrl>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        return make<DidChangeUrl>(move(page_id), move(url));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidChangeUrl));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:

    u64 m_page_id;

    URL::URL m_url;

};

class DidRequestTooltipOverride final : public IPC::Message {
public:

    DidRequestTooltipOverride(DidRequestTooltipOverride const&) = default;
    DidRequestTooltipOverride(DidRequestTooltipOverride&&) = default;
    DidRequestTooltipOverride& operator=(DidRequestTooltipOverride const&) = default;
    DidRequestTooltipOverride(u64 page_id, Gfx::IntPoint position, ByteString title) : m_page_id(move(page_id)), m_position(move(position)), m_title(move(title)) {}

    virtual ~DidRequestTooltipOverride() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestTooltipOverride; }
    static i32 static_message_id() { return (int)MessageID::DidRequestTooltipOverride; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestTooltipOverride"; }

    static ErrorOr<NonnullOwnPtr<DidRequestTooltipOverride>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto title = TRY((decoder.decode<ByteString>()));

        return make<DidRequestTooltipOverride>(move(page_id), move(position), move(title));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestTooltipOverride));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_position));

        TRY(stream.encode(m_title));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& position() const { return m_position; }
    Gfx::IntPoint take_position() { return move(m_position); }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_position;

    ByteString m_title;

};

class DidStopTooltipOverride final : public IPC::Message {
public:

    DidStopTooltipOverride(DidStopTooltipOverride const&) = default;
    DidStopTooltipOverride(DidStopTooltipOverride&&) = default;
    DidStopTooltipOverride& operator=(DidStopTooltipOverride const&) = default;
    DidStopTooltipOverride(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidStopTooltipOverride(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidStopTooltipOverride() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidStopTooltipOverride; }
    static i32 static_message_id() { return (int)MessageID::DidStopTooltipOverride; }
    virtual const char* message_name() const override { return "WebContentClient::DidStopTooltipOverride"; }

    static ErrorOr<NonnullOwnPtr<DidStopTooltipOverride>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidStopTooltipOverride>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidStopTooltipOverride));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidEnterTooltipArea final : public IPC::Message {
public:

    DidEnterTooltipArea(DidEnterTooltipArea const&) = default;
    DidEnterTooltipArea(DidEnterTooltipArea&&) = default;
    DidEnterTooltipArea& operator=(DidEnterTooltipArea const&) = default;
    DidEnterTooltipArea(u64 page_id, ByteString title) : m_page_id(move(page_id)), m_title(move(title)) {}

    virtual ~DidEnterTooltipArea() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidEnterTooltipArea; }
    static i32 static_message_id() { return (int)MessageID::DidEnterTooltipArea; }
    virtual const char* message_name() const override { return "WebContentClient::DidEnterTooltipArea"; }

    static ErrorOr<NonnullOwnPtr<DidEnterTooltipArea>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto title = TRY((decoder.decode<ByteString>()));

        return make<DidEnterTooltipArea>(move(page_id), move(title));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidEnterTooltipArea));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_title));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& title() const { return m_title; }
    ByteString take_title() { return move(m_title); }

private:

    u64 m_page_id;

    ByteString m_title;

};

class DidLeaveTooltipArea final : public IPC::Message {
public:

    DidLeaveTooltipArea(DidLeaveTooltipArea const&) = default;
    DidLeaveTooltipArea(DidLeaveTooltipArea&&) = default;
    DidLeaveTooltipArea& operator=(DidLeaveTooltipArea const&) = default;
    DidLeaveTooltipArea(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidLeaveTooltipArea(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidLeaveTooltipArea() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidLeaveTooltipArea; }
    static i32 static_message_id() { return (int)MessageID::DidLeaveTooltipArea; }
    virtual const char* message_name() const override { return "WebContentClient::DidLeaveTooltipArea"; }

    static ErrorOr<NonnullOwnPtr<DidLeaveTooltipArea>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidLeaveTooltipArea>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidLeaveTooltipArea));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidHoverLink final : public IPC::Message {
public:

    DidHoverLink(DidHoverLink const&) = default;
    DidHoverLink(DidHoverLink&&) = default;
    DidHoverLink& operator=(DidHoverLink const&) = default;
    DidHoverLink(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~DidHoverLink() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidHoverLink; }
    static i32 static_message_id() { return (int)MessageID::DidHoverLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidHoverLink"; }

    static ErrorOr<NonnullOwnPtr<DidHoverLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        return make<DidHoverLink>(move(page_id), move(url));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidHoverLink));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:

    u64 m_page_id;

    URL::URL m_url;

};

class DidUnhoverLink final : public IPC::Message {
public:

    DidUnhoverLink(DidUnhoverLink const&) = default;
    DidUnhoverLink(DidUnhoverLink&&) = default;
    DidUnhoverLink& operator=(DidUnhoverLink const&) = default;
    DidUnhoverLink(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidUnhoverLink(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidUnhoverLink() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidUnhoverLink; }
    static i32 static_message_id() { return (int)MessageID::DidUnhoverLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidUnhoverLink"; }

    static ErrorOr<NonnullOwnPtr<DidUnhoverLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidUnhoverLink>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidUnhoverLink));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidClickLink final : public IPC::Message {
public:

    DidClickLink(DidClickLink const&) = default;
    DidClickLink(DidClickLink&&) = default;
    DidClickLink& operator=(DidClickLink const&) = default;
    DidClickLink(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidClickLink() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidClickLink; }
    static i32 static_message_id() { return (int)MessageID::DidClickLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidClickLink"; }

    static ErrorOr<NonnullOwnPtr<DidClickLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto target = TRY((decoder.decode<ByteString>()));

        auto modifiers = TRY((decoder.decode<unsigned>()));

        return make<DidClickLink>(move(page_id), move(url), move(target), move(modifiers));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidClickLink));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_target));

        TRY(stream.encode(m_modifiers));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    const unsigned& modifiers() const { return m_modifiers; }
    unsigned take_modifiers() { return move(m_modifiers); }

private:

    u64 m_page_id;

    URL::URL m_url;

    ByteString m_target;

    unsigned m_modifiers;

};

class DidMiddleClickLink final : public IPC::Message {
public:

    DidMiddleClickLink(DidMiddleClickLink const&) = default;
    DidMiddleClickLink(DidMiddleClickLink&&) = default;
    DidMiddleClickLink& operator=(DidMiddleClickLink const&) = default;
    DidMiddleClickLink(u64 page_id, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidMiddleClickLink() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidMiddleClickLink; }
    static i32 static_message_id() { return (int)MessageID::DidMiddleClickLink; }
    virtual const char* message_name() const override { return "WebContentClient::DidMiddleClickLink"; }

    static ErrorOr<NonnullOwnPtr<DidMiddleClickLink>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto target = TRY((decoder.decode<ByteString>()));

        auto modifiers = TRY((decoder.decode<unsigned>()));

        return make<DidMiddleClickLink>(move(page_id), move(url), move(target), move(modifiers));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidMiddleClickLink));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_target));

        TRY(stream.encode(m_modifiers));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    const unsigned& modifiers() const { return m_modifiers; }
    unsigned take_modifiers() { return move(m_modifiers); }

private:

    u64 m_page_id;

    URL::URL m_url;

    ByteString m_target;

    unsigned m_modifiers;

};

class DidRequestContextMenu final : public IPC::Message {
public:

    DidRequestContextMenu(DidRequestContextMenu const&) = default;
    DidRequestContextMenu(DidRequestContextMenu&&) = default;
    DidRequestContextMenu& operator=(DidRequestContextMenu const&) = default;
    DidRequestContextMenu(u64 page_id, Gfx::IntPoint content_position) : m_page_id(move(page_id)), m_content_position(move(content_position)) {}

    virtual ~DidRequestContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));

        return make<DidRequestContextMenu>(move(page_id), move(content_position));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_position));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& content_position() const { return m_content_position; }
    Gfx::IntPoint take_content_position() { return move(m_content_position); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_content_position;

};

class DidRequestLinkContextMenu final : public IPC::Message {
public:

    DidRequestLinkContextMenu(DidRequestLinkContextMenu const&) = default;
    DidRequestLinkContextMenu(DidRequestLinkContextMenu&&) = default;
    DidRequestLinkContextMenu& operator=(DidRequestLinkContextMenu const&) = default;
    DidRequestLinkContextMenu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)) {}

    virtual ~DidRequestLinkContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestLinkContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestLinkContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestLinkContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestLinkContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto target = TRY((decoder.decode<ByteString>()));

        auto modifiers = TRY((decoder.decode<unsigned>()));

        return make<DidRequestLinkContextMenu>(move(page_id), move(content_position), move(url), move(target), move(modifiers));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestLinkContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_position));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_target));

        TRY(stream.encode(m_modifiers));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& content_position() const { return m_content_position; }
    Gfx::IntPoint take_content_position() { return move(m_content_position); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    const unsigned& modifiers() const { return m_modifiers; }
    unsigned take_modifiers() { return move(m_modifiers); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_content_position;

    URL::URL m_url;

    ByteString m_target;

    unsigned m_modifiers;

};

class DidRequestImageContextMenu final : public IPC::Message {
public:

    DidRequestImageContextMenu(DidRequestImageContextMenu const&) = default;
    DidRequestImageContextMenu(DidRequestImageContextMenu&&) = default;
    DidRequestImageContextMenu& operator=(DidRequestImageContextMenu const&) = default;
    DidRequestImageContextMenu(u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers, Optional<Gfx::ShareableBitmap> bitmap) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_url(move(url)), m_target(move(target)), m_modifiers(move(modifiers)), m_bitmap(move(bitmap)) {}

    virtual ~DidRequestImageContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestImageContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestImageContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestImageContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestImageContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto target = TRY((decoder.decode<ByteString>()));

        auto modifiers = TRY((decoder.decode<unsigned>()));

        auto bitmap = TRY((decoder.decode<Optional<Gfx::ShareableBitmap>>()));

        return make<DidRequestImageContextMenu>(move(page_id), move(content_position), move(url), move(target), move(modifiers), move(bitmap));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestImageContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_position));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_target));

        TRY(stream.encode(m_modifiers));

        TRY(stream.encode(m_bitmap));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& content_position() const { return m_content_position; }
    Gfx::IntPoint take_content_position() { return move(m_content_position); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    const unsigned& modifiers() const { return m_modifiers; }
    unsigned take_modifiers() { return move(m_modifiers); }

    const Optional<Gfx::ShareableBitmap>& bitmap() const { return m_bitmap; }
    Optional<Gfx::ShareableBitmap> take_bitmap() { return move(m_bitmap); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_content_position;

    URL::URL m_url;

    ByteString m_target;

    unsigned m_modifiers;

    Optional<Gfx::ShareableBitmap> m_bitmap;

};

class DidRequestMediaContextMenu final : public IPC::Message {
public:

    DidRequestMediaContextMenu(DidRequestMediaContextMenu const&) = default;
    DidRequestMediaContextMenu(DidRequestMediaContextMenu&&) = default;
    DidRequestMediaContextMenu& operator=(DidRequestMediaContextMenu const&) = default;
    DidRequestMediaContextMenu(u64 page_id, Gfx::IntPoint content_position, ByteString target, unsigned modifiers, Web::Page::MediaContextMenu menu) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_target(move(target)), m_modifiers(move(modifiers)), m_menu(move(menu)) {}

    virtual ~DidRequestMediaContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMediaContextMenu; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMediaContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMediaContextMenu"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMediaContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto target = TRY((decoder.decode<ByteString>()));

        auto modifiers = TRY((decoder.decode<unsigned>()));

        auto menu = TRY((decoder.decode<Web::Page::MediaContextMenu>()));

        return make<DidRequestMediaContextMenu>(move(page_id), move(content_position), move(target), move(modifiers), move(menu));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestMediaContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_position));

        TRY(stream.encode(m_target));

        TRY(stream.encode(m_modifiers));

        TRY(stream.encode(m_menu));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& content_position() const { return m_content_position; }
    Gfx::IntPoint take_content_position() { return move(m_content_position); }

    const ByteString& target() const { return m_target; }
    ByteString take_target() { return move(m_target); }

    const unsigned& modifiers() const { return m_modifiers; }
    unsigned take_modifiers() { return move(m_modifiers); }

    const Web::Page::MediaContextMenu& menu() const { return m_menu; }
    Web::Page::MediaContextMenu take_menu() { return move(m_menu); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_content_position;

    ByteString m_target;

    unsigned m_modifiers;

    Web::Page::MediaContextMenu m_menu;

};

class DidRequestAlert final : public IPC::Message {
public:

    DidRequestAlert(DidRequestAlert const&) = default;
    DidRequestAlert(DidRequestAlert&&) = default;
    DidRequestAlert& operator=(DidRequestAlert const&) = default;
    DidRequestAlert(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestAlert() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAlert; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAlert; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAlert"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAlert>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto message = TRY((decoder.decode<String>()));

        return make<DidRequestAlert>(move(page_id), move(message));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestAlert));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_message));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:

    u64 m_page_id;

    String m_message;

};

class DidRequestConfirm final : public IPC::Message {
public:

    DidRequestConfirm(DidRequestConfirm const&) = default;
    DidRequestConfirm(DidRequestConfirm&&) = default;
    DidRequestConfirm& operator=(DidRequestConfirm const&) = default;
    DidRequestConfirm(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestConfirm() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestConfirm; }
    static i32 static_message_id() { return (int)MessageID::DidRequestConfirm; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestConfirm"; }

    static ErrorOr<NonnullOwnPtr<DidRequestConfirm>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto message = TRY((decoder.decode<String>()));

        return make<DidRequestConfirm>(move(page_id), move(message));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestConfirm));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_message));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:

    u64 m_page_id;

    String m_message;

};

class DidRequestPrompt final : public IPC::Message {
public:

    DidRequestPrompt(DidRequestPrompt const&) = default;
    DidRequestPrompt(DidRequestPrompt&&) = default;
    DidRequestPrompt& operator=(DidRequestPrompt const&) = default;
    DidRequestPrompt(u64 page_id, String message, String default_) : m_page_id(move(page_id)), m_message(move(message)), m_default_(move(default_)) {}

    virtual ~DidRequestPrompt() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestPrompt; }
    static i32 static_message_id() { return (int)MessageID::DidRequestPrompt; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestPrompt"; }

    static ErrorOr<NonnullOwnPtr<DidRequestPrompt>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto message = TRY((decoder.decode<String>()));

        auto default_ = TRY((decoder.decode<String>()));

        return make<DidRequestPrompt>(move(page_id), move(message), move(default_));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestPrompt));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_message));

        TRY(stream.encode(m_default_));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

    const String& default_() const { return m_default_; }
    String take_default_() { return move(m_default_); }

private:

    u64 m_page_id;

    String m_message;

    String m_default_;

};

class DidRequestSetPromptText final : public IPC::Message {
public:

    DidRequestSetPromptText(DidRequestSetPromptText const&) = default;
    DidRequestSetPromptText(DidRequestSetPromptText&&) = default;
    DidRequestSetPromptText& operator=(DidRequestSetPromptText const&) = default;
    DidRequestSetPromptText(u64 page_id, String message) : m_page_id(move(page_id)), m_message(move(message)) {}

    virtual ~DidRequestSetPromptText() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestSetPromptText; }
    static i32 static_message_id() { return (int)MessageID::DidRequestSetPromptText; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestSetPromptText"; }

    static ErrorOr<NonnullOwnPtr<DidRequestSetPromptText>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto message = TRY((decoder.decode<String>()));

        return make<DidRequestSetPromptText>(move(page_id), move(message));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestSetPromptText));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_message));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& message() const { return m_message; }
    String take_message() { return move(m_message); }

private:

    u64 m_page_id;

    String m_message;

};

class DidRequestAcceptDialog final : public IPC::Message {
public:

    DidRequestAcceptDialog(DidRequestAcceptDialog const&) = default;
    DidRequestAcceptDialog(DidRequestAcceptDialog&&) = default;
    DidRequestAcceptDialog& operator=(DidRequestAcceptDialog const&) = default;
    DidRequestAcceptDialog(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestAcceptDialog(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAcceptDialog() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAcceptDialog; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAcceptDialog; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAcceptDialog"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAcceptDialog>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestAcceptDialog>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestAcceptDialog));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestDismissDialog final : public IPC::Message {
public:

    DidRequestDismissDialog(DidRequestDismissDialog const&) = default;
    DidRequestDismissDialog(DidRequestDismissDialog&&) = default;
    DidRequestDismissDialog& operator=(DidRequestDismissDialog const&) = default;
    DidRequestDismissDialog(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestDismissDialog(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestDismissDialog() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestDismissDialog; }
    static i32 static_message_id() { return (int)MessageID::DidRequestDismissDialog; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestDismissDialog"; }

    static ErrorOr<NonnullOwnPtr<DidRequestDismissDialog>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestDismissDialog>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestDismissDialog));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidGetSource final : public IPC::Message {
public:

    DidGetSource(DidGetSource const&) = default;
    DidGetSource(DidGetSource&&) = default;
    DidGetSource& operator=(DidGetSource const&) = default;
    DidGetSource(u64 page_id, URL::URL url, URL::URL base_url, String source) : m_page_id(move(page_id)), m_url(move(url)), m_base_url(move(base_url)), m_source(move(source)) {}

    virtual ~DidGetSource() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetSource; }
    static i32 static_message_id() { return (int)MessageID::DidGetSource; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetSource"; }

    static ErrorOr<NonnullOwnPtr<DidGetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        auto base_url = TRY((decoder.decode<URL::URL>()));

        auto source = TRY((decoder.decode<String>()));

        return make<DidGetSource>(move(page_id), move(url), move(base_url), move(source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetSource));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_base_url));

        TRY(stream.encode(m_source));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const URL::URL& base_url() const { return m_base_url; }
    URL::URL take_base_url() { return move(m_base_url); }

    const String& source() const { return m_source; }
    String take_source() { return move(m_source); }

private:

    u64 m_page_id;

    URL::URL m_url;

    URL::URL m_base_url;

    String m_source;

};

class DidInspectDomTree final : public IPC::Message {
public:

    DidInspectDomTree(DidInspectDomTree const&) = default;
    DidInspectDomTree(DidInspectDomTree&&) = default;
    DidInspectDomTree& operator=(DidInspectDomTree const&) = default;
    DidInspectDomTree(u64 page_id, ByteString dom_tree) : m_page_id(move(page_id)), m_dom_tree(move(dom_tree)) {}

    virtual ~DidInspectDomTree() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectDomTree; }
    static i32 static_message_id() { return (int)MessageID::DidInspectDomTree; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectDomTree"; }

    static ErrorOr<NonnullOwnPtr<DidInspectDomTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto dom_tree = TRY((decoder.decode<ByteString>()));

        return make<DidInspectDomTree>(move(page_id), move(dom_tree));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidInspectDomTree));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_dom_tree));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& dom_tree() const { return m_dom_tree; }
    ByteString take_dom_tree() { return move(m_dom_tree); }

private:

    u64 m_page_id;

    ByteString m_dom_tree;

};

class DidInspectDomNode final : public IPC::Message {
public:

    DidInspectDomNode(DidInspectDomNode const&) = default;
    DidInspectDomNode(DidInspectDomNode&&) = default;
    DidInspectDomNode& operator=(DidInspectDomNode const&) = default;
    DidInspectDomNode(u64 page_id, bool has_style, ByteString computed_style, ByteString resolved_style, ByteString custom_properties, ByteString node_box_sizing, ByteString aria_properties_state, ByteString fonts) : m_page_id(move(page_id)), m_has_style(move(has_style)), m_computed_style(move(computed_style)), m_resolved_style(move(resolved_style)), m_custom_properties(move(custom_properties)), m_node_box_sizing(move(node_box_sizing)), m_aria_properties_state(move(aria_properties_state)), m_fonts(move(fonts)) {}

    virtual ~DidInspectDomNode() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectDomNode; }
    static i32 static_message_id() { return (int)MessageID::DidInspectDomNode; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectDomNode"; }

    static ErrorOr<NonnullOwnPtr<DidInspectDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto has_style = TRY((decoder.decode<bool>()));

        auto computed_style = TRY((decoder.decode<ByteString>()));

        auto resolved_style = TRY((decoder.decode<ByteString>()));

        auto custom_properties = TRY((decoder.decode<ByteString>()));

        auto node_box_sizing = TRY((decoder.decode<ByteString>()));

        auto aria_properties_state = TRY((decoder.decode<ByteString>()));

        auto fonts = TRY((decoder.decode<ByteString>()));

        return make<DidInspectDomNode>(move(page_id), move(has_style), move(computed_style), move(resolved_style), move(custom_properties), move(node_box_sizing), move(aria_properties_state), move(fonts));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidInspectDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_has_style));

        TRY(stream.encode(m_computed_style));

        TRY(stream.encode(m_resolved_style));

        TRY(stream.encode(m_custom_properties));

        TRY(stream.encode(m_node_box_sizing));

        TRY(stream.encode(m_aria_properties_state));

        TRY(stream.encode(m_fonts));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& has_style() const { return m_has_style; }
    bool take_has_style() { return move(m_has_style); }

    const ByteString& computed_style() const { return m_computed_style; }
    ByteString take_computed_style() { return move(m_computed_style); }

    const ByteString& resolved_style() const { return m_resolved_style; }
    ByteString take_resolved_style() { return move(m_resolved_style); }

    const ByteString& custom_properties() const { return m_custom_properties; }
    ByteString take_custom_properties() { return move(m_custom_properties); }

    const ByteString& node_box_sizing() const { return m_node_box_sizing; }
    ByteString take_node_box_sizing() { return move(m_node_box_sizing); }

    const ByteString& aria_properties_state() const { return m_aria_properties_state; }
    ByteString take_aria_properties_state() { return move(m_aria_properties_state); }

    const ByteString& fonts() const { return m_fonts; }
    ByteString take_fonts() { return move(m_fonts); }

private:

    u64 m_page_id;

    bool m_has_style;

    ByteString m_computed_style;

    ByteString m_resolved_style;

    ByteString m_custom_properties;

    ByteString m_node_box_sizing;

    ByteString m_aria_properties_state;

    ByteString m_fonts;

};

class DidInspectAccessibilityTree final : public IPC::Message {
public:

    DidInspectAccessibilityTree(DidInspectAccessibilityTree const&) = default;
    DidInspectAccessibilityTree(DidInspectAccessibilityTree&&) = default;
    DidInspectAccessibilityTree& operator=(DidInspectAccessibilityTree const&) = default;
    DidInspectAccessibilityTree(u64 page_id, ByteString accessibility_tree) : m_page_id(move(page_id)), m_accessibility_tree(move(accessibility_tree)) {}

    virtual ~DidInspectAccessibilityTree() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidInspectAccessibilityTree; }
    static i32 static_message_id() { return (int)MessageID::DidInspectAccessibilityTree; }
    virtual const char* message_name() const override { return "WebContentClient::DidInspectAccessibilityTree"; }

    static ErrorOr<NonnullOwnPtr<DidInspectAccessibilityTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto accessibility_tree = TRY((decoder.decode<ByteString>()));

        return make<DidInspectAccessibilityTree>(move(page_id), move(accessibility_tree));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidInspectAccessibilityTree));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_accessibility_tree));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& accessibility_tree() const { return m_accessibility_tree; }
    ByteString take_accessibility_tree() { return move(m_accessibility_tree); }

private:

    u64 m_page_id;

    ByteString m_accessibility_tree;

};

class DidGetHoveredNodeId final : public IPC::Message {
public:

    DidGetHoveredNodeId(DidGetHoveredNodeId const&) = default;
    DidGetHoveredNodeId(DidGetHoveredNodeId&&) = default;
    DidGetHoveredNodeId& operator=(DidGetHoveredNodeId const&) = default;
    DidGetHoveredNodeId(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~DidGetHoveredNodeId() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetHoveredNodeId; }
    static i32 static_message_id() { return (int)MessageID::DidGetHoveredNodeId; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetHoveredNodeId"; }

    static ErrorOr<NonnullOwnPtr<DidGetHoveredNodeId>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<DidGetHoveredNodeId>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetHoveredNodeId));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class DidFinishEditingDomNode final : public IPC::Message {
public:

    DidFinishEditingDomNode(DidFinishEditingDomNode const&) = default;
    DidFinishEditingDomNode(DidFinishEditingDomNode&&) = default;
    DidFinishEditingDomNode& operator=(DidFinishEditingDomNode const&) = default;
    DidFinishEditingDomNode(u64 page_id, Optional<Web::UniqueNodeID> node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~DidFinishEditingDomNode() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishEditingDomNode; }
    static i32 static_message_id() { return (int)MessageID::DidFinishEditingDomNode; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishEditingDomNode"; }

    static ErrorOr<NonnullOwnPtr<DidFinishEditingDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Optional<Web::UniqueNodeID>>()));

        return make<DidFinishEditingDomNode>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidFinishEditingDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Optional<Web::UniqueNodeID>& node_id() const { return m_node_id; }
    Optional<Web::UniqueNodeID> take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Optional<Web::UniqueNodeID> m_node_id;

};

class DidGetDomNodeHtml final : public IPC::Message {
public:

    DidGetDomNodeHtml(DidGetDomNodeHtml const&) = default;
    DidGetDomNodeHtml(DidGetDomNodeHtml&&) = default;
    DidGetDomNodeHtml& operator=(DidGetDomNodeHtml const&) = default;
    DidGetDomNodeHtml(u64 page_id, String html) : m_page_id(move(page_id)), m_html(move(html)) {}

    virtual ~DidGetDomNodeHtml() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetDomNodeHtml; }
    static i32 static_message_id() { return (int)MessageID::DidGetDomNodeHtml; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetDomNodeHtml"; }

    static ErrorOr<NonnullOwnPtr<DidGetDomNodeHtml>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto html = TRY((decoder.decode<String>()));

        return make<DidGetDomNodeHtml>(move(page_id), move(html));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetDomNodeHtml));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_html));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& html() const { return m_html; }
    String take_html() { return move(m_html); }

private:

    u64 m_page_id;

    String m_html;

};

class InspectorDidListStyleSheets final : public IPC::Message {
public:

    InspectorDidListStyleSheets(InspectorDidListStyleSheets const&) = default;
    InspectorDidListStyleSheets(InspectorDidListStyleSheets&&) = default;
    InspectorDidListStyleSheets& operator=(InspectorDidListStyleSheets const&) = default;
    InspectorDidListStyleSheets(u64 page_id, Vector<Web::CSS::StyleSheetIdentifier> style_sheets) : m_page_id(move(page_id)), m_style_sheets(move(style_sheets)) {}

    virtual ~InspectorDidListStyleSheets() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidListStyleSheets; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidListStyleSheets; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidListStyleSheets"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidListStyleSheets>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto style_sheets = TRY((decoder.decode<Vector<Web::CSS::StyleSheetIdentifier>>()));

        return make<InspectorDidListStyleSheets>(move(page_id), move(style_sheets));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidListStyleSheets));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_style_sheets));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<Web::CSS::StyleSheetIdentifier>& style_sheets() const { return m_style_sheets; }
    Vector<Web::CSS::StyleSheetIdentifier> take_style_sheets() { return move(m_style_sheets); }

private:

    u64 m_page_id;

    Vector<Web::CSS::StyleSheetIdentifier> m_style_sheets;

};

class InspectorDidRequestStyleSheetSource final : public IPC::Message {
public:

    InspectorDidRequestStyleSheetSource(InspectorDidRequestStyleSheetSource const&) = default;
    InspectorDidRequestStyleSheetSource(InspectorDidRequestStyleSheetSource&&) = default;
    InspectorDidRequestStyleSheetSource& operator=(InspectorDidRequestStyleSheetSource const&) = default;
    InspectorDidRequestStyleSheetSource(u64 page_id, Web::CSS::StyleSheetIdentifier identifier) : m_page_id(move(page_id)), m_identifier(move(identifier)) {}

    virtual ~InspectorDidRequestStyleSheetSource() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidRequestStyleSheetSource; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidRequestStyleSheetSource; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidRequestStyleSheetSource"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidRequestStyleSheetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto identifier = TRY((decoder.decode<Web::CSS::StyleSheetIdentifier>()));

        return make<InspectorDidRequestStyleSheetSource>(move(page_id), move(identifier));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidRequestStyleSheetSource));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_identifier));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::StyleSheetIdentifier& identifier() const { return m_identifier; }
    Web::CSS::StyleSheetIdentifier take_identifier() { return move(m_identifier); }

private:

    u64 m_page_id;

    Web::CSS::StyleSheetIdentifier m_identifier;

};

class DidGetStyleSheetSource final : public IPC::Message {
public:

    DidGetStyleSheetSource(DidGetStyleSheetSource const&) = default;
    DidGetStyleSheetSource(DidGetStyleSheetSource&&) = default;
    DidGetStyleSheetSource& operator=(DidGetStyleSheetSource const&) = default;
    DidGetStyleSheetSource(u64 page_id, Web::CSS::StyleSheetIdentifier identifier, URL::URL base_url, String source) : m_page_id(move(page_id)), m_identifier(move(identifier)), m_base_url(move(base_url)), m_source(move(source)) {}

    virtual ~DidGetStyleSheetSource() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetStyleSheetSource; }
    static i32 static_message_id() { return (int)MessageID::DidGetStyleSheetSource; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetStyleSheetSource"; }

    static ErrorOr<NonnullOwnPtr<DidGetStyleSheetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto identifier = TRY((decoder.decode<Web::CSS::StyleSheetIdentifier>()));

        auto base_url = TRY((decoder.decode<URL::URL>()));

        auto source = TRY((decoder.decode<String>()));

        return make<DidGetStyleSheetSource>(move(page_id), move(identifier), move(base_url), move(source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetStyleSheetSource));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_identifier));

        TRY(stream.encode(m_base_url));

        TRY(stream.encode(m_source));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::StyleSheetIdentifier& identifier() const { return m_identifier; }
    Web::CSS::StyleSheetIdentifier take_identifier() { return move(m_identifier); }

    const URL::URL& base_url() const { return m_base_url; }
    URL::URL take_base_url() { return move(m_base_url); }

    const String& source() const { return m_source; }
    String take_source() { return move(m_source); }

private:

    u64 m_page_id;

    Web::CSS::StyleSheetIdentifier m_identifier;

    URL::URL m_base_url;

    String m_source;

};

class DidTakeScreenshot final : public IPC::Message {
public:

    DidTakeScreenshot(DidTakeScreenshot const&) = default;
    DidTakeScreenshot(DidTakeScreenshot&&) = default;
    DidTakeScreenshot& operator=(DidTakeScreenshot const&) = default;
    DidTakeScreenshot(u64 page_id, Gfx::ShareableBitmap screenshot) : m_page_id(move(page_id)), m_screenshot(move(screenshot)) {}

    virtual ~DidTakeScreenshot() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidTakeScreenshot; }
    static i32 static_message_id() { return (int)MessageID::DidTakeScreenshot; }
    virtual const char* message_name() const override { return "WebContentClient::DidTakeScreenshot"; }

    static ErrorOr<NonnullOwnPtr<DidTakeScreenshot>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto screenshot = TRY((decoder.decode<Gfx::ShareableBitmap>()));

        return make<DidTakeScreenshot>(move(page_id), move(screenshot));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidTakeScreenshot));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_screenshot));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::ShareableBitmap& screenshot() const { return m_screenshot; }
    Gfx::ShareableBitmap take_screenshot() { return move(m_screenshot); }

private:

    u64 m_page_id;

    Gfx::ShareableBitmap m_screenshot;

};

class DidGetInternalPageInfo final : public IPC::Message {
public:

    DidGetInternalPageInfo(DidGetInternalPageInfo const&) = default;
    DidGetInternalPageInfo(DidGetInternalPageInfo&&) = default;
    DidGetInternalPageInfo& operator=(DidGetInternalPageInfo const&) = default;
    DidGetInternalPageInfo(u64 page_id, WebView::PageInfoType type, String info) : m_page_id(move(page_id)), m_type(move(type)), m_info(move(info)) {}

    virtual ~DidGetInternalPageInfo() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetInternalPageInfo; }
    static i32 static_message_id() { return (int)MessageID::DidGetInternalPageInfo; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetInternalPageInfo"; }

    static ErrorOr<NonnullOwnPtr<DidGetInternalPageInfo>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto type = TRY((decoder.decode<WebView::PageInfoType>()));

        auto info = TRY((decoder.decode<String>()));

        return make<DidGetInternalPageInfo>(move(page_id), move(type), move(info));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetInternalPageInfo));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_type));

        TRY(stream.encode(m_info));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const WebView::PageInfoType& type() const { return m_type; }
    WebView::PageInfoType take_type() { return move(m_type); }

    const String& info() const { return m_info; }
    String take_info() { return move(m_info); }

private:

    u64 m_page_id;

    WebView::PageInfoType m_type;

    String m_info;

};

class DidChangeFavicon final : public IPC::Message {
public:

    DidChangeFavicon(DidChangeFavicon const&) = default;
    DidChangeFavicon(DidChangeFavicon&&) = default;
    DidChangeFavicon& operator=(DidChangeFavicon const&) = default;
    DidChangeFavicon(u64 page_id, Gfx::ShareableBitmap favicon) : m_page_id(move(page_id)), m_favicon(move(favicon)) {}

    virtual ~DidChangeFavicon() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeFavicon; }
    static i32 static_message_id() { return (int)MessageID::DidChangeFavicon; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeFavicon"; }

    static ErrorOr<NonnullOwnPtr<DidChangeFavicon>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto favicon = TRY((decoder.decode<Gfx::ShareableBitmap>()));

        return make<DidChangeFavicon>(move(page_id), move(favicon));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidChangeFavicon));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_favicon));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::ShareableBitmap& favicon() const { return m_favicon; }
    Gfx::ShareableBitmap take_favicon() { return move(m_favicon); }

private:

    u64 m_page_id;

    Gfx::ShareableBitmap m_favicon;

};

class DidRequestAllCookiesResponse final : public IPC::Message {
public:

    DidRequestAllCookiesResponse(DidRequestAllCookiesResponse const&) = default;
    DidRequestAllCookiesResponse(DidRequestAllCookiesResponse&&) = default;
    DidRequestAllCookiesResponse& operator=(DidRequestAllCookiesResponse const&) = default;
    DidRequestAllCookiesResponse(Vector<Web::Cookie::Cookie> cookies) : m_cookies(move(cookies)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Vector<Web::Cookie::Cookie>>)
    DidRequestAllCookiesResponse(WrappedReturnType&& value)
        : m_cookies(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAllCookiesResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAllCookiesResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAllCookiesResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAllCookiesResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAllCookiesResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto cookies = TRY((decoder.decode<Vector<Web::Cookie::Cookie>>()));

        return make<DidRequestAllCookiesResponse>(move(cookies));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestAllCookiesResponse));

        TRY(stream.encode(m_cookies));

        return buffer;
    }

    const Vector<Web::Cookie::Cookie>& cookies() const { return m_cookies; }
    Vector<Web::Cookie::Cookie> take_cookies() { return move(m_cookies); }

private:

    Vector<Web::Cookie::Cookie> m_cookies;

};

class DidRequestAllCookies final : public IPC::Message {
public:

   typedef class DidRequestAllCookiesResponse ResponseType;

    DidRequestAllCookies(DidRequestAllCookies const&) = default;
    DidRequestAllCookies(DidRequestAllCookies&&) = default;
    DidRequestAllCookies& operator=(DidRequestAllCookies const&) = default;
    DidRequestAllCookies(URL::URL url) : m_url(move(url)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, URL::URL>)
    DidRequestAllCookies(WrappedReturnType&& value)
        : m_url(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestAllCookies() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestAllCookies; }
    static i32 static_message_id() { return (int)MessageID::DidRequestAllCookies; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestAllCookies"; }

    static ErrorOr<NonnullOwnPtr<DidRequestAllCookies>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto url = TRY((decoder.decode<URL::URL>()));

        return make<DidRequestAllCookies>(move(url));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestAllCookies));

        TRY(stream.encode(m_url));

        return buffer;
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:

    URL::URL m_url;

};

class DidRequestNamedCookieResponse final : public IPC::Message {
public:

    DidRequestNamedCookieResponse(DidRequestNamedCookieResponse const&) = default;
    DidRequestNamedCookieResponse(DidRequestNamedCookieResponse&&) = default;
    DidRequestNamedCookieResponse& operator=(DidRequestNamedCookieResponse const&) = default;
    DidRequestNamedCookieResponse(Optional<Web::Cookie::Cookie> cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Optional<Web::Cookie::Cookie>>)
    DidRequestNamedCookieResponse(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestNamedCookieResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNamedCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNamedCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNamedCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNamedCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto cookie = TRY((decoder.decode<Optional<Web::Cookie::Cookie>>()));

        return make<DidRequestNamedCookieResponse>(move(cookie));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestNamedCookieResponse));

        TRY(stream.encode(m_cookie));

        return buffer;
    }

    const Optional<Web::Cookie::Cookie>& cookie() const { return m_cookie; }
    Optional<Web::Cookie::Cookie> take_cookie() { return move(m_cookie); }

private:

    Optional<Web::Cookie::Cookie> m_cookie;

};

class DidRequestNamedCookie final : public IPC::Message {
public:

   typedef class DidRequestNamedCookieResponse ResponseType;

    DidRequestNamedCookie(DidRequestNamedCookie const&) = default;
    DidRequestNamedCookie(DidRequestNamedCookie&&) = default;
    DidRequestNamedCookie& operator=(DidRequestNamedCookie const&) = default;
    DidRequestNamedCookie(URL::URL url, String name) : m_url(move(url)), m_name(move(name)) {}

    virtual ~DidRequestNamedCookie() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNamedCookie; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNamedCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNamedCookie"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNamedCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto url = TRY((decoder.decode<URL::URL>()));

        auto name = TRY((decoder.decode<String>()));

        return make<DidRequestNamedCookie>(move(url), move(name));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestNamedCookie));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_name));

        return buffer;
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

private:

    URL::URL m_url;

    String m_name;

};

class DidRequestCookieResponse final : public IPC::Message {
public:

    DidRequestCookieResponse(DidRequestCookieResponse const&) = default;
    DidRequestCookieResponse(DidRequestCookieResponse&&) = default;
    DidRequestCookieResponse& operator=(DidRequestCookieResponse const&) = default;
    DidRequestCookieResponse(String cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, String>)
    DidRequestCookieResponse(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestCookieResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto cookie = TRY((decoder.decode<String>()));

        return make<DidRequestCookieResponse>(move(cookie));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestCookieResponse));

        TRY(stream.encode(m_cookie));

        return buffer;
    }

    const String& cookie() const { return m_cookie; }
    String take_cookie() { return move(m_cookie); }

private:

    String m_cookie;

};

class DidRequestCookie final : public IPC::Message {
public:

   typedef class DidRequestCookieResponse ResponseType;

    DidRequestCookie(DidRequestCookie const&) = default;
    DidRequestCookie(DidRequestCookie&&) = default;
    DidRequestCookie& operator=(DidRequestCookie const&) = default;
    DidRequestCookie(URL::URL url, Web::Cookie::Source source) : m_url(move(url)), m_source(move(source)) {}

    virtual ~DidRequestCookie() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestCookie; }
    static i32 static_message_id() { return (int)MessageID::DidRequestCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestCookie"; }

    static ErrorOr<NonnullOwnPtr<DidRequestCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto url = TRY((decoder.decode<URL::URL>()));

        auto source = TRY((decoder.decode<Web::Cookie::Source>()));

        return make<DidRequestCookie>(move(url), move(source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestCookie));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_source));

        return buffer;
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const Web::Cookie::Source& source() const { return m_source; }
    Web::Cookie::Source take_source() { return move(m_source); }

private:

    URL::URL m_url;

    Web::Cookie::Source m_source;

};

class DidSetCookieResponse final : public IPC::Message {
public:

    DidSetCookieResponse(DidSetCookieResponse const&) = default;
    DidSetCookieResponse(DidSetCookieResponse&&) = default;
    DidSetCookieResponse& operator=(DidSetCookieResponse const&) = default;
    DidSetCookieResponse() {}

    virtual ~DidSetCookieResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetCookieResponse; }
    static i32 static_message_id() { return (int)MessageID::DidSetCookieResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetCookieResponse"; }

    static ErrorOr<NonnullOwnPtr<DidSetCookieResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        return make<DidSetCookieResponse>();
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidSetCookieResponse));

        return buffer;
    }

private:

};

class DidSetCookie final : public IPC::Message {
public:

   typedef class DidSetCookieResponse ResponseType;

    DidSetCookie(DidSetCookie const&) = default;
    DidSetCookie(DidSetCookie&&) = default;
    DidSetCookie& operator=(DidSetCookie const&) = default;
    DidSetCookie(URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) : m_url(move(url)), m_cookie(move(cookie)), m_source(move(source)) {}

    virtual ~DidSetCookie() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetCookie; }
    static i32 static_message_id() { return (int)MessageID::DidSetCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetCookie"; }

    static ErrorOr<NonnullOwnPtr<DidSetCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto url = TRY((decoder.decode<URL::URL>()));

        auto cookie = TRY((decoder.decode<Web::Cookie::ParsedCookie>()));

        auto source = TRY((decoder.decode<Web::Cookie::Source>()));

        return make<DidSetCookie>(move(url), move(cookie), move(source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidSetCookie));

        TRY(stream.encode(m_url));

        TRY(stream.encode(m_cookie));

        TRY(stream.encode(m_source));

        return buffer;
    }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

    const Web::Cookie::ParsedCookie& cookie() const { return m_cookie; }
    Web::Cookie::ParsedCookie take_cookie() { return move(m_cookie); }

    const Web::Cookie::Source& source() const { return m_source; }
    Web::Cookie::Source take_source() { return move(m_source); }

private:

    URL::URL m_url;

    Web::Cookie::ParsedCookie m_cookie;

    Web::Cookie::Source m_source;

};

class DidUpdateCookie final : public IPC::Message {
public:

    DidUpdateCookie(DidUpdateCookie const&) = default;
    DidUpdateCookie(DidUpdateCookie&&) = default;
    DidUpdateCookie& operator=(DidUpdateCookie const&) = default;
    DidUpdateCookie(Web::Cookie::Cookie cookie) : m_cookie(move(cookie)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, Web::Cookie::Cookie>)
    DidUpdateCookie(WrappedReturnType&& value)
        : m_cookie(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidUpdateCookie() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateCookie; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateCookie; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateCookie"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateCookie>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto cookie = TRY((decoder.decode<Web::Cookie::Cookie>()));

        return make<DidUpdateCookie>(move(cookie));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidUpdateCookie));

        TRY(stream.encode(m_cookie));

        return buffer;
    }

    const Web::Cookie::Cookie& cookie() const { return m_cookie; }
    Web::Cookie::Cookie take_cookie() { return move(m_cookie); }

private:

    Web::Cookie::Cookie m_cookie;

};

class DidExpireCookiesWithTimeOffset final : public IPC::Message {
public:

    DidExpireCookiesWithTimeOffset(DidExpireCookiesWithTimeOffset const&) = default;
    DidExpireCookiesWithTimeOffset(DidExpireCookiesWithTimeOffset&&) = default;
    DidExpireCookiesWithTimeOffset& operator=(DidExpireCookiesWithTimeOffset const&) = default;
    DidExpireCookiesWithTimeOffset(AK::Duration offset) : m_offset(move(offset)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, AK::Duration>)
    DidExpireCookiesWithTimeOffset(WrappedReturnType&& value)
        : m_offset(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidExpireCookiesWithTimeOffset() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidExpireCookiesWithTimeOffset; }
    static i32 static_message_id() { return (int)MessageID::DidExpireCookiesWithTimeOffset; }
    virtual const char* message_name() const override { return "WebContentClient::DidExpireCookiesWithTimeOffset"; }

    static ErrorOr<NonnullOwnPtr<DidExpireCookiesWithTimeOffset>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto offset = TRY((decoder.decode<AK::Duration>()));

        return make<DidExpireCookiesWithTimeOffset>(move(offset));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidExpireCookiesWithTimeOffset));

        TRY(stream.encode(m_offset));

        return buffer;
    }

    const AK::Duration& offset() const { return m_offset; }
    AK::Duration take_offset() { return move(m_offset); }

private:

    AK::Duration m_offset;

};

class DidUpdateResourceCount final : public IPC::Message {
public:

    DidUpdateResourceCount(DidUpdateResourceCount const&) = default;
    DidUpdateResourceCount(DidUpdateResourceCount&&) = default;
    DidUpdateResourceCount& operator=(DidUpdateResourceCount const&) = default;
    DidUpdateResourceCount(u64 page_id, i32 count_waiting) : m_page_id(move(page_id)), m_count_waiting(move(count_waiting)) {}

    virtual ~DidUpdateResourceCount() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateResourceCount; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateResourceCount; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateResourceCount"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateResourceCount>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto count_waiting = TRY((decoder.decode<i32>()));

        return make<DidUpdateResourceCount>(move(page_id), move(count_waiting));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidUpdateResourceCount));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_count_waiting));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& count_waiting() const { return m_count_waiting; }
    i32 take_count_waiting() { return move(m_count_waiting); }

private:

    u64 m_page_id;

    i32 m_count_waiting;

};

class DidRequestNewWebViewResponse final : public IPC::Message {
public:

    DidRequestNewWebViewResponse(DidRequestNewWebViewResponse const&) = default;
    DidRequestNewWebViewResponse(DidRequestNewWebViewResponse&&) = default;
    DidRequestNewWebViewResponse& operator=(DidRequestNewWebViewResponse const&) = default;
    DidRequestNewWebViewResponse(String handle) : m_handle(move(handle)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, String>)
    DidRequestNewWebViewResponse(WrappedReturnType&& value)
        : m_handle(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestNewWebViewResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNewWebViewResponse; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNewWebViewResponse; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNewWebViewResponse"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNewWebViewResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto handle = TRY((decoder.decode<String>()));

        return make<DidRequestNewWebViewResponse>(move(handle));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestNewWebViewResponse));

        TRY(stream.encode(m_handle));

        return buffer;
    }

    const String& handle() const { return m_handle; }
    String take_handle() { return move(m_handle); }

private:

    String m_handle;

};

class DidRequestNewWebView final : public IPC::Message {
public:

   typedef class DidRequestNewWebViewResponse ResponseType;

    DidRequestNewWebView(DidRequestNewWebView const&) = default;
    DidRequestNewWebView(DidRequestNewWebView&&) = default;
    DidRequestNewWebView& operator=(DidRequestNewWebView const&) = default;
    DidRequestNewWebView(u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) : m_page_id(move(page_id)), m_activate_tab(move(activate_tab)), m_hints(move(hints)), m_page_index(move(page_index)) {}

    virtual ~DidRequestNewWebView() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestNewWebView; }
    static i32 static_message_id() { return (int)MessageID::DidRequestNewWebView; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestNewWebView"; }

    static ErrorOr<NonnullOwnPtr<DidRequestNewWebView>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto activate_tab = TRY((decoder.decode<Web::HTML::ActivateTab>()));

        auto hints = TRY((decoder.decode<Web::HTML::WebViewHints>()));

        auto page_index = TRY((decoder.decode<Optional<u64>>()));

        return make<DidRequestNewWebView>(move(page_id), move(activate_tab), move(hints), move(page_index));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestNewWebView));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_activate_tab));

        TRY(stream.encode(m_hints));

        TRY(stream.encode(m_page_index));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::HTML::ActivateTab& activate_tab() const { return m_activate_tab; }
    Web::HTML::ActivateTab take_activate_tab() { return move(m_activate_tab); }

    const Web::HTML::WebViewHints& hints() const { return m_hints; }
    Web::HTML::WebViewHints take_hints() { return move(m_hints); }

    const Optional<u64>& page_index() const { return m_page_index; }
    Optional<u64> take_page_index() { return move(m_page_index); }

private:

    u64 m_page_id;

    Web::HTML::ActivateTab m_activate_tab;

    Web::HTML::WebViewHints m_hints;

    Optional<u64> m_page_index;

};

class DidRequestActivateTab final : public IPC::Message {
public:

    DidRequestActivateTab(DidRequestActivateTab const&) = default;
    DidRequestActivateTab(DidRequestActivateTab&&) = default;
    DidRequestActivateTab& operator=(DidRequestActivateTab const&) = default;
    DidRequestActivateTab(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestActivateTab(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestActivateTab() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestActivateTab; }
    static i32 static_message_id() { return (int)MessageID::DidRequestActivateTab; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestActivateTab"; }

    static ErrorOr<NonnullOwnPtr<DidRequestActivateTab>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestActivateTab>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestActivateTab));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidCloseBrowsingContext final : public IPC::Message {
public:

    DidCloseBrowsingContext(DidCloseBrowsingContext const&) = default;
    DidCloseBrowsingContext(DidCloseBrowsingContext&&) = default;
    DidCloseBrowsingContext& operator=(DidCloseBrowsingContext const&) = default;
    DidCloseBrowsingContext(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidCloseBrowsingContext(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidCloseBrowsingContext() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidCloseBrowsingContext; }
    static i32 static_message_id() { return (int)MessageID::DidCloseBrowsingContext; }
    virtual const char* message_name() const override { return "WebContentClient::DidCloseBrowsingContext"; }

    static ErrorOr<NonnullOwnPtr<DidCloseBrowsingContext>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidCloseBrowsingContext>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidCloseBrowsingContext));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestRestoreWindow final : public IPC::Message {
public:

    DidRequestRestoreWindow(DidRequestRestoreWindow const&) = default;
    DidRequestRestoreWindow(DidRequestRestoreWindow&&) = default;
    DidRequestRestoreWindow& operator=(DidRequestRestoreWindow const&) = default;
    DidRequestRestoreWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestRestoreWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestRestoreWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRestoreWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRestoreWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRestoreWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRestoreWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestRestoreWindow>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestRestoreWindow));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestRepositionWindow final : public IPC::Message {
public:

    DidRequestRepositionWindow(DidRequestRepositionWindow const&) = default;
    DidRequestRepositionWindow(DidRequestRepositionWindow&&) = default;
    DidRequestRepositionWindow& operator=(DidRequestRepositionWindow const&) = default;
    DidRequestRepositionWindow(u64 page_id, Gfx::IntPoint position) : m_page_id(move(page_id)), m_position(move(position)) {}

    virtual ~DidRequestRepositionWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestRepositionWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestRepositionWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestRepositionWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestRepositionWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto position = TRY((decoder.decode<Gfx::IntPoint>()));

        return make<DidRequestRepositionWindow>(move(page_id), move(position));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestRepositionWindow));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_position));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& position() const { return m_position; }
    Gfx::IntPoint take_position() { return move(m_position); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_position;

};

class DidRequestResizeWindow final : public IPC::Message {
public:

    DidRequestResizeWindow(DidRequestResizeWindow const&) = default;
    DidRequestResizeWindow(DidRequestResizeWindow&&) = default;
    DidRequestResizeWindow& operator=(DidRequestResizeWindow const&) = default;
    DidRequestResizeWindow(u64 page_id, Gfx::IntSize size) : m_page_id(move(page_id)), m_size(move(size)) {}

    virtual ~DidRequestResizeWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestResizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestResizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestResizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestResizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto size = TRY((decoder.decode<Gfx::IntSize>()));

        return make<DidRequestResizeWindow>(move(page_id), move(size));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestResizeWindow));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_size));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntSize& size() const { return m_size; }
    Gfx::IntSize take_size() { return move(m_size); }

private:

    u64 m_page_id;

    Gfx::IntSize m_size;

};

class DidRequestMaximizeWindow final : public IPC::Message {
public:

    DidRequestMaximizeWindow(DidRequestMaximizeWindow const&) = default;
    DidRequestMaximizeWindow(DidRequestMaximizeWindow&&) = default;
    DidRequestMaximizeWindow& operator=(DidRequestMaximizeWindow const&) = default;
    DidRequestMaximizeWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestMaximizeWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestMaximizeWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMaximizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMaximizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMaximizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMaximizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestMaximizeWindow>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestMaximizeWindow));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestMinimizeWindow final : public IPC::Message {
public:

    DidRequestMinimizeWindow(DidRequestMinimizeWindow const&) = default;
    DidRequestMinimizeWindow(DidRequestMinimizeWindow&&) = default;
    DidRequestMinimizeWindow& operator=(DidRequestMinimizeWindow const&) = default;
    DidRequestMinimizeWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestMinimizeWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestMinimizeWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestMinimizeWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestMinimizeWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestMinimizeWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestMinimizeWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestMinimizeWindow>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestMinimizeWindow));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestFullscreenWindow final : public IPC::Message {
public:

    DidRequestFullscreenWindow(DidRequestFullscreenWindow const&) = default;
    DidRequestFullscreenWindow(DidRequestFullscreenWindow&&) = default;
    DidRequestFullscreenWindow& operator=(DidRequestFullscreenWindow const&) = default;
    DidRequestFullscreenWindow(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidRequestFullscreenWindow(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidRequestFullscreenWindow() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFullscreenWindow; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFullscreenWindow; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFullscreenWindow"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFullscreenWindow>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidRequestFullscreenWindow>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestFullscreenWindow));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class DidRequestFile final : public IPC::Message {
public:

    DidRequestFile(DidRequestFile const&) = default;
    DidRequestFile(DidRequestFile&&) = default;
    DidRequestFile& operator=(DidRequestFile const&) = default;
    DidRequestFile(u64 page_id, ByteString path, i32 request_id) : m_page_id(move(page_id)), m_path(move(path)), m_request_id(move(request_id)) {}

    virtual ~DidRequestFile() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFile; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFile; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFile"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFile>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto path = TRY((decoder.decode<ByteString>()));

        auto request_id = TRY((decoder.decode<i32>()));

        return make<DidRequestFile>(move(page_id), move(path), move(request_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestFile));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_path));

        TRY(stream.encode(m_request_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& path() const { return m_path; }
    ByteString take_path() { return move(m_path); }

    const i32& request_id() const { return m_request_id; }
    i32 take_request_id() { return move(m_request_id); }

private:

    u64 m_page_id;

    ByteString m_path;

    i32 m_request_id;

};

class DidRequestColorPicker final : public IPC::Message {
public:

    DidRequestColorPicker(DidRequestColorPicker const&) = default;
    DidRequestColorPicker(DidRequestColorPicker&&) = default;
    DidRequestColorPicker& operator=(DidRequestColorPicker const&) = default;
    DidRequestColorPicker(u64 page_id, Color current_color) : m_page_id(move(page_id)), m_current_color(move(current_color)) {}

    virtual ~DidRequestColorPicker() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestColorPicker; }
    static i32 static_message_id() { return (int)MessageID::DidRequestColorPicker; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestColorPicker"; }

    static ErrorOr<NonnullOwnPtr<DidRequestColorPicker>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto current_color = TRY((decoder.decode<Color>()));

        return make<DidRequestColorPicker>(move(page_id), move(current_color));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestColorPicker));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_current_color));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Color& current_color() const { return m_current_color; }
    Color take_current_color() { return move(m_current_color); }

private:

    u64 m_page_id;

    Color m_current_color;

};

class DidRequestFilePicker final : public IPC::Message {
public:

    DidRequestFilePicker(DidRequestFilePicker const&) = default;
    DidRequestFilePicker(DidRequestFilePicker&&) = default;
    DidRequestFilePicker& operator=(DidRequestFilePicker const&) = default;
    DidRequestFilePicker(u64 page_id, Web::HTML::FileFilter accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files) : m_page_id(move(page_id)), m_accepted_file_types(move(accepted_file_types)), m_allow_multiple_files(move(allow_multiple_files)) {}

    virtual ~DidRequestFilePicker() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestFilePicker; }
    static i32 static_message_id() { return (int)MessageID::DidRequestFilePicker; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestFilePicker"; }

    static ErrorOr<NonnullOwnPtr<DidRequestFilePicker>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto accepted_file_types = TRY((decoder.decode<Web::HTML::FileFilter>()));

        auto allow_multiple_files = TRY((decoder.decode<Web::HTML::AllowMultipleFiles>()));

        return make<DidRequestFilePicker>(move(page_id), move(accepted_file_types), move(allow_multiple_files));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestFilePicker));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_accepted_file_types));

        TRY(stream.encode(m_allow_multiple_files));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::HTML::FileFilter& accepted_file_types() const { return m_accepted_file_types; }
    Web::HTML::FileFilter take_accepted_file_types() { return move(m_accepted_file_types); }

    const Web::HTML::AllowMultipleFiles& allow_multiple_files() const { return m_allow_multiple_files; }
    Web::HTML::AllowMultipleFiles take_allow_multiple_files() { return move(m_allow_multiple_files); }

private:

    u64 m_page_id;

    Web::HTML::FileFilter m_accepted_file_types;

    Web::HTML::AllowMultipleFiles m_allow_multiple_files;

};

class DidRequestSelectDropdown final : public IPC::Message {
public:

    DidRequestSelectDropdown(DidRequestSelectDropdown const&) = default;
    DidRequestSelectDropdown(DidRequestSelectDropdown&&) = default;
    DidRequestSelectDropdown& operator=(DidRequestSelectDropdown const&) = default;
    DidRequestSelectDropdown(u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, Vector<Web::HTML::SelectItem> items) : m_page_id(move(page_id)), m_content_position(move(content_position)), m_minimum_width(move(minimum_width)), m_items(move(items)) {}

    virtual ~DidRequestSelectDropdown() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidRequestSelectDropdown; }
    static i32 static_message_id() { return (int)MessageID::DidRequestSelectDropdown; }
    virtual const char* message_name() const override { return "WebContentClient::DidRequestSelectDropdown"; }

    static ErrorOr<NonnullOwnPtr<DidRequestSelectDropdown>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto content_position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto minimum_width = TRY((decoder.decode<i32>()));

        auto items = TRY((decoder.decode<Vector<Web::HTML::SelectItem>>()));

        return make<DidRequestSelectDropdown>(move(page_id), move(content_position), move(minimum_width), move(items));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidRequestSelectDropdown));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_content_position));

        TRY(stream.encode(m_minimum_width));

        TRY(stream.encode(m_items));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::IntPoint& content_position() const { return m_content_position; }
    Gfx::IntPoint take_content_position() { return move(m_content_position); }

    const i32& minimum_width() const { return m_minimum_width; }
    i32 take_minimum_width() { return move(m_minimum_width); }

    const Vector<Web::HTML::SelectItem>& items() const { return m_items; }
    Vector<Web::HTML::SelectItem> take_items() { return move(m_items); }

private:

    u64 m_page_id;

    Gfx::IntPoint m_content_position;

    i32 m_minimum_width;

    Vector<Web::HTML::SelectItem> m_items;

};

class DidFinishHandlingInputEvent final : public IPC::Message {
public:

    DidFinishHandlingInputEvent(DidFinishHandlingInputEvent const&) = default;
    DidFinishHandlingInputEvent(DidFinishHandlingInputEvent&&) = default;
    DidFinishHandlingInputEvent& operator=(DidFinishHandlingInputEvent const&) = default;
    DidFinishHandlingInputEvent(u64 page_id, Web::EventResult event_result) : m_page_id(move(page_id)), m_event_result(move(event_result)) {}

    virtual ~DidFinishHandlingInputEvent() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishHandlingInputEvent; }
    static i32 static_message_id() { return (int)MessageID::DidFinishHandlingInputEvent; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishHandlingInputEvent"; }

    static ErrorOr<NonnullOwnPtr<DidFinishHandlingInputEvent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto event_result = TRY((decoder.decode<Web::EventResult>()));

        return make<DidFinishHandlingInputEvent>(move(page_id), move(event_result));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidFinishHandlingInputEvent));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_event_result));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::EventResult& event_result() const { return m_event_result; }
    Web::EventResult take_event_result() { return move(m_event_result); }

private:

    u64 m_page_id;

    Web::EventResult m_event_result;

};

class DidChangeThemeColor final : public IPC::Message {
public:

    DidChangeThemeColor(DidChangeThemeColor const&) = default;
    DidChangeThemeColor(DidChangeThemeColor&&) = default;
    DidChangeThemeColor& operator=(DidChangeThemeColor const&) = default;
    DidChangeThemeColor(u64 page_id, Gfx::Color color) : m_page_id(move(page_id)), m_color(move(color)) {}

    virtual ~DidChangeThemeColor() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeThemeColor; }
    static i32 static_message_id() { return (int)MessageID::DidChangeThemeColor; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeThemeColor"; }

    static ErrorOr<NonnullOwnPtr<DidChangeThemeColor>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto color = TRY((decoder.decode<Gfx::Color>()));

        return make<DidChangeThemeColor>(move(page_id), move(color));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidChangeThemeColor));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_color));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Gfx::Color& color() const { return m_color; }
    Gfx::Color take_color() { return move(m_color); }

private:

    u64 m_page_id;

    Gfx::Color m_color;

};

class DidInsertClipboardEntry final : public IPC::Message {
public:

    DidInsertClipboardEntry(DidInsertClipboardEntry const&) = default;
    DidInsertClipboardEntry(DidInsertClipboardEntry&&) = default;
    DidInsertClipboardEntry& operator=(DidInsertClipboardEntry const&) = default;
    DidInsertClipboardEntry(u64 page_id, String data, String presentation_style, String mime_type) : m_page_id(move(page_id)), m_data(move(data)), m_presentation_style(move(presentation_style)), m_mime_type(move(mime_type)) {}

    virtual ~DidInsertClipboardEntry() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidInsertClipboardEntry; }
    static i32 static_message_id() { return (int)MessageID::DidInsertClipboardEntry; }
    virtual const char* message_name() const override { return "WebContentClient::DidInsertClipboardEntry"; }

    static ErrorOr<NonnullOwnPtr<DidInsertClipboardEntry>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto data = TRY((decoder.decode<String>()));

        auto presentation_style = TRY((decoder.decode<String>()));

        auto mime_type = TRY((decoder.decode<String>()));

        return make<DidInsertClipboardEntry>(move(page_id), move(data), move(presentation_style), move(mime_type));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidInsertClipboardEntry));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_data));

        TRY(stream.encode(m_presentation_style));

        TRY(stream.encode(m_mime_type));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& data() const { return m_data; }
    String take_data() { return move(m_data); }

    const String& presentation_style() const { return m_presentation_style; }
    String take_presentation_style() { return move(m_presentation_style); }

    const String& mime_type() const { return m_mime_type; }
    String take_mime_type() { return move(m_mime_type); }

private:

    u64 m_page_id;

    String m_data;

    String m_presentation_style;

    String m_mime_type;

};

class DidUpdateNavigationButtonsState final : public IPC::Message {
public:

    DidUpdateNavigationButtonsState(DidUpdateNavigationButtonsState const&) = default;
    DidUpdateNavigationButtonsState(DidUpdateNavigationButtonsState&&) = default;
    DidUpdateNavigationButtonsState& operator=(DidUpdateNavigationButtonsState const&) = default;
    DidUpdateNavigationButtonsState(u64 page_id, bool back_enabled, bool forward_enabled) : m_page_id(move(page_id)), m_back_enabled(move(back_enabled)), m_forward_enabled(move(forward_enabled)) {}

    virtual ~DidUpdateNavigationButtonsState() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateNavigationButtonsState; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateNavigationButtonsState; }
    virtual const char* message_name() const override { return "WebContentClient::DidUpdateNavigationButtonsState"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateNavigationButtonsState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto back_enabled = TRY((decoder.decode<bool>()));

        auto forward_enabled = TRY((decoder.decode<bool>()));

        return make<DidUpdateNavigationButtonsState>(move(page_id), move(back_enabled), move(forward_enabled));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidUpdateNavigationButtonsState));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_back_enabled));

        TRY(stream.encode(m_forward_enabled));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& back_enabled() const { return m_back_enabled; }
    bool take_back_enabled() { return move(m_back_enabled); }

    const bool& forward_enabled() const { return m_forward_enabled; }
    bool take_forward_enabled() { return move(m_forward_enabled); }

private:

    u64 m_page_id;

    bool m_back_enabled;

    bool m_forward_enabled;

};

class DidAllocateBackingStores final : public IPC::Message {
public:

    DidAllocateBackingStores(DidAllocateBackingStores const&) = default;
    DidAllocateBackingStores(DidAllocateBackingStores&&) = default;
    DidAllocateBackingStores& operator=(DidAllocateBackingStores const&) = default;
    DidAllocateBackingStores(u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap back_bitmap) : m_page_id(move(page_id)), m_front_bitmap_id(move(front_bitmap_id)), m_front_bitmap(move(front_bitmap)), m_back_bitmap_id(move(back_bitmap_id)), m_back_bitmap(move(back_bitmap)) {}

    virtual ~DidAllocateBackingStores() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidAllocateBackingStores; }
    static i32 static_message_id() { return (int)MessageID::DidAllocateBackingStores; }
    virtual const char* message_name() const override { return "WebContentClient::DidAllocateBackingStores"; }

    static ErrorOr<NonnullOwnPtr<DidAllocateBackingStores>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto front_bitmap_id = TRY((decoder.decode<i32>()));

        auto front_bitmap = TRY((decoder.decode<Gfx::ShareableBitmap>()));

        auto back_bitmap_id = TRY((decoder.decode<i32>()));

        auto back_bitmap = TRY((decoder.decode<Gfx::ShareableBitmap>()));

        return make<DidAllocateBackingStores>(move(page_id), move(front_bitmap_id), move(front_bitmap), move(back_bitmap_id), move(back_bitmap));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidAllocateBackingStores));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_front_bitmap_id));

        TRY(stream.encode(m_front_bitmap));

        TRY(stream.encode(m_back_bitmap_id));

        TRY(stream.encode(m_back_bitmap));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& front_bitmap_id() const { return m_front_bitmap_id; }
    i32 take_front_bitmap_id() { return move(m_front_bitmap_id); }

    const Gfx::ShareableBitmap& front_bitmap() const { return m_front_bitmap; }
    Gfx::ShareableBitmap take_front_bitmap() { return move(m_front_bitmap); }

    const i32& back_bitmap_id() const { return m_back_bitmap_id; }
    i32 take_back_bitmap_id() { return move(m_back_bitmap_id); }

    const Gfx::ShareableBitmap& back_bitmap() const { return m_back_bitmap; }
    Gfx::ShareableBitmap take_back_bitmap() { return move(m_back_bitmap); }

private:

    u64 m_page_id;

    i32 m_front_bitmap_id;

    Gfx::ShareableBitmap m_front_bitmap;

    i32 m_back_bitmap_id;

    Gfx::ShareableBitmap m_back_bitmap;

};

class DidChangeAudioPlayState final : public IPC::Message {
public:

    DidChangeAudioPlayState(DidChangeAudioPlayState const&) = default;
    DidChangeAudioPlayState(DidChangeAudioPlayState&&) = default;
    DidChangeAudioPlayState& operator=(DidChangeAudioPlayState const&) = default;
    DidChangeAudioPlayState(u64 page_id, Web::HTML::AudioPlayState play_state) : m_page_id(move(page_id)), m_play_state(move(play_state)) {}

    virtual ~DidChangeAudioPlayState() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidChangeAudioPlayState; }
    static i32 static_message_id() { return (int)MessageID::DidChangeAudioPlayState; }
    virtual const char* message_name() const override { return "WebContentClient::DidChangeAudioPlayState"; }

    static ErrorOr<NonnullOwnPtr<DidChangeAudioPlayState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto play_state = TRY((decoder.decode<Web::HTML::AudioPlayState>()));

        return make<DidChangeAudioPlayState>(move(page_id), move(play_state));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidChangeAudioPlayState));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_play_state));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::HTML::AudioPlayState& play_state() const { return m_play_state; }
    Web::HTML::AudioPlayState take_play_state() { return move(m_play_state); }

private:

    u64 m_page_id;

    Web::HTML::AudioPlayState m_play_state;

};

class DidOutputJsConsoleMessage final : public IPC::Message {
public:

    DidOutputJsConsoleMessage(DidOutputJsConsoleMessage const&) = default;
    DidOutputJsConsoleMessage(DidOutputJsConsoleMessage&&) = default;
    DidOutputJsConsoleMessage& operator=(DidOutputJsConsoleMessage const&) = default;
    DidOutputJsConsoleMessage(u64 page_id, i32 message_index) : m_page_id(move(page_id)), m_message_index(move(message_index)) {}

    virtual ~DidOutputJsConsoleMessage() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidOutputJsConsoleMessage; }
    static i32 static_message_id() { return (int)MessageID::DidOutputJsConsoleMessage; }
    virtual const char* message_name() const override { return "WebContentClient::DidOutputJsConsoleMessage"; }

    static ErrorOr<NonnullOwnPtr<DidOutputJsConsoleMessage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto message_index = TRY((decoder.decode<i32>()));

        return make<DidOutputJsConsoleMessage>(move(page_id), move(message_index));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidOutputJsConsoleMessage));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_message_index));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& message_index() const { return m_message_index; }
    i32 take_message_index() { return move(m_message_index); }

private:

    u64 m_page_id;

    i32 m_message_index;

};

class DidGetJsConsoleMessages final : public IPC::Message {
public:

    DidGetJsConsoleMessages(DidGetJsConsoleMessages const&) = default;
    DidGetJsConsoleMessages(DidGetJsConsoleMessages&&) = default;
    DidGetJsConsoleMessages& operator=(DidGetJsConsoleMessages const&) = default;
    DidGetJsConsoleMessages(u64 page_id, i32 start_index, Vector<ByteString> message_types, Vector<ByteString> messages) : m_page_id(move(page_id)), m_start_index(move(start_index)), m_message_types(move(message_types)), m_messages(move(messages)) {}

    virtual ~DidGetJsConsoleMessages() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidGetJsConsoleMessages; }
    static i32 static_message_id() { return (int)MessageID::DidGetJsConsoleMessages; }
    virtual const char* message_name() const override { return "WebContentClient::DidGetJsConsoleMessages"; }

    static ErrorOr<NonnullOwnPtr<DidGetJsConsoleMessages>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto start_index = TRY((decoder.decode<i32>()));

        auto message_types = TRY((decoder.decode<Vector<ByteString>>()));

        auto messages = TRY((decoder.decode<Vector<ByteString>>()));

        return make<DidGetJsConsoleMessages>(move(page_id), move(start_index), move(message_types), move(messages));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidGetJsConsoleMessages));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_start_index));

        TRY(stream.encode(m_message_types));

        TRY(stream.encode(m_messages));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& start_index() const { return m_start_index; }
    i32 take_start_index() { return move(m_start_index); }

    const Vector<ByteString>& message_types() const { return m_message_types; }
    Vector<ByteString> take_message_types() { return move(m_message_types); }

    const Vector<ByteString>& messages() const { return m_messages; }
    Vector<ByteString> take_messages() { return move(m_messages); }

private:

    u64 m_page_id;

    i32 m_start_index;

    Vector<ByteString> m_message_types;

    Vector<ByteString> m_messages;

};

class DidFinishTextTest final : public IPC::Message {
public:

    DidFinishTextTest(DidFinishTextTest const&) = default;
    DidFinishTextTest(DidFinishTextTest&&) = default;
    DidFinishTextTest& operator=(DidFinishTextTest const&) = default;
    DidFinishTextTest(u64 page_id, String text) : m_page_id(move(page_id)), m_text(move(text)) {}

    virtual ~DidFinishTextTest() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidFinishTextTest; }
    static i32 static_message_id() { return (int)MessageID::DidFinishTextTest; }
    virtual const char* message_name() const override { return "WebContentClient::DidFinishTextTest"; }

    static ErrorOr<NonnullOwnPtr<DidFinishTextTest>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto text = TRY((decoder.decode<String>()));

        return make<DidFinishTextTest>(move(page_id), move(text));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidFinishTextTest));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_text));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& text() const { return m_text; }
    String take_text() { return move(m_text); }

private:

    u64 m_page_id;

    String m_text;

};

class DidSetTestTimeout final : public IPC::Message {
public:

    DidSetTestTimeout(DidSetTestTimeout const&) = default;
    DidSetTestTimeout(DidSetTestTimeout&&) = default;
    DidSetTestTimeout& operator=(DidSetTestTimeout const&) = default;
    DidSetTestTimeout(u64 page_id, double milliseconds) : m_page_id(move(page_id)), m_milliseconds(move(milliseconds)) {}

    virtual ~DidSetTestTimeout() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetTestTimeout; }
    static i32 static_message_id() { return (int)MessageID::DidSetTestTimeout; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetTestTimeout"; }

    static ErrorOr<NonnullOwnPtr<DidSetTestTimeout>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto milliseconds = TRY((decoder.decode<double>()));

        return make<DidSetTestTimeout>(move(page_id), move(milliseconds));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidSetTestTimeout));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_milliseconds));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const double& milliseconds() const { return m_milliseconds; }
    double take_milliseconds() { return move(m_milliseconds); }

private:

    u64 m_page_id;

    double m_milliseconds;

};

class DidSetBrowserZoom final : public IPC::Message {
public:

    DidSetBrowserZoom(DidSetBrowserZoom const&) = default;
    DidSetBrowserZoom(DidSetBrowserZoom&&) = default;
    DidSetBrowserZoom& operator=(DidSetBrowserZoom const&) = default;
    DidSetBrowserZoom(u64 page_id, double factor) : m_page_id(move(page_id)), m_factor(move(factor)) {}

    virtual ~DidSetBrowserZoom() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidSetBrowserZoom; }
    static i32 static_message_id() { return (int)MessageID::DidSetBrowserZoom; }
    virtual const char* message_name() const override { return "WebContentClient::DidSetBrowserZoom"; }

    static ErrorOr<NonnullOwnPtr<DidSetBrowserZoom>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto factor = TRY((decoder.decode<double>()));

        return make<DidSetBrowserZoom>(move(page_id), move(factor));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidSetBrowserZoom));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_factor));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const double& factor() const { return m_factor; }
    double take_factor() { return move(m_factor); }

private:

    u64 m_page_id;

    double m_factor;

};

class DidFindInPage final : public IPC::Message {
public:

    DidFindInPage(DidFindInPage const&) = default;
    DidFindInPage(DidFindInPage&&) = default;
    DidFindInPage& operator=(DidFindInPage const&) = default;
    DidFindInPage(u64 page_id, size_t current_match_index, Optional<size_t> total_match_count) : m_page_id(move(page_id)), m_current_match_index(move(current_match_index)), m_total_match_count(move(total_match_count)) {}

    virtual ~DidFindInPage() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::DidFindInPage; }
    static i32 static_message_id() { return (int)MessageID::DidFindInPage; }
    virtual const char* message_name() const override { return "WebContentClient::DidFindInPage"; }

    static ErrorOr<NonnullOwnPtr<DidFindInPage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto current_match_index = TRY((decoder.decode<size_t>()));

        auto total_match_count = TRY((decoder.decode<Optional<size_t>>()));

        return make<DidFindInPage>(move(page_id), move(current_match_index), move(total_match_count));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidFindInPage));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_current_match_index));

        TRY(stream.encode(m_total_match_count));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const size_t& current_match_index() const { return m_current_match_index; }
    size_t take_current_match_index() { return move(m_current_match_index); }

    const Optional<size_t>& total_match_count() const { return m_total_match_count; }
    Optional<size_t> take_total_match_count() { return move(m_total_match_count); }

private:

    u64 m_page_id;

    size_t m_current_match_index;

    Optional<size_t> m_total_match_count;

};

class RequestWorkerAgentResponse final : public IPC::Message {
public:

    RequestWorkerAgentResponse(RequestWorkerAgentResponse const&) = default;
    RequestWorkerAgentResponse(RequestWorkerAgentResponse&&) = default;
    RequestWorkerAgentResponse& operator=(RequestWorkerAgentResponse const&) = default;
    RequestWorkerAgentResponse(IPC::File socket) : m_socket(move(socket)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, IPC::File>)
    RequestWorkerAgentResponse(WrappedReturnType&& value)
        : m_socket(forward<WrappedReturnType>(value))
    {
    }

    virtual ~RequestWorkerAgentResponse() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::RequestWorkerAgentResponse; }
    static i32 static_message_id() { return (int)MessageID::RequestWorkerAgentResponse; }
    virtual const char* message_name() const override { return "WebContentClient::RequestWorkerAgentResponse"; }

    static ErrorOr<NonnullOwnPtr<RequestWorkerAgentResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto socket = TRY((decoder.decode<IPC::File>()));

        return make<RequestWorkerAgentResponse>(move(socket));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RequestWorkerAgentResponse));

        TRY(stream.encode(m_socket));

        return buffer;
    }

    const IPC::File& socket() const { return m_socket; }
    IPC::File take_socket() { return move(m_socket); }

private:

    IPC::File m_socket;

};

class RequestWorkerAgent final : public IPC::Message {
public:

   typedef class RequestWorkerAgentResponse ResponseType;

    RequestWorkerAgent(RequestWorkerAgent const&) = default;
    RequestWorkerAgent(RequestWorkerAgent&&) = default;
    RequestWorkerAgent& operator=(RequestWorkerAgent const&) = default;
    RequestWorkerAgent(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    RequestWorkerAgent(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~RequestWorkerAgent() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::RequestWorkerAgent; }
    static i32 static_message_id() { return (int)MessageID::RequestWorkerAgent; }
    virtual const char* message_name() const override { return "WebContentClient::RequestWorkerAgent"; }

    static ErrorOr<NonnullOwnPtr<RequestWorkerAgent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<RequestWorkerAgent>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RequestWorkerAgent));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class InspectorDidLoad final : public IPC::Message {
public:

    InspectorDidLoad(InspectorDidLoad const&) = default;
    InspectorDidLoad(InspectorDidLoad&&) = default;
    InspectorDidLoad& operator=(InspectorDidLoad const&) = default;
    InspectorDidLoad(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    InspectorDidLoad(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InspectorDidLoad() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidLoad; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidLoad; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidLoad"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidLoad>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<InspectorDidLoad>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidLoad));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class InspectorDidSelectDomNode final : public IPC::Message {
public:

    InspectorDidSelectDomNode(InspectorDidSelectDomNode const&) = default;
    InspectorDidSelectDomNode(InspectorDidSelectDomNode&&) = default;
    InspectorDidSelectDomNode& operator=(InspectorDidSelectDomNode const&) = default;
    InspectorDidSelectDomNode(u64 page_id, Web::UniqueNodeID node_id, Optional<Web::CSS::Selector::PseudoElement::Type> pseudo_element) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_pseudo_element(move(pseudo_element)) {}

    virtual ~InspectorDidSelectDomNode() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidSelectDomNode; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidSelectDomNode; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidSelectDomNode"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidSelectDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto pseudo_element = TRY((decoder.decode<Optional<Web::CSS::Selector::PseudoElement::Type>>()));

        return make<InspectorDidSelectDomNode>(move(page_id), move(node_id), move(pseudo_element));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidSelectDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_pseudo_element));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const Optional<Web::CSS::Selector::PseudoElement::Type>& pseudo_element() const { return m_pseudo_element; }
    Optional<Web::CSS::Selector::PseudoElement::Type> take_pseudo_element() { return move(m_pseudo_element); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    Optional<Web::CSS::Selector::PseudoElement::Type> m_pseudo_element;

};

class InspectorDidSetDomNodeText final : public IPC::Message {
public:

    InspectorDidSetDomNodeText(InspectorDidSetDomNodeText const&) = default;
    InspectorDidSetDomNodeText(InspectorDidSetDomNodeText&&) = default;
    InspectorDidSetDomNodeText& operator=(InspectorDidSetDomNodeText const&) = default;
    InspectorDidSetDomNodeText(u64 page_id, Web::UniqueNodeID node_id, String text) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_text(move(text)) {}

    virtual ~InspectorDidSetDomNodeText() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidSetDomNodeText; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidSetDomNodeText; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidSetDomNodeText"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidSetDomNodeText>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto text = TRY((decoder.decode<String>()));

        return make<InspectorDidSetDomNodeText>(move(page_id), move(node_id), move(text));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidSetDomNodeText));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_text));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const String& text() const { return m_text; }
    String take_text() { return move(m_text); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    String m_text;

};

class InspectorDidSetDomNodeTag final : public IPC::Message {
public:

    InspectorDidSetDomNodeTag(InspectorDidSetDomNodeTag const&) = default;
    InspectorDidSetDomNodeTag(InspectorDidSetDomNodeTag&&) = default;
    InspectorDidSetDomNodeTag& operator=(InspectorDidSetDomNodeTag const&) = default;
    InspectorDidSetDomNodeTag(u64 page_id, Web::UniqueNodeID node_id, String tag) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_tag(move(tag)) {}

    virtual ~InspectorDidSetDomNodeTag() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidSetDomNodeTag; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidSetDomNodeTag; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidSetDomNodeTag"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidSetDomNodeTag>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto tag = TRY((decoder.decode<String>()));

        return make<InspectorDidSetDomNodeTag>(move(page_id), move(node_id), move(tag));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidSetDomNodeTag));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_tag));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const String& tag() const { return m_tag; }
    String take_tag() { return move(m_tag); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    String m_tag;

};

class InspectorDidAddDomNodeAttributes final : public IPC::Message {
public:

    InspectorDidAddDomNodeAttributes(InspectorDidAddDomNodeAttributes const&) = default;
    InspectorDidAddDomNodeAttributes(InspectorDidAddDomNodeAttributes&&) = default;
    InspectorDidAddDomNodeAttributes& operator=(InspectorDidAddDomNodeAttributes const&) = default;
    InspectorDidAddDomNodeAttributes(u64 page_id, Web::UniqueNodeID node_id, Vector<WebView::Attribute> attributes) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_attributes(move(attributes)) {}

    virtual ~InspectorDidAddDomNodeAttributes() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidAddDomNodeAttributes; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidAddDomNodeAttributes; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidAddDomNodeAttributes"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidAddDomNodeAttributes>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto attributes = TRY((decoder.decode<Vector<WebView::Attribute>>()));

        return make<InspectorDidAddDomNodeAttributes>(move(page_id), move(node_id), move(attributes));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidAddDomNodeAttributes));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_attributes));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const Vector<WebView::Attribute>& attributes() const { return m_attributes; }
    Vector<WebView::Attribute> take_attributes() { return move(m_attributes); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    Vector<WebView::Attribute> m_attributes;

};

class InspectorDidReplaceDomNodeAttribute final : public IPC::Message {
public:

    InspectorDidReplaceDomNodeAttribute(InspectorDidReplaceDomNodeAttribute const&) = default;
    InspectorDidReplaceDomNodeAttribute(InspectorDidReplaceDomNodeAttribute&&) = default;
    InspectorDidReplaceDomNodeAttribute& operator=(InspectorDidReplaceDomNodeAttribute const&) = default;
    InspectorDidReplaceDomNodeAttribute(u64 page_id, Web::UniqueNodeID node_id, size_t attribute_index, Vector<WebView::Attribute> replacement_attributes) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_attribute_index(move(attribute_index)), m_replacement_attributes(move(replacement_attributes)) {}

    virtual ~InspectorDidReplaceDomNodeAttribute() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidReplaceDomNodeAttribute; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidReplaceDomNodeAttribute; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidReplaceDomNodeAttribute"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidReplaceDomNodeAttribute>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto attribute_index = TRY((decoder.decode<size_t>()));

        auto replacement_attributes = TRY((decoder.decode<Vector<WebView::Attribute>>()));

        return make<InspectorDidReplaceDomNodeAttribute>(move(page_id), move(node_id), move(attribute_index), move(replacement_attributes));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidReplaceDomNodeAttribute));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_attribute_index));

        TRY(stream.encode(m_replacement_attributes));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const size_t& attribute_index() const { return m_attribute_index; }
    size_t take_attribute_index() { return move(m_attribute_index); }

    const Vector<WebView::Attribute>& replacement_attributes() const { return m_replacement_attributes; }
    Vector<WebView::Attribute> take_replacement_attributes() { return move(m_replacement_attributes); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    size_t m_attribute_index;

    Vector<WebView::Attribute> m_replacement_attributes;

};

class InspectorDidRequestDomTreeContextMenu final : public IPC::Message {
public:

    InspectorDidRequestDomTreeContextMenu(InspectorDidRequestDomTreeContextMenu const&) = default;
    InspectorDidRequestDomTreeContextMenu(InspectorDidRequestDomTreeContextMenu&&) = default;
    InspectorDidRequestDomTreeContextMenu& operator=(InspectorDidRequestDomTreeContextMenu const&) = default;
    InspectorDidRequestDomTreeContextMenu(u64 page_id, Web::UniqueNodeID node_id, Gfx::IntPoint position, String type, Optional<String> tag, Optional<size_t> attribute_index) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_position(move(position)), m_type(move(type)), m_tag(move(tag)), m_attribute_index(move(attribute_index)) {}

    virtual ~InspectorDidRequestDomTreeContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidRequestDomTreeContextMenu; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidRequestDomTreeContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidRequestDomTreeContextMenu"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidRequestDomTreeContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto position = TRY((decoder.decode<Gfx::IntPoint>()));

        auto type = TRY((decoder.decode<String>()));

        auto tag = TRY((decoder.decode<Optional<String>>()));

        auto attribute_index = TRY((decoder.decode<Optional<size_t>>()));

        return make<InspectorDidRequestDomTreeContextMenu>(move(page_id), move(node_id), move(position), move(type), move(tag), move(attribute_index));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidRequestDomTreeContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_position));

        TRY(stream.encode(m_type));

        TRY(stream.encode(m_tag));

        TRY(stream.encode(m_attribute_index));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const Gfx::IntPoint& position() const { return m_position; }
    Gfx::IntPoint take_position() { return move(m_position); }

    const String& type() const { return m_type; }
    String take_type() { return move(m_type); }

    const Optional<String>& tag() const { return m_tag; }
    Optional<String> take_tag() { return move(m_tag); }

    const Optional<size_t>& attribute_index() const { return m_attribute_index; }
    Optional<size_t> take_attribute_index() { return move(m_attribute_index); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    Gfx::IntPoint m_position;

    String m_type;

    Optional<String> m_tag;

    Optional<size_t> m_attribute_index;

};

class InspectorDidRequestCookieContextMenu final : public IPC::Message {
public:

    InspectorDidRequestCookieContextMenu(InspectorDidRequestCookieContextMenu const&) = default;
    InspectorDidRequestCookieContextMenu(InspectorDidRequestCookieContextMenu&&) = default;
    InspectorDidRequestCookieContextMenu& operator=(InspectorDidRequestCookieContextMenu const&) = default;
    InspectorDidRequestCookieContextMenu(u64 page_id, size_t cookie_index, Gfx::IntPoint position) : m_page_id(move(page_id)), m_cookie_index(move(cookie_index)), m_position(move(position)) {}

    virtual ~InspectorDidRequestCookieContextMenu() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidRequestCookieContextMenu; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidRequestCookieContextMenu; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidRequestCookieContextMenu"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidRequestCookieContextMenu>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto cookie_index = TRY((decoder.decode<size_t>()));

        auto position = TRY((decoder.decode<Gfx::IntPoint>()));

        return make<InspectorDidRequestCookieContextMenu>(move(page_id), move(cookie_index), move(position));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidRequestCookieContextMenu));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_cookie_index));

        TRY(stream.encode(m_position));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const size_t& cookie_index() const { return m_cookie_index; }
    size_t take_cookie_index() { return move(m_cookie_index); }

    const Gfx::IntPoint& position() const { return m_position; }
    Gfx::IntPoint take_position() { return move(m_position); }

private:

    u64 m_page_id;

    size_t m_cookie_index;

    Gfx::IntPoint m_position;

};

class InspectorDidExecuteConsoleScript final : public IPC::Message {
public:

    InspectorDidExecuteConsoleScript(InspectorDidExecuteConsoleScript const&) = default;
    InspectorDidExecuteConsoleScript(InspectorDidExecuteConsoleScript&&) = default;
    InspectorDidExecuteConsoleScript& operator=(InspectorDidExecuteConsoleScript const&) = default;
    InspectorDidExecuteConsoleScript(u64 page_id, String script) : m_page_id(move(page_id)), m_script(move(script)) {}

    virtual ~InspectorDidExecuteConsoleScript() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidExecuteConsoleScript; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidExecuteConsoleScript; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidExecuteConsoleScript"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidExecuteConsoleScript>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto script = TRY((decoder.decode<String>()));

        return make<InspectorDidExecuteConsoleScript>(move(page_id), move(script));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidExecuteConsoleScript));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_script));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& script() const { return m_script; }
    String take_script() { return move(m_script); }

private:

    u64 m_page_id;

    String m_script;

};

class InspectorDidExportInspectorHtml final : public IPC::Message {
public:

    InspectorDidExportInspectorHtml(InspectorDidExportInspectorHtml const&) = default;
    InspectorDidExportInspectorHtml(InspectorDidExportInspectorHtml&&) = default;
    InspectorDidExportInspectorHtml& operator=(InspectorDidExportInspectorHtml const&) = default;
    InspectorDidExportInspectorHtml(u64 page_id, String html) : m_page_id(move(page_id)), m_html(move(html)) {}

    virtual ~InspectorDidExportInspectorHtml() override {}

    virtual u32 endpoint_magic() const override { return 2565717604; }
    virtual i32 message_id() const override { return (int)MessageID::InspectorDidExportInspectorHtml; }
    static i32 static_message_id() { return (int)MessageID::InspectorDidExportInspectorHtml; }
    virtual const char* message_name() const override { return "WebContentClient::InspectorDidExportInspectorHtml"; }

    static ErrorOr<NonnullOwnPtr<InspectorDidExportInspectorHtml>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto html = TRY((decoder.decode<String>()));

        return make<InspectorDidExportInspectorHtml>(move(page_id), move(html));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectorDidExportInspectorHtml));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_html));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& html() const { return m_html; }
    String take_html() { return move(m_html); }

private:

    u64 m_page_id;

    String m_html;

};

} // namespace Messages::WebContentClient

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentClientProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    WebContentClientProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    void async_did_start_loading(
u64 page_id, URL::URL url, bool is_redirect) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidStartLoading { page_id, move(url), is_redirect }));

    }

    void async_did_finish_loading(
u64 page_id, URL::URL url) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidFinishLoading { page_id, move(url) }));

    }

    void async_did_request_refresh(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestRefresh { page_id }));

    }

    void async_did_paint(
u64 page_id, Gfx::IntRect content_rect, i32 bitmap_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidPaint { page_id, move(content_rect), bitmap_id }));

    }

    void async_did_request_cursor_change(
u64 page_id, i32 cursor_type) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestCursorChange { page_id, cursor_type }));

    }

    void async_did_change_title(
u64 page_id, ByteString title) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidChangeTitle { page_id, move(title) }));

    }

    void async_did_change_url(
u64 page_id, URL::URL url) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidChangeUrl { page_id, move(url) }));

    }

    void async_did_request_tooltip_override(
u64 page_id, Gfx::IntPoint position, ByteString title) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestTooltipOverride { page_id, position, move(title) }));

    }

    void async_did_stop_tooltip_override(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidStopTooltipOverride { page_id }));

    }

    void async_did_enter_tooltip_area(
u64 page_id, ByteString title) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidEnterTooltipArea { page_id, move(title) }));

    }

    void async_did_leave_tooltip_area(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidLeaveTooltipArea { page_id }));

    }

    void async_did_hover_link(
u64 page_id, URL::URL url) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidHoverLink { page_id, move(url) }));

    }

    void async_did_unhover_link(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidUnhoverLink { page_id }));

    }

    void async_did_click_link(
u64 page_id, URL::URL url, ByteString target, unsigned modifiers) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidClickLink { page_id, move(url), move(target), modifiers }));

    }

    void async_did_middle_click_link(
u64 page_id, URL::URL url, ByteString target, unsigned modifiers) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidMiddleClickLink { page_id, move(url), move(target), modifiers }));

    }

    void async_did_request_context_menu(
u64 page_id, Gfx::IntPoint content_position) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestContextMenu { page_id, content_position }));

    }

    void async_did_request_link_context_menu(
u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestLinkContextMenu { page_id, content_position, move(url), move(target), modifiers }));

    }

    void async_did_request_image_context_menu(
u64 page_id, Gfx::IntPoint content_position, URL::URL url, ByteString target, unsigned modifiers, Optional<Gfx::ShareableBitmap> bitmap) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestImageContextMenu { page_id, content_position, move(url), move(target), modifiers, move(bitmap) }));

    }

    void async_did_request_media_context_menu(
u64 page_id, Gfx::IntPoint content_position, ByteString target, unsigned modifiers, Web::Page::MediaContextMenu menu) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestMediaContextMenu { page_id, content_position, move(target), modifiers, move(menu) }));

    }

    void async_did_request_alert(
u64 page_id, String message) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestAlert { page_id, move(message) }));

    }

    void async_did_request_confirm(
u64 page_id, String message) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestConfirm { page_id, move(message) }));

    }

    void async_did_request_prompt(
u64 page_id, String message, String default_) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestPrompt { page_id, move(message), move(default_) }));

    }

    void async_did_request_set_prompt_text(
u64 page_id, String message) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestSetPromptText { page_id, move(message) }));

    }

    void async_did_request_accept_dialog(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestAcceptDialog { page_id }));

    }

    void async_did_request_dismiss_dialog(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestDismissDialog { page_id }));

    }

    void async_did_get_source(
u64 page_id, URL::URL url, URL::URL base_url, String source) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetSource { page_id, move(url), move(base_url), move(source) }));

    }

    void async_did_inspect_dom_tree(
u64 page_id, ByteString dom_tree) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidInspectDomTree { page_id, move(dom_tree) }));

    }

    void async_did_inspect_dom_node(
u64 page_id, bool has_style, ByteString computed_style, ByteString resolved_style, ByteString custom_properties, ByteString node_box_sizing, ByteString aria_properties_state, ByteString fonts) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidInspectDomNode { page_id, has_style, move(computed_style), move(resolved_style), move(custom_properties), move(node_box_sizing), move(aria_properties_state), move(fonts) }));

    }

    void async_did_inspect_accessibility_tree(
u64 page_id, ByteString accessibility_tree) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidInspectAccessibilityTree { page_id, move(accessibility_tree) }));

    }

    void async_did_get_hovered_node_id(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetHoveredNodeId { page_id, move(node_id) }));

    }

    void async_did_finish_editing_dom_node(
u64 page_id, Optional<Web::UniqueNodeID> node_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidFinishEditingDomNode { page_id, move(node_id) }));

    }

    void async_did_get_dom_node_html(
u64 page_id, String html) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetDomNodeHtml { page_id, move(html) }));

    }

    void async_inspector_did_list_style_sheets(
u64 page_id, Vector<Web::CSS::StyleSheetIdentifier> style_sheets) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidListStyleSheets { page_id, move(style_sheets) }));

    }

    void async_inspector_did_request_style_sheet_source(
u64 page_id, Web::CSS::StyleSheetIdentifier identifier) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidRequestStyleSheetSource { page_id, move(identifier) }));

    }

    void async_did_get_style_sheet_source(
u64 page_id, Web::CSS::StyleSheetIdentifier identifier, URL::URL base_url, String source) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetStyleSheetSource { page_id, move(identifier), move(base_url), move(source) }));

    }

    void async_did_take_screenshot(
u64 page_id, Gfx::ShareableBitmap screenshot) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidTakeScreenshot { page_id, move(screenshot) }));

    }

    void async_did_get_internal_page_info(
u64 page_id, WebView::PageInfoType type, String info) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetInternalPageInfo { page_id, type, move(info) }));

    }

    void async_did_change_favicon(
u64 page_id, Gfx::ShareableBitmap favicon) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidChangeFavicon { page_id, move(favicon) }));

    }

    Vector<Web::Cookie::Cookie> did_request_all_cookies(
URL::URL url) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestAllCookies>(move(url))->take_cookies();
    }

    void async_did_request_all_cookies(
URL::URL url) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestAllCookies { move(url) }));

    }

    IPC::IPCErrorOr<Vector<Web::Cookie::Cookie>> try_did_request_all_cookies(
URL::URL url) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestAllCookies>(move(url));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    Optional<Web::Cookie::Cookie> did_request_named_cookie(
URL::URL url, String name) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestNamedCookie>(move(url), move(name))->take_cookie();
    }

    void async_did_request_named_cookie(
URL::URL url, String name) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestNamedCookie { move(url), move(name) }));

    }

    IPC::IPCErrorOr<Optional<Web::Cookie::Cookie>> try_did_request_named_cookie(
URL::URL url, String name) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestNamedCookie>(move(url), move(name));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    String did_request_cookie(
URL::URL url, Web::Cookie::Source source) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestCookie>(move(url), source)->take_cookie();
    }

    void async_did_request_cookie(
URL::URL url, Web::Cookie::Source source) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestCookie { move(url), source }));

    }

    IPC::IPCErrorOr<String> try_did_request_cookie(
URL::URL url, Web::Cookie::Source source) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestCookie>(move(url), source);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void did_set_cookie(
URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) {
        (void) m_connection.template send_sync<Messages::WebContentClient::DidSetCookie>(move(url), move(cookie), source);
    }

    void async_did_set_cookie(
URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidSetCookie { move(url), move(cookie), source }));

    }

    IPC::IPCErrorOr<void> try_did_set_cookie(
URL::URL url, Web::Cookie::ParsedCookie cookie, Web::Cookie::Source source) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidSetCookie>(move(url), move(cookie), source);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return { };

    }

    void async_did_update_cookie(
Web::Cookie::Cookie cookie) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidUpdateCookie { move(cookie) }));

    }

    void async_did_expire_cookies_with_time_offset(
AK::Duration offset) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidExpireCookiesWithTimeOffset { offset }));

    }

    void async_did_update_resource_count(
u64 page_id, i32 count_waiting) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidUpdateResourceCount { page_id, count_waiting }));

    }

    String did_request_new_web_view(
u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) {
        return m_connection.template send_sync<Messages::WebContentClient::DidRequestNewWebView>(page_id, move(activate_tab), move(hints), move(page_index))->take_handle();
    }

    void async_did_request_new_web_view(
u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestNewWebView { page_id, move(activate_tab), move(hints), move(page_index) }));

    }

    IPC::IPCErrorOr<String> try_did_request_new_web_view(
u64 page_id, Web::HTML::ActivateTab activate_tab, Web::HTML::WebViewHints hints, Optional<u64> page_index) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::DidRequestNewWebView>(page_id, move(activate_tab), move(hints), move(page_index));
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_did_request_activate_tab(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestActivateTab { page_id }));

    }

    void async_did_close_browsing_context(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidCloseBrowsingContext { page_id }));

    }

    void async_did_request_restore_window(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestRestoreWindow { page_id }));

    }

    void async_did_request_reposition_window(
u64 page_id, Gfx::IntPoint position) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestRepositionWindow { page_id, position }));

    }

    void async_did_request_resize_window(
u64 page_id, Gfx::IntSize size) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestResizeWindow { page_id, size }));

    }

    void async_did_request_maximize_window(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestMaximizeWindow { page_id }));

    }

    void async_did_request_minimize_window(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestMinimizeWindow { page_id }));

    }

    void async_did_request_fullscreen_window(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestFullscreenWindow { page_id }));

    }

    void async_did_request_file(
u64 page_id, ByteString path, i32 request_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestFile { page_id, move(path), request_id }));

    }

    void async_did_request_color_picker(
u64 page_id, Color current_color) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestColorPicker { page_id, move(current_color) }));

    }

    void async_did_request_file_picker(
u64 page_id, Web::HTML::FileFilter accepted_file_types, Web::HTML::AllowMultipleFiles allow_multiple_files) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestFilePicker { page_id, move(accepted_file_types), allow_multiple_files }));

    }

    void async_did_request_select_dropdown(
u64 page_id, Gfx::IntPoint content_position, i32 minimum_width, Vector<Web::HTML::SelectItem> items) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidRequestSelectDropdown { page_id, content_position, minimum_width, move(items) }));

    }

    void async_did_finish_handling_input_event(
u64 page_id, Web::EventResult event_result) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidFinishHandlingInputEvent { page_id, event_result }));

    }

    void async_did_change_theme_color(
u64 page_id, Gfx::Color color) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidChangeThemeColor { page_id, color }));

    }

    void async_did_insert_clipboard_entry(
u64 page_id, String data, String presentation_style, String mime_type) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidInsertClipboardEntry { page_id, move(data), move(presentation_style), move(mime_type) }));

    }

    void async_did_update_navigation_buttons_state(
u64 page_id, bool back_enabled, bool forward_enabled) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidUpdateNavigationButtonsState { page_id, back_enabled, forward_enabled }));

    }

    void async_did_allocate_backing_stores(
u64 page_id, i32 front_bitmap_id, Gfx::ShareableBitmap front_bitmap, i32 back_bitmap_id, Gfx::ShareableBitmap back_bitmap) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidAllocateBackingStores { page_id, front_bitmap_id, move(front_bitmap), back_bitmap_id, move(back_bitmap) }));

    }

    void async_did_change_audio_play_state(
u64 page_id, Web::HTML::AudioPlayState play_state) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidChangeAudioPlayState { page_id, play_state }));

    }

    void async_did_output_js_console_message(
u64 page_id, i32 message_index) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidOutputJsConsoleMessage { page_id, message_index }));

    }

    void async_did_get_js_console_messages(
u64 page_id, i32 start_index, Vector<ByteString> message_types, Vector<ByteString> messages) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidGetJsConsoleMessages { page_id, start_index, move(message_types), move(messages) }));

    }

    void async_did_finish_text_test(
u64 page_id, String text) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidFinishTextTest { page_id, move(text) }));

    }

    void async_did_set_test_timeout(
u64 page_id, double milliseconds) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidSetTestTimeout { page_id, milliseconds }));

    }

    void async_did_set_browser_zoom(
u64 page_id, double factor) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidSetBrowserZoom { page_id, factor }));

    }

    void async_did_find_in_page(
u64 page_id, size_t current_match_index, Optional<size_t> total_match_count) {
        MUST(m_connection.post_message(Messages::WebContentClient::DidFindInPage { page_id, current_match_index, move(total_match_count) }));

    }

    IPC::File request_worker_agent(
u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentClient::RequestWorkerAgent>(page_id)->take_socket();
    }

    void async_request_worker_agent(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::RequestWorkerAgent { page_id }));

    }

    IPC::IPCErrorOr<IPC::File> try_request_worker_agent(
u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentClient::RequestWorkerAgent>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_inspector_did_load(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidLoad { page_id }));

    }

    void async_inspector_did_select_dom_node(
u64 page_id, Web::UniqueNodeID node_id, Optional<Web::CSS::Selector::PseudoElement::Type> pseudo_element) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidSelectDomNode { page_id, move(node_id), move(pseudo_element) }));

    }

    void async_inspector_did_set_dom_node_text(
u64 page_id, Web::UniqueNodeID node_id, String text) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidSetDomNodeText { page_id, move(node_id), move(text) }));

    }

    void async_inspector_did_set_dom_node_tag(
u64 page_id, Web::UniqueNodeID node_id, String tag) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidSetDomNodeTag { page_id, move(node_id), move(tag) }));

    }

    void async_inspector_did_add_dom_node_attributes(
u64 page_id, Web::UniqueNodeID node_id, Vector<WebView::Attribute> attributes) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidAddDomNodeAttributes { page_id, move(node_id), move(attributes) }));

    }

    void async_inspector_did_replace_dom_node_attribute(
u64 page_id, Web::UniqueNodeID node_id, size_t attribute_index, Vector<WebView::Attribute> replacement_attributes) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidReplaceDomNodeAttribute { page_id, move(node_id), attribute_index, move(replacement_attributes) }));

    }

    void async_inspector_did_request_dom_tree_context_menu(
u64 page_id, Web::UniqueNodeID node_id, Gfx::IntPoint position, String type, Optional<String> tag, Optional<size_t> attribute_index) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidRequestDomTreeContextMenu { page_id, move(node_id), position, move(type), move(tag), move(attribute_index) }));

    }

    void async_inspector_did_request_cookie_context_menu(
u64 page_id, size_t cookie_index, Gfx::IntPoint position) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidRequestCookieContextMenu { page_id, cookie_index, position }));

    }

    void async_inspector_did_execute_console_script(
u64 page_id, String script) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidExecuteConsoleScript { page_id, move(script) }));

    }

    void async_inspector_did_export_inspector_html(
u64 page_id, String html) {
        MUST(m_connection.post_message(Messages::WebContentClient::InspectorDidExportInspectorHtml { page_id, move(html) }));

    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentClientProxy;
class WebContentClientStub;

class WebContentClientEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = WebContentClientProxy<LocalEndpoint, WebContentClientEndpoint>;
    using Stub = WebContentClientStub;

    static u32 static_magic() { return 2565717604; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 2565717604) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::WebContentClient::MessageID::DidStartLoading:
            return TRY(Messages::WebContentClient::DidStartLoading::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishLoading:
            return TRY(Messages::WebContentClient::DidFinishLoading::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRefresh:
            return TRY(Messages::WebContentClient::DidRequestRefresh::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidPaint:
            return TRY(Messages::WebContentClient::DidPaint::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCursorChange:
            return TRY(Messages::WebContentClient::DidRequestCursorChange::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeTitle:
            return TRY(Messages::WebContentClient::DidChangeTitle::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeUrl:
            return TRY(Messages::WebContentClient::DidChangeUrl::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestTooltipOverride:
            return TRY(Messages::WebContentClient::DidRequestTooltipOverride::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidStopTooltipOverride:
            return TRY(Messages::WebContentClient::DidStopTooltipOverride::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidEnterTooltipArea:
            return TRY(Messages::WebContentClient::DidEnterTooltipArea::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidLeaveTooltipArea:
            return TRY(Messages::WebContentClient::DidLeaveTooltipArea::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidHoverLink:
            return TRY(Messages::WebContentClient::DidHoverLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUnhoverLink:
            return TRY(Messages::WebContentClient::DidUnhoverLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidClickLink:
            return TRY(Messages::WebContentClient::DidClickLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidMiddleClickLink:
            return TRY(Messages::WebContentClient::DidMiddleClickLink::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestContextMenu:
            return TRY(Messages::WebContentClient::DidRequestContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestLinkContextMenu:
            return TRY(Messages::WebContentClient::DidRequestLinkContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestImageContextMenu:
            return TRY(Messages::WebContentClient::DidRequestImageContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMediaContextMenu:
            return TRY(Messages::WebContentClient::DidRequestMediaContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAlert:
            return TRY(Messages::WebContentClient::DidRequestAlert::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestConfirm:
            return TRY(Messages::WebContentClient::DidRequestConfirm::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestPrompt:
            return TRY(Messages::WebContentClient::DidRequestPrompt::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestSetPromptText:
            return TRY(Messages::WebContentClient::DidRequestSetPromptText::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAcceptDialog:
            return TRY(Messages::WebContentClient::DidRequestAcceptDialog::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestDismissDialog:
            return TRY(Messages::WebContentClient::DidRequestDismissDialog::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetSource:
            return TRY(Messages::WebContentClient::DidGetSource::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectDomTree:
            return TRY(Messages::WebContentClient::DidInspectDomTree::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectDomNode:
            return TRY(Messages::WebContentClient::DidInspectDomNode::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInspectAccessibilityTree:
            return TRY(Messages::WebContentClient::DidInspectAccessibilityTree::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetHoveredNodeId:
            return TRY(Messages::WebContentClient::DidGetHoveredNodeId::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishEditingDomNode:
            return TRY(Messages::WebContentClient::DidFinishEditingDomNode::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetDomNodeHtml:
            return TRY(Messages::WebContentClient::DidGetDomNodeHtml::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidListStyleSheets:
            return TRY(Messages::WebContentClient::InspectorDidListStyleSheets::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestStyleSheetSource:
            return TRY(Messages::WebContentClient::InspectorDidRequestStyleSheetSource::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetStyleSheetSource:
            return TRY(Messages::WebContentClient::DidGetStyleSheetSource::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidTakeScreenshot:
            return TRY(Messages::WebContentClient::DidTakeScreenshot::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetInternalPageInfo:
            return TRY(Messages::WebContentClient::DidGetInternalPageInfo::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeFavicon:
            return TRY(Messages::WebContentClient::DidChangeFavicon::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookies:
            return TRY(Messages::WebContentClient::DidRequestAllCookies::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookiesResponse:
            return TRY(Messages::WebContentClient::DidRequestAllCookiesResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookie:
            return TRY(Messages::WebContentClient::DidRequestNamedCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookieResponse:
            return TRY(Messages::WebContentClient::DidRequestNamedCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCookie:
            return TRY(Messages::WebContentClient::DidRequestCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestCookieResponse:
            return TRY(Messages::WebContentClient::DidRequestCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetCookie:
            return TRY(Messages::WebContentClient::DidSetCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetCookieResponse:
            return TRY(Messages::WebContentClient::DidSetCookieResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateCookie:
            return TRY(Messages::WebContentClient::DidUpdateCookie::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidExpireCookiesWithTimeOffset:
            return TRY(Messages::WebContentClient::DidExpireCookiesWithTimeOffset::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateResourceCount:
            return TRY(Messages::WebContentClient::DidUpdateResourceCount::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebView:
            return TRY(Messages::WebContentClient::DidRequestNewWebView::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebViewResponse:
            return TRY(Messages::WebContentClient::DidRequestNewWebViewResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestActivateTab:
            return TRY(Messages::WebContentClient::DidRequestActivateTab::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidCloseBrowsingContext:
            return TRY(Messages::WebContentClient::DidCloseBrowsingContext::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRestoreWindow:
            return TRY(Messages::WebContentClient::DidRequestRestoreWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestRepositionWindow:
            return TRY(Messages::WebContentClient::DidRequestRepositionWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestResizeWindow:
            return TRY(Messages::WebContentClient::DidRequestResizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMaximizeWindow:
            return TRY(Messages::WebContentClient::DidRequestMaximizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestMinimizeWindow:
            return TRY(Messages::WebContentClient::DidRequestMinimizeWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFullscreenWindow:
            return TRY(Messages::WebContentClient::DidRequestFullscreenWindow::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFile:
            return TRY(Messages::WebContentClient::DidRequestFile::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestColorPicker:
            return TRY(Messages::WebContentClient::DidRequestColorPicker::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestFilePicker:
            return TRY(Messages::WebContentClient::DidRequestFilePicker::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidRequestSelectDropdown:
            return TRY(Messages::WebContentClient::DidRequestSelectDropdown::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishHandlingInputEvent:
            return TRY(Messages::WebContentClient::DidFinishHandlingInputEvent::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeThemeColor:
            return TRY(Messages::WebContentClient::DidChangeThemeColor::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidInsertClipboardEntry:
            return TRY(Messages::WebContentClient::DidInsertClipboardEntry::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidUpdateNavigationButtonsState:
            return TRY(Messages::WebContentClient::DidUpdateNavigationButtonsState::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidAllocateBackingStores:
            return TRY(Messages::WebContentClient::DidAllocateBackingStores::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidChangeAudioPlayState:
            return TRY(Messages::WebContentClient::DidChangeAudioPlayState::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidOutputJsConsoleMessage:
            return TRY(Messages::WebContentClient::DidOutputJsConsoleMessage::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidGetJsConsoleMessages:
            return TRY(Messages::WebContentClient::DidGetJsConsoleMessages::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFinishTextTest:
            return TRY(Messages::WebContentClient::DidFinishTextTest::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetTestTimeout:
            return TRY(Messages::WebContentClient::DidSetTestTimeout::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidSetBrowserZoom:
            return TRY(Messages::WebContentClient::DidSetBrowserZoom::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::DidFindInPage:
            return TRY(Messages::WebContentClient::DidFindInPage::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgent:
            return TRY(Messages::WebContentClient::RequestWorkerAgent::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgentResponse:
            return TRY(Messages::WebContentClient::RequestWorkerAgentResponse::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidLoad:
            return TRY(Messages::WebContentClient::InspectorDidLoad::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidSelectDomNode:
            return TRY(Messages::WebContentClient::InspectorDidSelectDomNode::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidSetDomNodeText:
            return TRY(Messages::WebContentClient::InspectorDidSetDomNodeText::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidSetDomNodeTag:
            return TRY(Messages::WebContentClient::InspectorDidSetDomNodeTag::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidAddDomNodeAttributes:
            return TRY(Messages::WebContentClient::InspectorDidAddDomNodeAttributes::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidReplaceDomNodeAttribute:
            return TRY(Messages::WebContentClient::InspectorDidReplaceDomNodeAttribute::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestDomTreeContextMenu:
            return TRY(Messages::WebContentClient::InspectorDidRequestDomTreeContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestCookieContextMenu:
            return TRY(Messages::WebContentClient::InspectorDidRequestCookieContextMenu::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidExecuteConsoleScript:
            return TRY(Messages::WebContentClient::InspectorDidExecuteConsoleScript::decode(stream, files));
        case (int)Messages::WebContentClient::MessageID::InspectorDidExportInspectorHtml:
            return TRY(Messages::WebContentClient::InspectorDidExportInspectorHtml::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode WebContentClient message");
        }

        VERIFY_NOT_REACHED();
    }

};

class WebContentClientStub : public IPC::Stub {
public:
    WebContentClientStub() { }
    virtual ~WebContentClientStub() override { }

    virtual u32 magic() const override { return 2565717604; }
    virtual ByteString name() const override { return "WebContentClient"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(const IPC::Message& message) override
    {
        switch (message.message_id()) {

        case (int)Messages::WebContentClient::MessageID::DidStartLoading: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidStartLoading&>(message);
            did_start_loading(request.page_id(), request.url(), request.is_redirect());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidFinishLoading: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidFinishLoading&>(message);
            did_finish_loading(request.page_id(), request.url());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestRefresh: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestRefresh&>(message);
            did_request_refresh(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidPaint: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidPaint&>(message);
            did_paint(request.page_id(), request.content_rect(), request.bitmap_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestCursorChange: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestCursorChange&>(message);
            did_request_cursor_change(request.page_id(), request.cursor_type());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidChangeTitle: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidChangeTitle&>(message);
            did_change_title(request.page_id(), request.title());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidChangeUrl: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidChangeUrl&>(message);
            did_change_url(request.page_id(), request.url());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestTooltipOverride: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestTooltipOverride&>(message);
            did_request_tooltip_override(request.page_id(), request.position(), request.title());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidStopTooltipOverride: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidStopTooltipOverride&>(message);
            did_stop_tooltip_override(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidEnterTooltipArea: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidEnterTooltipArea&>(message);
            did_enter_tooltip_area(request.page_id(), request.title());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidLeaveTooltipArea: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidLeaveTooltipArea&>(message);
            did_leave_tooltip_area(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidHoverLink: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidHoverLink&>(message);
            did_hover_link(request.page_id(), request.url());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidUnhoverLink: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidUnhoverLink&>(message);
            did_unhover_link(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidClickLink: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidClickLink&>(message);
            did_click_link(request.page_id(), request.url(), request.target(), request.modifiers());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidMiddleClickLink: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidMiddleClickLink&>(message);
            did_middle_click_link(request.page_id(), request.url(), request.target(), request.modifiers());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestContextMenu&>(message);
            did_request_context_menu(request.page_id(), request.content_position());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestLinkContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestLinkContextMenu&>(message);
            did_request_link_context_menu(request.page_id(), request.content_position(), request.url(), request.target(), request.modifiers());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestImageContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestImageContextMenu&>(message);
            did_request_image_context_menu(request.page_id(), request.content_position(), request.url(), request.target(), request.modifiers(), request.bitmap());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestMediaContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestMediaContextMenu&>(message);
            did_request_media_context_menu(request.page_id(), request.content_position(), request.target(), request.modifiers(), request.menu());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestAlert: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestAlert&>(message);
            did_request_alert(request.page_id(), request.message());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestConfirm: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestConfirm&>(message);
            did_request_confirm(request.page_id(), request.message());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestPrompt: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestPrompt&>(message);
            did_request_prompt(request.page_id(), request.message(), request.default_());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestSetPromptText: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestSetPromptText&>(message);
            did_request_set_prompt_text(request.page_id(), request.message());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestAcceptDialog: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestAcceptDialog&>(message);
            did_request_accept_dialog(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestDismissDialog: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestDismissDialog&>(message);
            did_request_dismiss_dialog(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetSource: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetSource&>(message);
            did_get_source(request.page_id(), request.url(), request.base_url(), request.source());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidInspectDomTree: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidInspectDomTree&>(message);
            did_inspect_dom_tree(request.page_id(), request.dom_tree());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidInspectDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidInspectDomNode&>(message);
            did_inspect_dom_node(request.page_id(), request.has_style(), request.computed_style(), request.resolved_style(), request.custom_properties(), request.node_box_sizing(), request.aria_properties_state(), request.fonts());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidInspectAccessibilityTree: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidInspectAccessibilityTree&>(message);
            did_inspect_accessibility_tree(request.page_id(), request.accessibility_tree());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetHoveredNodeId: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetHoveredNodeId&>(message);
            did_get_hovered_node_id(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidFinishEditingDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidFinishEditingDomNode&>(message);
            did_finish_editing_dom_node(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetDomNodeHtml: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetDomNodeHtml&>(message);
            did_get_dom_node_html(request.page_id(), request.html());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidListStyleSheets: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidListStyleSheets&>(message);
            inspector_did_list_style_sheets(request.page_id(), request.style_sheets());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestStyleSheetSource: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidRequestStyleSheetSource&>(message);
            inspector_did_request_style_sheet_source(request.page_id(), request.identifier());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetStyleSheetSource: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetStyleSheetSource&>(message);
            did_get_style_sheet_source(request.page_id(), request.identifier(), request.base_url(), request.source());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidTakeScreenshot: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidTakeScreenshot&>(message);
            did_take_screenshot(request.page_id(), request.screenshot());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetInternalPageInfo: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetInternalPageInfo&>(message);
            did_get_internal_page_info(request.page_id(), request.type(), request.info());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidChangeFavicon: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidChangeFavicon&>(message);
            did_change_favicon(request.page_id(), request.favicon());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestAllCookies: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestAllCookies&>(message);
            auto response = did_request_all_cookies(request.url());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestNamedCookie: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestNamedCookie&>(message);
            auto response = did_request_named_cookie(request.url(), request.name());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestCookie: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestCookie&>(message);
            auto response = did_request_cookie(request.url(), request.source());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::DidSetCookie: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidSetCookie&>(message);
            did_set_cookie(request.url(), request.cookie(), request.source());
            auto response = Messages::WebContentClient::DidSetCookieResponse { };
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::DidUpdateCookie: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidUpdateCookie&>(message);
            did_update_cookie(request.cookie());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidExpireCookiesWithTimeOffset: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidExpireCookiesWithTimeOffset&>(message);
            did_expire_cookies_with_time_offset(request.offset());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidUpdateResourceCount: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidUpdateResourceCount&>(message);
            did_update_resource_count(request.page_id(), request.count_waiting());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestNewWebView: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestNewWebView&>(message);
            auto response = did_request_new_web_view(request.page_id(), request.activate_tab(), request.hints(), request.page_index());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestActivateTab: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestActivateTab&>(message);
            did_request_activate_tab(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidCloseBrowsingContext: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidCloseBrowsingContext&>(message);
            did_close_browsing_context(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestRestoreWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestRestoreWindow&>(message);
            did_request_restore_window(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestRepositionWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestRepositionWindow&>(message);
            did_request_reposition_window(request.page_id(), request.position());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestResizeWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestResizeWindow&>(message);
            did_request_resize_window(request.page_id(), request.size());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestMaximizeWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestMaximizeWindow&>(message);
            did_request_maximize_window(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestMinimizeWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestMinimizeWindow&>(message);
            did_request_minimize_window(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestFullscreenWindow: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestFullscreenWindow&>(message);
            did_request_fullscreen_window(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestFile: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestFile&>(message);
            did_request_file(request.page_id(), request.path(), request.request_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestColorPicker: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestColorPicker&>(message);
            did_request_color_picker(request.page_id(), request.current_color());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestFilePicker: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestFilePicker&>(message);
            did_request_file_picker(request.page_id(), request.accepted_file_types(), request.allow_multiple_files());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidRequestSelectDropdown: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidRequestSelectDropdown&>(message);
            did_request_select_dropdown(request.page_id(), request.content_position(), request.minimum_width(), request.items());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidFinishHandlingInputEvent: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidFinishHandlingInputEvent&>(message);
            did_finish_handling_input_event(request.page_id(), request.event_result());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidChangeThemeColor: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidChangeThemeColor&>(message);
            did_change_theme_color(request.page_id(), request.color());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidInsertClipboardEntry: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidInsertClipboardEntry&>(message);
            did_insert_clipboard_entry(request.page_id(), request.data(), request.presentation_style(), request.mime_type());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidUpdateNavigationButtonsState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidUpdateNavigationButtonsState&>(message);
            did_update_navigation_buttons_state(request.page_id(), request.back_enabled(), request.forward_enabled());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidAllocateBackingStores: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidAllocateBackingStores&>(message);
            did_allocate_backing_stores(request.page_id(), request.front_bitmap_id(), request.front_bitmap(), request.back_bitmap_id(), request.back_bitmap());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidChangeAudioPlayState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidChangeAudioPlayState&>(message);
            did_change_audio_play_state(request.page_id(), request.play_state());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidOutputJsConsoleMessage: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidOutputJsConsoleMessage&>(message);
            did_output_js_console_message(request.page_id(), request.message_index());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidGetJsConsoleMessages: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidGetJsConsoleMessages&>(message);
            did_get_js_console_messages(request.page_id(), request.start_index(), request.message_types(), request.messages());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidFinishTextTest: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidFinishTextTest&>(message);
            did_finish_text_test(request.page_id(), request.text());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidSetTestTimeout: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidSetTestTimeout&>(message);
            did_set_test_timeout(request.page_id(), request.milliseconds());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidSetBrowserZoom: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidSetBrowserZoom&>(message);
            did_set_browser_zoom(request.page_id(), request.factor());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::DidFindInPage: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::DidFindInPage&>(message);
            did_find_in_page(request.page_id(), request.current_match_index(), request.total_match_count());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::RequestWorkerAgent: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::RequestWorkerAgent&>(message);
            auto response = request_worker_agent(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidLoad: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidLoad&>(message);
            inspector_did_load(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidSelectDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidSelectDomNode&>(message);
            inspector_did_select_dom_node(request.page_id(), request.node_id(), request.pseudo_element());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidSetDomNodeText: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidSetDomNodeText&>(message);
            inspector_did_set_dom_node_text(request.page_id(), request.node_id(), request.text());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidSetDomNodeTag: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidSetDomNodeTag&>(message);
            inspector_did_set_dom_node_tag(request.page_id(), request.node_id(), request.tag());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidAddDomNodeAttributes: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidAddDomNodeAttributes&>(message);
            inspector_did_add_dom_node_attributes(request.page_id(), request.node_id(), request.attributes());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidReplaceDomNodeAttribute: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidReplaceDomNodeAttribute&>(message);
            inspector_did_replace_dom_node_attribute(request.page_id(), request.node_id(), request.attribute_index(), request.replacement_attributes());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestDomTreeContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidRequestDomTreeContextMenu&>(message);
            inspector_did_request_dom_tree_context_menu(request.page_id(), request.node_id(), request.position(), request.type(), request.tag(), request.attribute_index());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidRequestCookieContextMenu: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidRequestCookieContextMenu&>(message);
            inspector_did_request_cookie_context_menu(request.page_id(), request.cookie_index(), request.position());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidExecuteConsoleScript: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidExecuteConsoleScript&>(message);
            inspector_did_execute_console_script(request.page_id(), request.script());
            return nullptr;

        }

        case (int)Messages::WebContentClient::MessageID::InspectorDidExportInspectorHtml: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentClient::InspectorDidExportInspectorHtml&>(message);
            inspector_did_export_inspector_html(request.page_id(), request.html());
            return nullptr;

        }

        default:
            return Error::from_string_literal("Unknown message ID for WebContentClient endpoint");
        }
    }

    virtual void did_start_loading(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] bool is_redirect) = 0;

    virtual void did_finish_loading(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url) = 0;

    virtual void did_request_refresh(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_paint(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntRect const& content_rect, [[maybe_unused]] i32 bitmap_id) = 0;

    virtual void did_request_cursor_change(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 cursor_type) = 0;

    virtual void did_change_title(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& title) = 0;

    virtual void did_change_url(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url) = 0;

    virtual void did_request_tooltip_override(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint position, [[maybe_unused]] ByteString const& title) = 0;

    virtual void did_stop_tooltip_override(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_enter_tooltip_area(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& title) = 0;

    virtual void did_leave_tooltip_area(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_hover_link(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url) = 0;

    virtual void did_unhover_link(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_click_link(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] ByteString const& target, [[maybe_unused]] unsigned modifiers) = 0;

    virtual void did_middle_click_link(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] ByteString const& target, [[maybe_unused]] unsigned modifiers) = 0;

    virtual void did_request_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint content_position) = 0;

    virtual void did_request_link_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint content_position, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] ByteString const& target, [[maybe_unused]] unsigned modifiers) = 0;

    virtual void did_request_image_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint content_position, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] ByteString const& target, [[maybe_unused]] unsigned modifiers, [[maybe_unused]] Optional<Gfx::ShareableBitmap> const& bitmap) = 0;

    virtual void did_request_media_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint content_position, [[maybe_unused]] ByteString const& target, [[maybe_unused]] unsigned modifiers, [[maybe_unused]] Web::Page::MediaContextMenu const& menu) = 0;

    virtual void did_request_alert(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& message) = 0;

    virtual void did_request_confirm(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& message) = 0;

    virtual void did_request_prompt(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& message, [[maybe_unused]] String const& default_) = 0;

    virtual void did_request_set_prompt_text(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& message) = 0;

    virtual void did_request_accept_dialog(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_dismiss_dialog(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_get_source(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url, [[maybe_unused]] URL::URL const& base_url, [[maybe_unused]] String const& source) = 0;

    virtual void did_inspect_dom_tree(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& dom_tree) = 0;

    virtual void did_inspect_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool has_style, [[maybe_unused]] ByteString const& computed_style, [[maybe_unused]] ByteString const& resolved_style, [[maybe_unused]] ByteString const& custom_properties, [[maybe_unused]] ByteString const& node_box_sizing, [[maybe_unused]] ByteString const& aria_properties_state, [[maybe_unused]] ByteString const& fonts) = 0;

    virtual void did_inspect_accessibility_tree(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& accessibility_tree) = 0;

    virtual void did_get_hovered_node_id(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void did_finish_editing_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Optional<Web::UniqueNodeID> const& node_id) = 0;

    virtual void did_get_dom_node_html(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& html) = 0;

    virtual void inspector_did_list_style_sheets(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<Web::CSS::StyleSheetIdentifier> const& style_sheets) = 0;

    virtual void inspector_did_request_style_sheet_source(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::StyleSheetIdentifier const& identifier) = 0;

    virtual void did_get_style_sheet_source(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::StyleSheetIdentifier const& identifier, [[maybe_unused]] URL::URL const& base_url, [[maybe_unused]] String const& source) = 0;

    virtual void did_take_screenshot(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::ShareableBitmap const& screenshot) = 0;

    virtual void did_get_internal_page_info(
[[maybe_unused]] u64 page_id, [[maybe_unused]] WebView::PageInfoType type, [[maybe_unused]] String const& info) = 0;

    virtual void did_change_favicon(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::ShareableBitmap const& favicon) = 0;

    virtual Messages::WebContentClient::DidRequestAllCookiesResponse did_request_all_cookies(
[[maybe_unused]] URL::URL const& url) = 0;

    virtual Messages::WebContentClient::DidRequestNamedCookieResponse did_request_named_cookie(
[[maybe_unused]] URL::URL const& url, [[maybe_unused]] String const& name) = 0;

    virtual Messages::WebContentClient::DidRequestCookieResponse did_request_cookie(
[[maybe_unused]] URL::URL const& url, [[maybe_unused]] Web::Cookie::Source source) = 0;

    virtual void did_set_cookie(
[[maybe_unused]] URL::URL const& url, [[maybe_unused]] Web::Cookie::ParsedCookie const& cookie, [[maybe_unused]] Web::Cookie::Source source) = 0;

    virtual void did_update_cookie(
[[maybe_unused]] Web::Cookie::Cookie const& cookie) = 0;

    virtual void did_expire_cookies_with_time_offset(
[[maybe_unused]] AK::Duration offset) = 0;

    virtual void did_update_resource_count(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 count_waiting) = 0;

    virtual Messages::WebContentClient::DidRequestNewWebViewResponse did_request_new_web_view(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::HTML::ActivateTab const& activate_tab, [[maybe_unused]] Web::HTML::WebViewHints const& hints, [[maybe_unused]] Optional<u64> const& page_index) = 0;

    virtual void did_request_activate_tab(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_close_browsing_context(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_restore_window(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_reposition_window(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint position) = 0;

    virtual void did_request_resize_window(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntSize size) = 0;

    virtual void did_request_maximize_window(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_minimize_window(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_fullscreen_window(
[[maybe_unused]] u64 page_id) = 0;

    virtual void did_request_file(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& path, [[maybe_unused]] i32 request_id) = 0;

    virtual void did_request_color_picker(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Color const& current_color) = 0;

    virtual void did_request_file_picker(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::HTML::FileFilter const& accepted_file_types, [[maybe_unused]] Web::HTML::AllowMultipleFiles allow_multiple_files) = 0;

    virtual void did_request_select_dropdown(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::IntPoint content_position, [[maybe_unused]] i32 minimum_width, [[maybe_unused]] Vector<Web::HTML::SelectItem> const& items) = 0;

    virtual void did_finish_handling_input_event(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::EventResult event_result) = 0;

    virtual void did_change_theme_color(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Gfx::Color color) = 0;

    virtual void did_insert_clipboard_entry(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& data, [[maybe_unused]] String const& presentation_style, [[maybe_unused]] String const& mime_type) = 0;

    virtual void did_update_navigation_buttons_state(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool back_enabled, [[maybe_unused]] bool forward_enabled) = 0;

    virtual void did_allocate_backing_stores(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 front_bitmap_id, [[maybe_unused]] Gfx::ShareableBitmap const& front_bitmap, [[maybe_unused]] i32 back_bitmap_id, [[maybe_unused]] Gfx::ShareableBitmap const& back_bitmap) = 0;

    virtual void did_change_audio_play_state(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::HTML::AudioPlayState play_state) = 0;

    virtual void did_output_js_console_message(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 message_index) = 0;

    virtual void did_get_js_console_messages(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 start_index, [[maybe_unused]] Vector<ByteString> const& message_types, [[maybe_unused]] Vector<ByteString> const& messages) = 0;

    virtual void did_finish_text_test(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& text) = 0;

    virtual void did_set_test_timeout(
[[maybe_unused]] u64 page_id, [[maybe_unused]] double milliseconds) = 0;

    virtual void did_set_browser_zoom(
[[maybe_unused]] u64 page_id, [[maybe_unused]] double factor) = 0;

    virtual void did_find_in_page(
[[maybe_unused]] u64 page_id, [[maybe_unused]] size_t current_match_index, [[maybe_unused]] Optional<size_t> const& total_match_count) = 0;

    virtual Messages::WebContentClient::RequestWorkerAgentResponse request_worker_agent(
[[maybe_unused]] u64 page_id) = 0;

    virtual void inspector_did_load(
[[maybe_unused]] u64 page_id) = 0;

    virtual void inspector_did_select_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] Optional<Web::CSS::Selector::PseudoElement::Type> const& pseudo_element) = 0;

    virtual void inspector_did_set_dom_node_text(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] String const& text) = 0;

    virtual void inspector_did_set_dom_node_tag(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] String const& tag) = 0;

    virtual void inspector_did_add_dom_node_attributes(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] Vector<WebView::Attribute> const& attributes) = 0;

    virtual void inspector_did_replace_dom_node_attribute(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] size_t attribute_index, [[maybe_unused]] Vector<WebView::Attribute> const& replacement_attributes) = 0;

    virtual void inspector_did_request_dom_tree_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] Gfx::IntPoint position, [[maybe_unused]] String const& type, [[maybe_unused]] Optional<String> const& tag, [[maybe_unused]] Optional<size_t> const& attribute_index) = 0;

    virtual void inspector_did_request_cookie_context_menu(
[[maybe_unused]] u64 page_id, [[maybe_unused]] size_t cookie_index, [[maybe_unused]] Gfx::IntPoint position) = 0;

    virtual void inspector_did_execute_console_script(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& script) = 0;

    virtual void inspector_did_export_inspector_html(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& html) = 0;

private:
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
