#pragma once
#include <LibGfx/Rect.h>
#include <LibIPC/File.h>
#include <LibURL/URL.h>
#include <LibWeb/CSS/PreferredColorScheme.h>
#include <LibWeb/CSS/PreferredContrast.h>
#include <LibWeb/CSS/PreferredMotion.h>
#include <LibWeb/CSS/Selector.h>
#include <LibWeb/CSS/StyleSheetIdentifier.h>
#include <LibWeb/HTML/ColorPickerUpdateState.h>
#include <LibWeb/HTML/SelectedFile.h>
#include <LibWeb/HTML/VisibilityState.h>
#include <LibWeb/Page/InputEvent.h>
#include <LibWeb/WebDriver/ExecuteScript.h>
#include <LibWebView/Attribute.h>
#include <LibWebView/PageInfo.h>
#include <AK/Error.h>
#include <AK/MemoryStream.h>
#include <AK/OwnPtr.h>
#include <AK/Result.h>
#include <AK/Utf8View.h>
#include <LibIPC/Connection.h>
#include <LibIPC/Decoder.h>
#include <LibIPC/Encoder.h>
#include <LibIPC/File.h>
#include <LibIPC/Message.h>
#include <LibIPC/Stub.h>

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdefaulted-function-deleted"
#endif

namespace Messages::WebContentServer {

enum class MessageID : i32 {
    InitTransport = 1,
    InitTransportResponse = 2,
    CloseServer = 3,
    GetWindowHandle = 4,
    GetWindowHandleResponse = 5,
    SetWindowHandle = 6,
    ConnectToWebdriver = 7,
    ConnectToImageDecoder = 8,
    UpdateSystemTheme = 9,
    UpdateScreenRects = 10,
    LoadUrl = 11,
    LoadHtml = 12,
    Reload = 13,
    TraverseTheHistoryByDelta = 14,
    ReadyToPaint = 15,
    SetViewportSize = 16,
    KeyEvent = 17,
    MouseEvent = 18,
    DragEvent = 19,
    DebugRequest = 20,
    GetSource = 21,
    InspectDomTree = 22,
    InspectDomNode = 23,
    InspectAccessibilityTree = 24,
    GetHoveredNodeId = 25,
    JsConsoleInput = 26,
    JsConsoleRequestMessages = 27,
    ListStyleSheets = 28,
    RequestStyleSheetSource = 29,
    SetDomNodeText = 30,
    SetDomNodeTag = 31,
    AddDomNodeAttributes = 32,
    ReplaceDomNodeAttribute = 33,
    CreateChildElement = 34,
    CreateChildTextNode = 35,
    CloneDomNode = 36,
    RemoveDomNode = 37,
    GetDomNodeHtml = 38,
    TakeDocumentScreenshot = 39,
    TakeDomNodeScreenshot = 40,
    RequestInternalPageInfo = 41,
    RunJavascript = 42,
    GetSelectedText = 43,
    GetSelectedTextResponse = 44,
    SelectAll = 45,
    Paste = 46,
    FindInPage = 47,
    FindInPageNextMatch = 48,
    FindInPagePreviousMatch = 49,
    SetContentFilters = 50,
    SetAutoplayAllowedOnAllWebsites = 51,
    SetAutoplayAllowlist = 52,
    SetProxyMappings = 53,
    SetPreferredColorScheme = 54,
    SetPreferredContrast = 55,
    SetPreferredMotion = 56,
    SetPreferredLanguages = 57,
    SetEnableDoNotTrack = 58,
    SetHasFocus = 59,
    SetIsScriptingEnabled = 60,
    SetDevicePixelsPerCssPixel = 61,
    SetWindowPosition = 62,
    SetWindowSize = 63,
    DidUpdateWindowRect = 64,
    GetLocalStorageEntries = 65,
    GetLocalStorageEntriesResponse = 66,
    GetSessionStorageEntries = 67,
    GetSessionStorageEntriesResponse = 68,
    HandleFileReturn = 69,
    SetSystemVisibilityState = 70,
    AlertClosed = 71,
    ConfirmClosed = 72,
    PromptClosed = 73,
    ColorPickerUpdate = 74,
    FilePickerClosed = 75,
    SelectDropdownClosed = 76,
    ToggleMediaPlayState = 77,
    ToggleMediaMuteState = 78,
    ToggleMediaLoopState = 79,
    ToggleMediaControlsState = 80,
    TogglePageMuteState = 81,
    SetUserStyle = 82,
    EnableInspectorPrototype = 83,
    SystemTimeZoneChanged = 84,
};

class InitTransportResponse final : public IPC::Message {
public:

    InitTransportResponse(InitTransportResponse const&) = default;
    InitTransportResponse(InitTransportResponse&&) = default;
    InitTransportResponse& operator=(InitTransportResponse const&) = default;
    InitTransportResponse(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransportResponse(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransportResponse() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransportResponse; }
    static i32 static_message_id() { return (int)MessageID::InitTransportResponse; }
    virtual const char* message_name() const override { return "WebContentServer::InitTransportResponse"; }

    static ErrorOr<NonnullOwnPtr<InitTransportResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto peer_pid = TRY((decoder.decode<int>()));

        return make<InitTransportResponse>(move(peer_pid));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InitTransportResponse));

        TRY(stream.encode(m_peer_pid));

        return buffer;
    }

    const int& peer_pid() const { return m_peer_pid; }
    int take_peer_pid() { return move(m_peer_pid); }

private:

    int m_peer_pid;

};

class InitTransport final : public IPC::Message {
public:

   typedef class InitTransportResponse ResponseType;

    InitTransport(InitTransport const&) = default;
    InitTransport(InitTransport&&) = default;
    InitTransport& operator=(InitTransport const&) = default;
    InitTransport(int peer_pid) : m_peer_pid(move(peer_pid)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, int>)
    InitTransport(WrappedReturnType&& value)
        : m_peer_pid(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InitTransport() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::InitTransport; }
    static i32 static_message_id() { return (int)MessageID::InitTransport; }
    virtual const char* message_name() const override { return "WebContentServer::InitTransport"; }

    static ErrorOr<NonnullOwnPtr<InitTransport>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto peer_pid = TRY((decoder.decode<int>()));

        return make<InitTransport>(move(peer_pid));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InitTransport));

        TRY(stream.encode(m_peer_pid));

        return buffer;
    }

    const int& peer_pid() const { return m_peer_pid; }
    int take_peer_pid() { return move(m_peer_pid); }

private:

    int m_peer_pid;

};

class CloseServer final : public IPC::Message {
public:

    CloseServer(CloseServer const&) = default;
    CloseServer(CloseServer&&) = default;
    CloseServer& operator=(CloseServer const&) = default;
    CloseServer() {}

    virtual ~CloseServer() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::CloseServer; }
    static i32 static_message_id() { return (int)MessageID::CloseServer; }
    virtual const char* message_name() const override { return "WebContentServer::CloseServer"; }

    static ErrorOr<NonnullOwnPtr<CloseServer>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        return make<CloseServer>();
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::CloseServer));

        return buffer;
    }

private:

};

class GetWindowHandleResponse final : public IPC::Message {
public:

    GetWindowHandleResponse(GetWindowHandleResponse const&) = default;
    GetWindowHandleResponse(GetWindowHandleResponse&&) = default;
    GetWindowHandleResponse& operator=(GetWindowHandleResponse const&) = default;
    GetWindowHandleResponse(String handle) : m_handle(move(handle)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, String>)
    GetWindowHandleResponse(WrappedReturnType&& value)
        : m_handle(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetWindowHandleResponse() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetWindowHandleResponse; }
    static i32 static_message_id() { return (int)MessageID::GetWindowHandleResponse; }
    virtual const char* message_name() const override { return "WebContentServer::GetWindowHandleResponse"; }

    static ErrorOr<NonnullOwnPtr<GetWindowHandleResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto handle = TRY((decoder.decode<String>()));

        return make<GetWindowHandleResponse>(move(handle));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetWindowHandleResponse));

        TRY(stream.encode(m_handle));

        return buffer;
    }

    const String& handle() const { return m_handle; }
    String take_handle() { return move(m_handle); }

private:

    String m_handle;

};

class GetWindowHandle final : public IPC::Message {
public:

   typedef class GetWindowHandleResponse ResponseType;

    GetWindowHandle(GetWindowHandle const&) = default;
    GetWindowHandle(GetWindowHandle&&) = default;
    GetWindowHandle& operator=(GetWindowHandle const&) = default;
    GetWindowHandle(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetWindowHandle(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetWindowHandle() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetWindowHandle; }
    static i32 static_message_id() { return (int)MessageID::GetWindowHandle; }
    virtual const char* message_name() const override { return "WebContentServer::GetWindowHandle"; }

    static ErrorOr<NonnullOwnPtr<GetWindowHandle>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetWindowHandle>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetWindowHandle));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SetWindowHandle final : public IPC::Message {
public:

    SetWindowHandle(SetWindowHandle const&) = default;
    SetWindowHandle(SetWindowHandle&&) = default;
    SetWindowHandle& operator=(SetWindowHandle const&) = default;
    SetWindowHandle(u64 page_id, String handle) : m_page_id(move(page_id)), m_handle(move(handle)) {}

    virtual ~SetWindowHandle() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetWindowHandle; }
    static i32 static_message_id() { return (int)MessageID::SetWindowHandle; }
    virtual const char* message_name() const override { return "WebContentServer::SetWindowHandle"; }

    static ErrorOr<NonnullOwnPtr<SetWindowHandle>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto handle = TRY((decoder.decode<String>()));

        return make<SetWindowHandle>(move(page_id), move(handle));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetWindowHandle));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_handle));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& handle() const { return m_handle; }
    String take_handle() { return move(m_handle); }

private:

    u64 m_page_id;

    String m_handle;

};

class ConnectToWebdriver final : public IPC::Message {
public:

    ConnectToWebdriver(ConnectToWebdriver const&) = default;
    ConnectToWebdriver(ConnectToWebdriver&&) = default;
    ConnectToWebdriver& operator=(ConnectToWebdriver const&) = default;
    ConnectToWebdriver(u64 page_id, ByteString webdriver_ipc_path) : m_page_id(move(page_id)), m_webdriver_ipc_path(move(webdriver_ipc_path)) {}

    virtual ~ConnectToWebdriver() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectToWebdriver; }
    static i32 static_message_id() { return (int)MessageID::ConnectToWebdriver; }
    virtual const char* message_name() const override { return "WebContentServer::ConnectToWebdriver"; }

    static ErrorOr<NonnullOwnPtr<ConnectToWebdriver>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto webdriver_ipc_path = TRY((decoder.decode<ByteString>()));

        return make<ConnectToWebdriver>(move(page_id), move(webdriver_ipc_path));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ConnectToWebdriver));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_webdriver_ipc_path));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& webdriver_ipc_path() const { return m_webdriver_ipc_path; }
    ByteString take_webdriver_ipc_path() { return move(m_webdriver_ipc_path); }

private:

    u64 m_page_id;

    ByteString m_webdriver_ipc_path;

};

class ConnectToImageDecoder final : public IPC::Message {
public:

    ConnectToImageDecoder(ConnectToImageDecoder const&) = default;
    ConnectToImageDecoder(ConnectToImageDecoder&&) = default;
    ConnectToImageDecoder& operator=(ConnectToImageDecoder const&) = default;
    ConnectToImageDecoder(IPC::File socket_fd) : m_socket_fd(move(socket_fd)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, IPC::File>)
    ConnectToImageDecoder(WrappedReturnType&& value)
        : m_socket_fd(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ConnectToImageDecoder() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ConnectToImageDecoder; }
    static i32 static_message_id() { return (int)MessageID::ConnectToImageDecoder; }
    virtual const char* message_name() const override { return "WebContentServer::ConnectToImageDecoder"; }

    static ErrorOr<NonnullOwnPtr<ConnectToImageDecoder>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto socket_fd = TRY((decoder.decode<IPC::File>()));

        return make<ConnectToImageDecoder>(move(socket_fd));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ConnectToImageDecoder));

        TRY(stream.encode(m_socket_fd));

        return buffer;
    }

    const IPC::File& socket_fd() const { return m_socket_fd; }
    IPC::File take_socket_fd() { return move(m_socket_fd); }

private:

    IPC::File m_socket_fd;

};

class UpdateSystemTheme final : public IPC::Message {
public:

    UpdateSystemTheme(UpdateSystemTheme const&) = default;
    UpdateSystemTheme(UpdateSystemTheme&&) = default;
    UpdateSystemTheme& operator=(UpdateSystemTheme const&) = default;
    UpdateSystemTheme(u64 page_id, Core::AnonymousBuffer theme_buffer) : m_page_id(move(page_id)), m_theme_buffer(move(theme_buffer)) {}

    virtual ~UpdateSystemTheme() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::UpdateSystemTheme; }
    static i32 static_message_id() { return (int)MessageID::UpdateSystemTheme; }
    virtual const char* message_name() const override { return "WebContentServer::UpdateSystemTheme"; }

    static ErrorOr<NonnullOwnPtr<UpdateSystemTheme>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto theme_buffer = TRY((decoder.decode<Core::AnonymousBuffer>()));

        return make<UpdateSystemTheme>(move(page_id), move(theme_buffer));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::UpdateSystemTheme));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_theme_buffer));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Core::AnonymousBuffer& theme_buffer() const { return m_theme_buffer; }
    Core::AnonymousBuffer take_theme_buffer() { return move(m_theme_buffer); }

private:

    u64 m_page_id;

    Core::AnonymousBuffer m_theme_buffer;

};

class UpdateScreenRects final : public IPC::Message {
public:

    UpdateScreenRects(UpdateScreenRects const&) = default;
    UpdateScreenRects(UpdateScreenRects&&) = default;
    UpdateScreenRects& operator=(UpdateScreenRects const&) = default;
    UpdateScreenRects(u64 page_id, Vector<Web::DevicePixelRect> rects, u32 main_screen_index) : m_page_id(move(page_id)), m_rects(move(rects)), m_main_screen_index(move(main_screen_index)) {}

    virtual ~UpdateScreenRects() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::UpdateScreenRects; }
    static i32 static_message_id() { return (int)MessageID::UpdateScreenRects; }
    virtual const char* message_name() const override { return "WebContentServer::UpdateScreenRects"; }

    static ErrorOr<NonnullOwnPtr<UpdateScreenRects>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto rects = TRY((decoder.decode<Vector<Web::DevicePixelRect>>()));

        auto main_screen_index = TRY((decoder.decode<u32>()));

        return make<UpdateScreenRects>(move(page_id), move(rects), move(main_screen_index));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::UpdateScreenRects));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_rects));

        TRY(stream.encode(m_main_screen_index));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<Web::DevicePixelRect>& rects() const { return m_rects; }
    Vector<Web::DevicePixelRect> take_rects() { return move(m_rects); }

    const u32& main_screen_index() const { return m_main_screen_index; }
    u32 take_main_screen_index() { return move(m_main_screen_index); }

private:

    u64 m_page_id;

    Vector<Web::DevicePixelRect> m_rects;

    u32 m_main_screen_index;

};

class LoadUrl final : public IPC::Message {
public:

    LoadUrl(LoadUrl const&) = default;
    LoadUrl(LoadUrl&&) = default;
    LoadUrl& operator=(LoadUrl const&) = default;
    LoadUrl(u64 page_id, URL::URL url) : m_page_id(move(page_id)), m_url(move(url)) {}

    virtual ~LoadUrl() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::LoadUrl; }
    static i32 static_message_id() { return (int)MessageID::LoadUrl; }
    virtual const char* message_name() const override { return "WebContentServer::LoadUrl"; }

    static ErrorOr<NonnullOwnPtr<LoadUrl>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto url = TRY((decoder.decode<URL::URL>()));

        return make<LoadUrl>(move(page_id), move(url));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::LoadUrl));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_url));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const URL::URL& url() const { return m_url; }
    URL::URL take_url() { return move(m_url); }

private:

    u64 m_page_id;

    URL::URL m_url;

};

class LoadHtml final : public IPC::Message {
public:

    LoadHtml(LoadHtml const&) = default;
    LoadHtml(LoadHtml&&) = default;
    LoadHtml& operator=(LoadHtml const&) = default;
    LoadHtml(u64 page_id, ByteString html) : m_page_id(move(page_id)), m_html(move(html)) {}

    virtual ~LoadHtml() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::LoadHtml; }
    static i32 static_message_id() { return (int)MessageID::LoadHtml; }
    virtual const char* message_name() const override { return "WebContentServer::LoadHtml"; }

    static ErrorOr<NonnullOwnPtr<LoadHtml>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto html = TRY((decoder.decode<ByteString>()));

        return make<LoadHtml>(move(page_id), move(html));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::LoadHtml));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_html));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& html() const { return m_html; }
    ByteString take_html() { return move(m_html); }

private:

    u64 m_page_id;

    ByteString m_html;

};

class Reload final : public IPC::Message {
public:

    Reload(Reload const&) = default;
    Reload(Reload&&) = default;
    Reload& operator=(Reload const&) = default;
    Reload(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    Reload(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~Reload() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::Reload; }
    static i32 static_message_id() { return (int)MessageID::Reload; }
    virtual const char* message_name() const override { return "WebContentServer::Reload"; }

    static ErrorOr<NonnullOwnPtr<Reload>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<Reload>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::Reload));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class TraverseTheHistoryByDelta final : public IPC::Message {
public:

    TraverseTheHistoryByDelta(TraverseTheHistoryByDelta const&) = default;
    TraverseTheHistoryByDelta(TraverseTheHistoryByDelta&&) = default;
    TraverseTheHistoryByDelta& operator=(TraverseTheHistoryByDelta const&) = default;
    TraverseTheHistoryByDelta(u64 page_id, i32 delta) : m_page_id(move(page_id)), m_delta(move(delta)) {}

    virtual ~TraverseTheHistoryByDelta() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::TraverseTheHistoryByDelta; }
    static i32 static_message_id() { return (int)MessageID::TraverseTheHistoryByDelta; }
    virtual const char* message_name() const override { return "WebContentServer::TraverseTheHistoryByDelta"; }

    static ErrorOr<NonnullOwnPtr<TraverseTheHistoryByDelta>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto delta = TRY((decoder.decode<i32>()));

        return make<TraverseTheHistoryByDelta>(move(page_id), move(delta));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::TraverseTheHistoryByDelta));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_delta));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& delta() const { return m_delta; }
    i32 take_delta() { return move(m_delta); }

private:

    u64 m_page_id;

    i32 m_delta;

};

class ReadyToPaint final : public IPC::Message {
public:

    ReadyToPaint(ReadyToPaint const&) = default;
    ReadyToPaint(ReadyToPaint&&) = default;
    ReadyToPaint& operator=(ReadyToPaint const&) = default;
    ReadyToPaint(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ReadyToPaint(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ReadyToPaint() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ReadyToPaint; }
    static i32 static_message_id() { return (int)MessageID::ReadyToPaint; }
    virtual const char* message_name() const override { return "WebContentServer::ReadyToPaint"; }

    static ErrorOr<NonnullOwnPtr<ReadyToPaint>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ReadyToPaint>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ReadyToPaint));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SetViewportSize final : public IPC::Message {
public:

    SetViewportSize(SetViewportSize const&) = default;
    SetViewportSize(SetViewportSize&&) = default;
    SetViewportSize& operator=(SetViewportSize const&) = default;
    SetViewportSize(u64 page_id, Web::DevicePixelSize size) : m_page_id(move(page_id)), m_size(move(size)) {}

    virtual ~SetViewportSize() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetViewportSize; }
    static i32 static_message_id() { return (int)MessageID::SetViewportSize; }
    virtual const char* message_name() const override { return "WebContentServer::SetViewportSize"; }

    static ErrorOr<NonnullOwnPtr<SetViewportSize>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto size = TRY((decoder.decode<Web::DevicePixelSize>()));

        return make<SetViewportSize>(move(page_id), move(size));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetViewportSize));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_size));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::DevicePixelSize& size() const { return m_size; }
    Web::DevicePixelSize take_size() { return move(m_size); }

private:

    u64 m_page_id;

    Web::DevicePixelSize m_size;

};

class KeyEvent final : public IPC::Message {
public:

    KeyEvent(KeyEvent const&) = default;
    KeyEvent(KeyEvent&&) = default;
    KeyEvent& operator=(KeyEvent const&) = default;
    KeyEvent(u64 page_id, Web::KeyEvent event) : m_page_id(move(page_id)), m_event(move(event)) {}

    virtual ~KeyEvent() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::KeyEvent; }
    static i32 static_message_id() { return (int)MessageID::KeyEvent; }
    virtual const char* message_name() const override { return "WebContentServer::KeyEvent"; }

    static ErrorOr<NonnullOwnPtr<KeyEvent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto event = TRY((decoder.decode<Web::KeyEvent>()));

        return make<KeyEvent>(move(page_id), move(event));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::KeyEvent));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_event));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::KeyEvent& event() const { return m_event; }
    Web::KeyEvent take_event() { return move(m_event); }

private:

    u64 m_page_id;

    Web::KeyEvent m_event;

};

class MouseEvent final : public IPC::Message {
public:

    MouseEvent(MouseEvent const&) = default;
    MouseEvent(MouseEvent&&) = default;
    MouseEvent& operator=(MouseEvent const&) = default;
    MouseEvent(u64 page_id, Web::MouseEvent event) : m_page_id(move(page_id)), m_event(move(event)) {}

    virtual ~MouseEvent() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::MouseEvent; }
    static i32 static_message_id() { return (int)MessageID::MouseEvent; }
    virtual const char* message_name() const override { return "WebContentServer::MouseEvent"; }

    static ErrorOr<NonnullOwnPtr<MouseEvent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto event = TRY((decoder.decode<Web::MouseEvent>()));

        return make<MouseEvent>(move(page_id), move(event));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::MouseEvent));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_event));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::MouseEvent& event() const { return m_event; }
    Web::MouseEvent take_event() { return move(m_event); }

private:

    u64 m_page_id;

    Web::MouseEvent m_event;

};

class DragEvent final : public IPC::Message {
public:

    DragEvent(DragEvent const&) = default;
    DragEvent(DragEvent&&) = default;
    DragEvent& operator=(DragEvent const&) = default;
    DragEvent(u64 page_id, Web::DragEvent event) : m_page_id(move(page_id)), m_event(move(event)) {}

    virtual ~DragEvent() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::DragEvent; }
    static i32 static_message_id() { return (int)MessageID::DragEvent; }
    virtual const char* message_name() const override { return "WebContentServer::DragEvent"; }

    static ErrorOr<NonnullOwnPtr<DragEvent>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto event = TRY((decoder.decode<Web::DragEvent>()));

        return make<DragEvent>(move(page_id), move(event));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DragEvent));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_event));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::DragEvent& event() const { return m_event; }
    Web::DragEvent take_event() { return move(m_event); }

private:

    u64 m_page_id;

    Web::DragEvent m_event;

};

class DebugRequest final : public IPC::Message {
public:

    DebugRequest(DebugRequest const&) = default;
    DebugRequest(DebugRequest&&) = default;
    DebugRequest& operator=(DebugRequest const&) = default;
    DebugRequest(u64 page_id, ByteString request, ByteString argument) : m_page_id(move(page_id)), m_request(move(request)), m_argument(move(argument)) {}

    virtual ~DebugRequest() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::DebugRequest; }
    static i32 static_message_id() { return (int)MessageID::DebugRequest; }
    virtual const char* message_name() const override { return "WebContentServer::DebugRequest"; }

    static ErrorOr<NonnullOwnPtr<DebugRequest>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto request = TRY((decoder.decode<ByteString>()));

        auto argument = TRY((decoder.decode<ByteString>()));

        return make<DebugRequest>(move(page_id), move(request), move(argument));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DebugRequest));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_request));

        TRY(stream.encode(m_argument));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& request() const { return m_request; }
    ByteString take_request() { return move(m_request); }

    const ByteString& argument() const { return m_argument; }
    ByteString take_argument() { return move(m_argument); }

private:

    u64 m_page_id;

    ByteString m_request;

    ByteString m_argument;

};

class GetSource final : public IPC::Message {
public:

    GetSource(GetSource const&) = default;
    GetSource(GetSource&&) = default;
    GetSource& operator=(GetSource const&) = default;
    GetSource(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetSource(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetSource() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetSource; }
    static i32 static_message_id() { return (int)MessageID::GetSource; }
    virtual const char* message_name() const override { return "WebContentServer::GetSource"; }

    static ErrorOr<NonnullOwnPtr<GetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetSource>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetSource));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class InspectDomTree final : public IPC::Message {
public:

    InspectDomTree(InspectDomTree const&) = default;
    InspectDomTree(InspectDomTree&&) = default;
    InspectDomTree& operator=(InspectDomTree const&) = default;
    InspectDomTree(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    InspectDomTree(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InspectDomTree() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::InspectDomTree; }
    static i32 static_message_id() { return (int)MessageID::InspectDomTree; }
    virtual const char* message_name() const override { return "WebContentServer::InspectDomTree"; }

    static ErrorOr<NonnullOwnPtr<InspectDomTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<InspectDomTree>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectDomTree));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class InspectDomNode final : public IPC::Message {
public:

    InspectDomNode(InspectDomNode const&) = default;
    InspectDomNode(InspectDomNode&&) = default;
    InspectDomNode& operator=(InspectDomNode const&) = default;
    InspectDomNode(u64 page_id, Web::UniqueNodeID node_id, Optional<Web::CSS::Selector::PseudoElement::Type> pseudo_element) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_pseudo_element(move(pseudo_element)) {}

    virtual ~InspectDomNode() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::InspectDomNode; }
    static i32 static_message_id() { return (int)MessageID::InspectDomNode; }
    virtual const char* message_name() const override { return "WebContentServer::InspectDomNode"; }

    static ErrorOr<NonnullOwnPtr<InspectDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto pseudo_element = TRY((decoder.decode<Optional<Web::CSS::Selector::PseudoElement::Type>>()));

        return make<InspectDomNode>(move(page_id), move(node_id), move(pseudo_element));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_pseudo_element));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const Optional<Web::CSS::Selector::PseudoElement::Type>& pseudo_element() const { return m_pseudo_element; }
    Optional<Web::CSS::Selector::PseudoElement::Type> take_pseudo_element() { return move(m_pseudo_element); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    Optional<Web::CSS::Selector::PseudoElement::Type> m_pseudo_element;

};

class InspectAccessibilityTree final : public IPC::Message {
public:

    InspectAccessibilityTree(InspectAccessibilityTree const&) = default;
    InspectAccessibilityTree(InspectAccessibilityTree&&) = default;
    InspectAccessibilityTree& operator=(InspectAccessibilityTree const&) = default;
    InspectAccessibilityTree(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    InspectAccessibilityTree(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~InspectAccessibilityTree() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::InspectAccessibilityTree; }
    static i32 static_message_id() { return (int)MessageID::InspectAccessibilityTree; }
    virtual const char* message_name() const override { return "WebContentServer::InspectAccessibilityTree"; }

    static ErrorOr<NonnullOwnPtr<InspectAccessibilityTree>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<InspectAccessibilityTree>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::InspectAccessibilityTree));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class GetHoveredNodeId final : public IPC::Message {
public:

    GetHoveredNodeId(GetHoveredNodeId const&) = default;
    GetHoveredNodeId(GetHoveredNodeId&&) = default;
    GetHoveredNodeId& operator=(GetHoveredNodeId const&) = default;
    GetHoveredNodeId(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetHoveredNodeId(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetHoveredNodeId() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetHoveredNodeId; }
    static i32 static_message_id() { return (int)MessageID::GetHoveredNodeId; }
    virtual const char* message_name() const override { return "WebContentServer::GetHoveredNodeId"; }

    static ErrorOr<NonnullOwnPtr<GetHoveredNodeId>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetHoveredNodeId>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetHoveredNodeId));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class JsConsoleInput final : public IPC::Message {
public:

    JsConsoleInput(JsConsoleInput const&) = default;
    JsConsoleInput(JsConsoleInput&&) = default;
    JsConsoleInput& operator=(JsConsoleInput const&) = default;
    JsConsoleInput(u64 page_id, ByteString js_source) : m_page_id(move(page_id)), m_js_source(move(js_source)) {}

    virtual ~JsConsoleInput() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::JsConsoleInput; }
    static i32 static_message_id() { return (int)MessageID::JsConsoleInput; }
    virtual const char* message_name() const override { return "WebContentServer::JsConsoleInput"; }

    static ErrorOr<NonnullOwnPtr<JsConsoleInput>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto js_source = TRY((decoder.decode<ByteString>()));

        return make<JsConsoleInput>(move(page_id), move(js_source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::JsConsoleInput));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_js_source));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& js_source() const { return m_js_source; }
    ByteString take_js_source() { return move(m_js_source); }

private:

    u64 m_page_id;

    ByteString m_js_source;

};

class JsConsoleRequestMessages final : public IPC::Message {
public:

    JsConsoleRequestMessages(JsConsoleRequestMessages const&) = default;
    JsConsoleRequestMessages(JsConsoleRequestMessages&&) = default;
    JsConsoleRequestMessages& operator=(JsConsoleRequestMessages const&) = default;
    JsConsoleRequestMessages(u64 page_id, i32 start_index) : m_page_id(move(page_id)), m_start_index(move(start_index)) {}

    virtual ~JsConsoleRequestMessages() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::JsConsoleRequestMessages; }
    static i32 static_message_id() { return (int)MessageID::JsConsoleRequestMessages; }
    virtual const char* message_name() const override { return "WebContentServer::JsConsoleRequestMessages"; }

    static ErrorOr<NonnullOwnPtr<JsConsoleRequestMessages>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto start_index = TRY((decoder.decode<i32>()));

        return make<JsConsoleRequestMessages>(move(page_id), move(start_index));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::JsConsoleRequestMessages));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_start_index));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& start_index() const { return m_start_index; }
    i32 take_start_index() { return move(m_start_index); }

private:

    u64 m_page_id;

    i32 m_start_index;

};

class ListStyleSheets final : public IPC::Message {
public:

    ListStyleSheets(ListStyleSheets const&) = default;
    ListStyleSheets(ListStyleSheets&&) = default;
    ListStyleSheets& operator=(ListStyleSheets const&) = default;
    ListStyleSheets(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ListStyleSheets(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ListStyleSheets() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ListStyleSheets; }
    static i32 static_message_id() { return (int)MessageID::ListStyleSheets; }
    virtual const char* message_name() const override { return "WebContentServer::ListStyleSheets"; }

    static ErrorOr<NonnullOwnPtr<ListStyleSheets>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ListStyleSheets>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ListStyleSheets));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class RequestStyleSheetSource final : public IPC::Message {
public:

    RequestStyleSheetSource(RequestStyleSheetSource const&) = default;
    RequestStyleSheetSource(RequestStyleSheetSource&&) = default;
    RequestStyleSheetSource& operator=(RequestStyleSheetSource const&) = default;
    RequestStyleSheetSource(u64 page_id, Web::CSS::StyleSheetIdentifier identifier) : m_page_id(move(page_id)), m_identifier(move(identifier)) {}

    virtual ~RequestStyleSheetSource() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::RequestStyleSheetSource; }
    static i32 static_message_id() { return (int)MessageID::RequestStyleSheetSource; }
    virtual const char* message_name() const override { return "WebContentServer::RequestStyleSheetSource"; }

    static ErrorOr<NonnullOwnPtr<RequestStyleSheetSource>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto identifier = TRY((decoder.decode<Web::CSS::StyleSheetIdentifier>()));

        return make<RequestStyleSheetSource>(move(page_id), move(identifier));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RequestStyleSheetSource));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_identifier));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::StyleSheetIdentifier& identifier() const { return m_identifier; }
    Web::CSS::StyleSheetIdentifier take_identifier() { return move(m_identifier); }

private:

    u64 m_page_id;

    Web::CSS::StyleSheetIdentifier m_identifier;

};

class SetDomNodeText final : public IPC::Message {
public:

    SetDomNodeText(SetDomNodeText const&) = default;
    SetDomNodeText(SetDomNodeText&&) = default;
    SetDomNodeText& operator=(SetDomNodeText const&) = default;
    SetDomNodeText(u64 page_id, Web::UniqueNodeID node_id, String text) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_text(move(text)) {}

    virtual ~SetDomNodeText() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetDomNodeText; }
    static i32 static_message_id() { return (int)MessageID::SetDomNodeText; }
    virtual const char* message_name() const override { return "WebContentServer::SetDomNodeText"; }

    static ErrorOr<NonnullOwnPtr<SetDomNodeText>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto text = TRY((decoder.decode<String>()));

        return make<SetDomNodeText>(move(page_id), move(node_id), move(text));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetDomNodeText));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_text));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const String& text() const { return m_text; }
    String take_text() { return move(m_text); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    String m_text;

};

class SetDomNodeTag final : public IPC::Message {
public:

    SetDomNodeTag(SetDomNodeTag const&) = default;
    SetDomNodeTag(SetDomNodeTag&&) = default;
    SetDomNodeTag& operator=(SetDomNodeTag const&) = default;
    SetDomNodeTag(u64 page_id, Web::UniqueNodeID node_id, String name) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_name(move(name)) {}

    virtual ~SetDomNodeTag() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetDomNodeTag; }
    static i32 static_message_id() { return (int)MessageID::SetDomNodeTag; }
    virtual const char* message_name() const override { return "WebContentServer::SetDomNodeTag"; }

    static ErrorOr<NonnullOwnPtr<SetDomNodeTag>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto name = TRY((decoder.decode<String>()));

        return make<SetDomNodeTag>(move(page_id), move(node_id), move(name));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetDomNodeTag));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_name));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    String m_name;

};

class AddDomNodeAttributes final : public IPC::Message {
public:

    AddDomNodeAttributes(AddDomNodeAttributes const&) = default;
    AddDomNodeAttributes(AddDomNodeAttributes&&) = default;
    AddDomNodeAttributes& operator=(AddDomNodeAttributes const&) = default;
    AddDomNodeAttributes(u64 page_id, Web::UniqueNodeID node_id, Vector<WebView::Attribute> attributes) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_attributes(move(attributes)) {}

    virtual ~AddDomNodeAttributes() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::AddDomNodeAttributes; }
    static i32 static_message_id() { return (int)MessageID::AddDomNodeAttributes; }
    virtual const char* message_name() const override { return "WebContentServer::AddDomNodeAttributes"; }

    static ErrorOr<NonnullOwnPtr<AddDomNodeAttributes>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto attributes = TRY((decoder.decode<Vector<WebView::Attribute>>()));

        return make<AddDomNodeAttributes>(move(page_id), move(node_id), move(attributes));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::AddDomNodeAttributes));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_attributes));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const Vector<WebView::Attribute>& attributes() const { return m_attributes; }
    Vector<WebView::Attribute> take_attributes() { return move(m_attributes); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    Vector<WebView::Attribute> m_attributes;

};

class ReplaceDomNodeAttribute final : public IPC::Message {
public:

    ReplaceDomNodeAttribute(ReplaceDomNodeAttribute const&) = default;
    ReplaceDomNodeAttribute(ReplaceDomNodeAttribute&&) = default;
    ReplaceDomNodeAttribute& operator=(ReplaceDomNodeAttribute const&) = default;
    ReplaceDomNodeAttribute(u64 page_id, Web::UniqueNodeID node_id, String name, Vector<WebView::Attribute> replacement_attributes) : m_page_id(move(page_id)), m_node_id(move(node_id)), m_name(move(name)), m_replacement_attributes(move(replacement_attributes)) {}

    virtual ~ReplaceDomNodeAttribute() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ReplaceDomNodeAttribute; }
    static i32 static_message_id() { return (int)MessageID::ReplaceDomNodeAttribute; }
    virtual const char* message_name() const override { return "WebContentServer::ReplaceDomNodeAttribute"; }

    static ErrorOr<NonnullOwnPtr<ReplaceDomNodeAttribute>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        auto name = TRY((decoder.decode<String>()));

        auto replacement_attributes = TRY((decoder.decode<Vector<WebView::Attribute>>()));

        return make<ReplaceDomNodeAttribute>(move(page_id), move(node_id), move(name), move(replacement_attributes));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ReplaceDomNodeAttribute));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        TRY(stream.encode(m_name));

        TRY(stream.encode(m_replacement_attributes));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

    const String& name() const { return m_name; }
    String take_name() { return move(m_name); }

    const Vector<WebView::Attribute>& replacement_attributes() const { return m_replacement_attributes; }
    Vector<WebView::Attribute> take_replacement_attributes() { return move(m_replacement_attributes); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

    String m_name;

    Vector<WebView::Attribute> m_replacement_attributes;

};

class CreateChildElement final : public IPC::Message {
public:

    CreateChildElement(CreateChildElement const&) = default;
    CreateChildElement(CreateChildElement&&) = default;
    CreateChildElement& operator=(CreateChildElement const&) = default;
    CreateChildElement(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~CreateChildElement() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::CreateChildElement; }
    static i32 static_message_id() { return (int)MessageID::CreateChildElement; }
    virtual const char* message_name() const override { return "WebContentServer::CreateChildElement"; }

    static ErrorOr<NonnullOwnPtr<CreateChildElement>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<CreateChildElement>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::CreateChildElement));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class CreateChildTextNode final : public IPC::Message {
public:

    CreateChildTextNode(CreateChildTextNode const&) = default;
    CreateChildTextNode(CreateChildTextNode&&) = default;
    CreateChildTextNode& operator=(CreateChildTextNode const&) = default;
    CreateChildTextNode(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~CreateChildTextNode() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::CreateChildTextNode; }
    static i32 static_message_id() { return (int)MessageID::CreateChildTextNode; }
    virtual const char* message_name() const override { return "WebContentServer::CreateChildTextNode"; }

    static ErrorOr<NonnullOwnPtr<CreateChildTextNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<CreateChildTextNode>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::CreateChildTextNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class CloneDomNode final : public IPC::Message {
public:

    CloneDomNode(CloneDomNode const&) = default;
    CloneDomNode(CloneDomNode&&) = default;
    CloneDomNode& operator=(CloneDomNode const&) = default;
    CloneDomNode(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~CloneDomNode() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::CloneDomNode; }
    static i32 static_message_id() { return (int)MessageID::CloneDomNode; }
    virtual const char* message_name() const override { return "WebContentServer::CloneDomNode"; }

    static ErrorOr<NonnullOwnPtr<CloneDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<CloneDomNode>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::CloneDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class RemoveDomNode final : public IPC::Message {
public:

    RemoveDomNode(RemoveDomNode const&) = default;
    RemoveDomNode(RemoveDomNode&&) = default;
    RemoveDomNode& operator=(RemoveDomNode const&) = default;
    RemoveDomNode(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~RemoveDomNode() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::RemoveDomNode; }
    static i32 static_message_id() { return (int)MessageID::RemoveDomNode; }
    virtual const char* message_name() const override { return "WebContentServer::RemoveDomNode"; }

    static ErrorOr<NonnullOwnPtr<RemoveDomNode>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<RemoveDomNode>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RemoveDomNode));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class GetDomNodeHtml final : public IPC::Message {
public:

    GetDomNodeHtml(GetDomNodeHtml const&) = default;
    GetDomNodeHtml(GetDomNodeHtml&&) = default;
    GetDomNodeHtml& operator=(GetDomNodeHtml const&) = default;
    GetDomNodeHtml(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~GetDomNodeHtml() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetDomNodeHtml; }
    static i32 static_message_id() { return (int)MessageID::GetDomNodeHtml; }
    virtual const char* message_name() const override { return "WebContentServer::GetDomNodeHtml"; }

    static ErrorOr<NonnullOwnPtr<GetDomNodeHtml>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<GetDomNodeHtml>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetDomNodeHtml));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class TakeDocumentScreenshot final : public IPC::Message {
public:

    TakeDocumentScreenshot(TakeDocumentScreenshot const&) = default;
    TakeDocumentScreenshot(TakeDocumentScreenshot&&) = default;
    TakeDocumentScreenshot& operator=(TakeDocumentScreenshot const&) = default;
    TakeDocumentScreenshot(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    TakeDocumentScreenshot(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~TakeDocumentScreenshot() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::TakeDocumentScreenshot; }
    static i32 static_message_id() { return (int)MessageID::TakeDocumentScreenshot; }
    virtual const char* message_name() const override { return "WebContentServer::TakeDocumentScreenshot"; }

    static ErrorOr<NonnullOwnPtr<TakeDocumentScreenshot>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<TakeDocumentScreenshot>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::TakeDocumentScreenshot));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class TakeDomNodeScreenshot final : public IPC::Message {
public:

    TakeDomNodeScreenshot(TakeDomNodeScreenshot const&) = default;
    TakeDomNodeScreenshot(TakeDomNodeScreenshot&&) = default;
    TakeDomNodeScreenshot& operator=(TakeDomNodeScreenshot const&) = default;
    TakeDomNodeScreenshot(u64 page_id, Web::UniqueNodeID node_id) : m_page_id(move(page_id)), m_node_id(move(node_id)) {}

    virtual ~TakeDomNodeScreenshot() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::TakeDomNodeScreenshot; }
    static i32 static_message_id() { return (int)MessageID::TakeDomNodeScreenshot; }
    virtual const char* message_name() const override { return "WebContentServer::TakeDomNodeScreenshot"; }

    static ErrorOr<NonnullOwnPtr<TakeDomNodeScreenshot>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto node_id = TRY((decoder.decode<Web::UniqueNodeID>()));

        return make<TakeDomNodeScreenshot>(move(page_id), move(node_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::TakeDomNodeScreenshot));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_node_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::UniqueNodeID& node_id() const { return m_node_id; }
    Web::UniqueNodeID take_node_id() { return move(m_node_id); }

private:

    u64 m_page_id;

    Web::UniqueNodeID m_node_id;

};

class RequestInternalPageInfo final : public IPC::Message {
public:

    RequestInternalPageInfo(RequestInternalPageInfo const&) = default;
    RequestInternalPageInfo(RequestInternalPageInfo&&) = default;
    RequestInternalPageInfo& operator=(RequestInternalPageInfo const&) = default;
    RequestInternalPageInfo(u64 page_id, WebView::PageInfoType type) : m_page_id(move(page_id)), m_type(move(type)) {}

    virtual ~RequestInternalPageInfo() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::RequestInternalPageInfo; }
    static i32 static_message_id() { return (int)MessageID::RequestInternalPageInfo; }
    virtual const char* message_name() const override { return "WebContentServer::RequestInternalPageInfo"; }

    static ErrorOr<NonnullOwnPtr<RequestInternalPageInfo>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto type = TRY((decoder.decode<WebView::PageInfoType>()));

        return make<RequestInternalPageInfo>(move(page_id), move(type));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RequestInternalPageInfo));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_type));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const WebView::PageInfoType& type() const { return m_type; }
    WebView::PageInfoType take_type() { return move(m_type); }

private:

    u64 m_page_id;

    WebView::PageInfoType m_type;

};

class RunJavascript final : public IPC::Message {
public:

    RunJavascript(RunJavascript const&) = default;
    RunJavascript(RunJavascript&&) = default;
    RunJavascript& operator=(RunJavascript const&) = default;
    RunJavascript(u64 page_id, ByteString js_source) : m_page_id(move(page_id)), m_js_source(move(js_source)) {}

    virtual ~RunJavascript() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::RunJavascript; }
    static i32 static_message_id() { return (int)MessageID::RunJavascript; }
    virtual const char* message_name() const override { return "WebContentServer::RunJavascript"; }

    static ErrorOr<NonnullOwnPtr<RunJavascript>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto js_source = TRY((decoder.decode<ByteString>()));

        return make<RunJavascript>(move(page_id), move(js_source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::RunJavascript));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_js_source));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const ByteString& js_source() const { return m_js_source; }
    ByteString take_js_source() { return move(m_js_source); }

private:

    u64 m_page_id;

    ByteString m_js_source;

};

class GetSelectedTextResponse final : public IPC::Message {
public:

    GetSelectedTextResponse(GetSelectedTextResponse const&) = default;
    GetSelectedTextResponse(GetSelectedTextResponse&&) = default;
    GetSelectedTextResponse& operator=(GetSelectedTextResponse const&) = default;
    GetSelectedTextResponse(ByteString selection) : m_selection(move(selection)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, ByteString>)
    GetSelectedTextResponse(WrappedReturnType&& value)
        : m_selection(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetSelectedTextResponse() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetSelectedTextResponse; }
    static i32 static_message_id() { return (int)MessageID::GetSelectedTextResponse; }
    virtual const char* message_name() const override { return "WebContentServer::GetSelectedTextResponse"; }

    static ErrorOr<NonnullOwnPtr<GetSelectedTextResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto selection = TRY((decoder.decode<ByteString>()));

        return make<GetSelectedTextResponse>(move(selection));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetSelectedTextResponse));

        TRY(stream.encode(m_selection));

        return buffer;
    }

    const ByteString& selection() const { return m_selection; }
    ByteString take_selection() { return move(m_selection); }

private:

    ByteString m_selection;

};

class GetSelectedText final : public IPC::Message {
public:

   typedef class GetSelectedTextResponse ResponseType;

    GetSelectedText(GetSelectedText const&) = default;
    GetSelectedText(GetSelectedText&&) = default;
    GetSelectedText& operator=(GetSelectedText const&) = default;
    GetSelectedText(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetSelectedText(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetSelectedText() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetSelectedText; }
    static i32 static_message_id() { return (int)MessageID::GetSelectedText; }
    virtual const char* message_name() const override { return "WebContentServer::GetSelectedText"; }

    static ErrorOr<NonnullOwnPtr<GetSelectedText>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetSelectedText>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetSelectedText));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SelectAll final : public IPC::Message {
public:

    SelectAll(SelectAll const&) = default;
    SelectAll(SelectAll&&) = default;
    SelectAll& operator=(SelectAll const&) = default;
    SelectAll(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    SelectAll(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~SelectAll() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SelectAll; }
    static i32 static_message_id() { return (int)MessageID::SelectAll; }
    virtual const char* message_name() const override { return "WebContentServer::SelectAll"; }

    static ErrorOr<NonnullOwnPtr<SelectAll>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<SelectAll>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SelectAll));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class Paste final : public IPC::Message {
public:

    Paste(Paste const&) = default;
    Paste(Paste&&) = default;
    Paste& operator=(Paste const&) = default;
    Paste(u64 page_id, String text) : m_page_id(move(page_id)), m_text(move(text)) {}

    virtual ~Paste() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::Paste; }
    static i32 static_message_id() { return (int)MessageID::Paste; }
    virtual const char* message_name() const override { return "WebContentServer::Paste"; }

    static ErrorOr<NonnullOwnPtr<Paste>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto text = TRY((decoder.decode<String>()));

        return make<Paste>(move(page_id), move(text));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::Paste));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_text));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& text() const { return m_text; }
    String take_text() { return move(m_text); }

private:

    u64 m_page_id;

    String m_text;

};

class FindInPage final : public IPC::Message {
public:

    FindInPage(FindInPage const&) = default;
    FindInPage(FindInPage&&) = default;
    FindInPage& operator=(FindInPage const&) = default;
    FindInPage(u64 page_id, String query, AK::CaseSensitivity case_sensitivity) : m_page_id(move(page_id)), m_query(move(query)), m_case_sensitivity(move(case_sensitivity)) {}

    virtual ~FindInPage() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::FindInPage; }
    static i32 static_message_id() { return (int)MessageID::FindInPage; }
    virtual const char* message_name() const override { return "WebContentServer::FindInPage"; }

    static ErrorOr<NonnullOwnPtr<FindInPage>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto query = TRY((decoder.decode<String>()));

        auto case_sensitivity = TRY((decoder.decode<AK::CaseSensitivity>()));

        return make<FindInPage>(move(page_id), move(query), move(case_sensitivity));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::FindInPage));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_query));

        TRY(stream.encode(m_case_sensitivity));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& query() const { return m_query; }
    String take_query() { return move(m_query); }

    const AK::CaseSensitivity& case_sensitivity() const { return m_case_sensitivity; }
    AK::CaseSensitivity take_case_sensitivity() { return move(m_case_sensitivity); }

private:

    u64 m_page_id;

    String m_query;

    AK::CaseSensitivity m_case_sensitivity;

};

class FindInPageNextMatch final : public IPC::Message {
public:

    FindInPageNextMatch(FindInPageNextMatch const&) = default;
    FindInPageNextMatch(FindInPageNextMatch&&) = default;
    FindInPageNextMatch& operator=(FindInPageNextMatch const&) = default;
    FindInPageNextMatch(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    FindInPageNextMatch(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~FindInPageNextMatch() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::FindInPageNextMatch; }
    static i32 static_message_id() { return (int)MessageID::FindInPageNextMatch; }
    virtual const char* message_name() const override { return "WebContentServer::FindInPageNextMatch"; }

    static ErrorOr<NonnullOwnPtr<FindInPageNextMatch>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<FindInPageNextMatch>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::FindInPageNextMatch));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class FindInPagePreviousMatch final : public IPC::Message {
public:

    FindInPagePreviousMatch(FindInPagePreviousMatch const&) = default;
    FindInPagePreviousMatch(FindInPagePreviousMatch&&) = default;
    FindInPagePreviousMatch& operator=(FindInPagePreviousMatch const&) = default;
    FindInPagePreviousMatch(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    FindInPagePreviousMatch(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~FindInPagePreviousMatch() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::FindInPagePreviousMatch; }
    static i32 static_message_id() { return (int)MessageID::FindInPagePreviousMatch; }
    virtual const char* message_name() const override { return "WebContentServer::FindInPagePreviousMatch"; }

    static ErrorOr<NonnullOwnPtr<FindInPagePreviousMatch>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<FindInPagePreviousMatch>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::FindInPagePreviousMatch));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SetContentFilters final : public IPC::Message {
public:

    SetContentFilters(SetContentFilters const&) = default;
    SetContentFilters(SetContentFilters&&) = default;
    SetContentFilters& operator=(SetContentFilters const&) = default;
    SetContentFilters(u64 page_id, Vector<String> filters) : m_page_id(move(page_id)), m_filters(move(filters)) {}

    virtual ~SetContentFilters() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetContentFilters; }
    static i32 static_message_id() { return (int)MessageID::SetContentFilters; }
    virtual const char* message_name() const override { return "WebContentServer::SetContentFilters"; }

    static ErrorOr<NonnullOwnPtr<SetContentFilters>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto filters = TRY((decoder.decode<Vector<String>>()));

        return make<SetContentFilters>(move(page_id), move(filters));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetContentFilters));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_filters));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<String>& filters() const { return m_filters; }
    Vector<String> take_filters() { return move(m_filters); }

private:

    u64 m_page_id;

    Vector<String> m_filters;

};

class SetAutoplayAllowedOnAllWebsites final : public IPC::Message {
public:

    SetAutoplayAllowedOnAllWebsites(SetAutoplayAllowedOnAllWebsites const&) = default;
    SetAutoplayAllowedOnAllWebsites(SetAutoplayAllowedOnAllWebsites&&) = default;
    SetAutoplayAllowedOnAllWebsites& operator=(SetAutoplayAllowedOnAllWebsites const&) = default;
    SetAutoplayAllowedOnAllWebsites(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    SetAutoplayAllowedOnAllWebsites(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~SetAutoplayAllowedOnAllWebsites() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetAutoplayAllowedOnAllWebsites; }
    static i32 static_message_id() { return (int)MessageID::SetAutoplayAllowedOnAllWebsites; }
    virtual const char* message_name() const override { return "WebContentServer::SetAutoplayAllowedOnAllWebsites"; }

    static ErrorOr<NonnullOwnPtr<SetAutoplayAllowedOnAllWebsites>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<SetAutoplayAllowedOnAllWebsites>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetAutoplayAllowedOnAllWebsites));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SetAutoplayAllowlist final : public IPC::Message {
public:

    SetAutoplayAllowlist(SetAutoplayAllowlist const&) = default;
    SetAutoplayAllowlist(SetAutoplayAllowlist&&) = default;
    SetAutoplayAllowlist& operator=(SetAutoplayAllowlist const&) = default;
    SetAutoplayAllowlist(u64 page_id, Vector<String> allowlist) : m_page_id(move(page_id)), m_allowlist(move(allowlist)) {}

    virtual ~SetAutoplayAllowlist() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetAutoplayAllowlist; }
    static i32 static_message_id() { return (int)MessageID::SetAutoplayAllowlist; }
    virtual const char* message_name() const override { return "WebContentServer::SetAutoplayAllowlist"; }

    static ErrorOr<NonnullOwnPtr<SetAutoplayAllowlist>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto allowlist = TRY((decoder.decode<Vector<String>>()));

        return make<SetAutoplayAllowlist>(move(page_id), move(allowlist));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetAutoplayAllowlist));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_allowlist));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<String>& allowlist() const { return m_allowlist; }
    Vector<String> take_allowlist() { return move(m_allowlist); }

private:

    u64 m_page_id;

    Vector<String> m_allowlist;

};

class SetProxyMappings final : public IPC::Message {
public:

    SetProxyMappings(SetProxyMappings const&) = default;
    SetProxyMappings(SetProxyMappings&&) = default;
    SetProxyMappings& operator=(SetProxyMappings const&) = default;
    SetProxyMappings(u64 page_id, Vector<ByteString> proxies, HashMap<ByteString, size_t> mappings) : m_page_id(move(page_id)), m_proxies(move(proxies)), m_mappings(move(mappings)) {}

    virtual ~SetProxyMappings() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetProxyMappings; }
    static i32 static_message_id() { return (int)MessageID::SetProxyMappings; }
    virtual const char* message_name() const override { return "WebContentServer::SetProxyMappings"; }

    static ErrorOr<NonnullOwnPtr<SetProxyMappings>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto proxies = TRY((decoder.decode<Vector<ByteString>>()));

        auto mappings = TRY((decoder.decode<HashMap<ByteString, size_t>>()));

        return make<SetProxyMappings>(move(page_id), move(proxies), move(mappings));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetProxyMappings));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_proxies));

        TRY(stream.encode(m_mappings));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<ByteString>& proxies() const { return m_proxies; }
    Vector<ByteString> take_proxies() { return move(m_proxies); }

    const HashMap<ByteString, size_t>& mappings() const { return m_mappings; }
    HashMap<ByteString, size_t> take_mappings() { return move(m_mappings); }

private:

    u64 m_page_id;

    Vector<ByteString> m_proxies;

    HashMap<ByteString, size_t> m_mappings;

};

class SetPreferredColorScheme final : public IPC::Message {
public:

    SetPreferredColorScheme(SetPreferredColorScheme const&) = default;
    SetPreferredColorScheme(SetPreferredColorScheme&&) = default;
    SetPreferredColorScheme& operator=(SetPreferredColorScheme const&) = default;
    SetPreferredColorScheme(u64 page_id, Web::CSS::PreferredColorScheme color_scheme) : m_page_id(move(page_id)), m_color_scheme(move(color_scheme)) {}

    virtual ~SetPreferredColorScheme() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetPreferredColorScheme; }
    static i32 static_message_id() { return (int)MessageID::SetPreferredColorScheme; }
    virtual const char* message_name() const override { return "WebContentServer::SetPreferredColorScheme"; }

    static ErrorOr<NonnullOwnPtr<SetPreferredColorScheme>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto color_scheme = TRY((decoder.decode<Web::CSS::PreferredColorScheme>()));

        return make<SetPreferredColorScheme>(move(page_id), move(color_scheme));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetPreferredColorScheme));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_color_scheme));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::PreferredColorScheme& color_scheme() const { return m_color_scheme; }
    Web::CSS::PreferredColorScheme take_color_scheme() { return move(m_color_scheme); }

private:

    u64 m_page_id;

    Web::CSS::PreferredColorScheme m_color_scheme;

};

class SetPreferredContrast final : public IPC::Message {
public:

    SetPreferredContrast(SetPreferredContrast const&) = default;
    SetPreferredContrast(SetPreferredContrast&&) = default;
    SetPreferredContrast& operator=(SetPreferredContrast const&) = default;
    SetPreferredContrast(u64 page_id, Web::CSS::PreferredContrast contrast) : m_page_id(move(page_id)), m_contrast(move(contrast)) {}

    virtual ~SetPreferredContrast() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetPreferredContrast; }
    static i32 static_message_id() { return (int)MessageID::SetPreferredContrast; }
    virtual const char* message_name() const override { return "WebContentServer::SetPreferredContrast"; }

    static ErrorOr<NonnullOwnPtr<SetPreferredContrast>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto contrast = TRY((decoder.decode<Web::CSS::PreferredContrast>()));

        return make<SetPreferredContrast>(move(page_id), move(contrast));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetPreferredContrast));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_contrast));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::PreferredContrast& contrast() const { return m_contrast; }
    Web::CSS::PreferredContrast take_contrast() { return move(m_contrast); }

private:

    u64 m_page_id;

    Web::CSS::PreferredContrast m_contrast;

};

class SetPreferredMotion final : public IPC::Message {
public:

    SetPreferredMotion(SetPreferredMotion const&) = default;
    SetPreferredMotion(SetPreferredMotion&&) = default;
    SetPreferredMotion& operator=(SetPreferredMotion const&) = default;
    SetPreferredMotion(u64 page_id, Web::CSS::PreferredMotion motion) : m_page_id(move(page_id)), m_motion(move(motion)) {}

    virtual ~SetPreferredMotion() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetPreferredMotion; }
    static i32 static_message_id() { return (int)MessageID::SetPreferredMotion; }
    virtual const char* message_name() const override { return "WebContentServer::SetPreferredMotion"; }

    static ErrorOr<NonnullOwnPtr<SetPreferredMotion>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto motion = TRY((decoder.decode<Web::CSS::PreferredMotion>()));

        return make<SetPreferredMotion>(move(page_id), move(motion));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetPreferredMotion));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_motion));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::CSS::PreferredMotion& motion() const { return m_motion; }
    Web::CSS::PreferredMotion take_motion() { return move(m_motion); }

private:

    u64 m_page_id;

    Web::CSS::PreferredMotion m_motion;

};

class SetPreferredLanguages final : public IPC::Message {
public:

    SetPreferredLanguages(SetPreferredLanguages const&) = default;
    SetPreferredLanguages(SetPreferredLanguages&&) = default;
    SetPreferredLanguages& operator=(SetPreferredLanguages const&) = default;
    SetPreferredLanguages(u64 page_id, Vector<String> preferred_languages) : m_page_id(move(page_id)), m_preferred_languages(move(preferred_languages)) {}

    virtual ~SetPreferredLanguages() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetPreferredLanguages; }
    static i32 static_message_id() { return (int)MessageID::SetPreferredLanguages; }
    virtual const char* message_name() const override { return "WebContentServer::SetPreferredLanguages"; }

    static ErrorOr<NonnullOwnPtr<SetPreferredLanguages>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto preferred_languages = TRY((decoder.decode<Vector<String>>()));

        return make<SetPreferredLanguages>(move(page_id), move(preferred_languages));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetPreferredLanguages));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_preferred_languages));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<String>& preferred_languages() const { return m_preferred_languages; }
    Vector<String> take_preferred_languages() { return move(m_preferred_languages); }

private:

    u64 m_page_id;

    Vector<String> m_preferred_languages;

};

class SetEnableDoNotTrack final : public IPC::Message {
public:

    SetEnableDoNotTrack(SetEnableDoNotTrack const&) = default;
    SetEnableDoNotTrack(SetEnableDoNotTrack&&) = default;
    SetEnableDoNotTrack& operator=(SetEnableDoNotTrack const&) = default;
    SetEnableDoNotTrack(u64 page_id, bool enable) : m_page_id(move(page_id)), m_enable(move(enable)) {}

    virtual ~SetEnableDoNotTrack() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetEnableDoNotTrack; }
    static i32 static_message_id() { return (int)MessageID::SetEnableDoNotTrack; }
    virtual const char* message_name() const override { return "WebContentServer::SetEnableDoNotTrack"; }

    static ErrorOr<NonnullOwnPtr<SetEnableDoNotTrack>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto enable = TRY((decoder.decode<bool>()));

        return make<SetEnableDoNotTrack>(move(page_id), move(enable));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetEnableDoNotTrack));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_enable));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& enable() const { return m_enable; }
    bool take_enable() { return move(m_enable); }

private:

    u64 m_page_id;

    bool m_enable;

};

class SetHasFocus final : public IPC::Message {
public:

    SetHasFocus(SetHasFocus const&) = default;
    SetHasFocus(SetHasFocus&&) = default;
    SetHasFocus& operator=(SetHasFocus const&) = default;
    SetHasFocus(u64 page_id, bool has_focus) : m_page_id(move(page_id)), m_has_focus(move(has_focus)) {}

    virtual ~SetHasFocus() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetHasFocus; }
    static i32 static_message_id() { return (int)MessageID::SetHasFocus; }
    virtual const char* message_name() const override { return "WebContentServer::SetHasFocus"; }

    static ErrorOr<NonnullOwnPtr<SetHasFocus>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto has_focus = TRY((decoder.decode<bool>()));

        return make<SetHasFocus>(move(page_id), move(has_focus));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetHasFocus));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_has_focus));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& has_focus() const { return m_has_focus; }
    bool take_has_focus() { return move(m_has_focus); }

private:

    u64 m_page_id;

    bool m_has_focus;

};

class SetIsScriptingEnabled final : public IPC::Message {
public:

    SetIsScriptingEnabled(SetIsScriptingEnabled const&) = default;
    SetIsScriptingEnabled(SetIsScriptingEnabled&&) = default;
    SetIsScriptingEnabled& operator=(SetIsScriptingEnabled const&) = default;
    SetIsScriptingEnabled(u64 page_id, bool is_scripting_enabled) : m_page_id(move(page_id)), m_is_scripting_enabled(move(is_scripting_enabled)) {}

    virtual ~SetIsScriptingEnabled() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetIsScriptingEnabled; }
    static i32 static_message_id() { return (int)MessageID::SetIsScriptingEnabled; }
    virtual const char* message_name() const override { return "WebContentServer::SetIsScriptingEnabled"; }

    static ErrorOr<NonnullOwnPtr<SetIsScriptingEnabled>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto is_scripting_enabled = TRY((decoder.decode<bool>()));

        return make<SetIsScriptingEnabled>(move(page_id), move(is_scripting_enabled));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetIsScriptingEnabled));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_is_scripting_enabled));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& is_scripting_enabled() const { return m_is_scripting_enabled; }
    bool take_is_scripting_enabled() { return move(m_is_scripting_enabled); }

private:

    u64 m_page_id;

    bool m_is_scripting_enabled;

};

class SetDevicePixelsPerCssPixel final : public IPC::Message {
public:

    SetDevicePixelsPerCssPixel(SetDevicePixelsPerCssPixel const&) = default;
    SetDevicePixelsPerCssPixel(SetDevicePixelsPerCssPixel&&) = default;
    SetDevicePixelsPerCssPixel& operator=(SetDevicePixelsPerCssPixel const&) = default;
    SetDevicePixelsPerCssPixel(u64 page_id, float device_pixels_per_css_pixel) : m_page_id(move(page_id)), m_device_pixels_per_css_pixel(move(device_pixels_per_css_pixel)) {}

    virtual ~SetDevicePixelsPerCssPixel() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetDevicePixelsPerCssPixel; }
    static i32 static_message_id() { return (int)MessageID::SetDevicePixelsPerCssPixel; }
    virtual const char* message_name() const override { return "WebContentServer::SetDevicePixelsPerCssPixel"; }

    static ErrorOr<NonnullOwnPtr<SetDevicePixelsPerCssPixel>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto device_pixels_per_css_pixel = TRY((decoder.decode<float>()));

        return make<SetDevicePixelsPerCssPixel>(move(page_id), move(device_pixels_per_css_pixel));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetDevicePixelsPerCssPixel));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_device_pixels_per_css_pixel));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const float& device_pixels_per_css_pixel() const { return m_device_pixels_per_css_pixel; }
    float take_device_pixels_per_css_pixel() { return move(m_device_pixels_per_css_pixel); }

private:

    u64 m_page_id;

    float m_device_pixels_per_css_pixel;

};

class SetWindowPosition final : public IPC::Message {
public:

    SetWindowPosition(SetWindowPosition const&) = default;
    SetWindowPosition(SetWindowPosition&&) = default;
    SetWindowPosition& operator=(SetWindowPosition const&) = default;
    SetWindowPosition(u64 page_id, Web::DevicePixelPoint position) : m_page_id(move(page_id)), m_position(move(position)) {}

    virtual ~SetWindowPosition() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetWindowPosition; }
    static i32 static_message_id() { return (int)MessageID::SetWindowPosition; }
    virtual const char* message_name() const override { return "WebContentServer::SetWindowPosition"; }

    static ErrorOr<NonnullOwnPtr<SetWindowPosition>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto position = TRY((decoder.decode<Web::DevicePixelPoint>()));

        return make<SetWindowPosition>(move(page_id), move(position));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetWindowPosition));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_position));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::DevicePixelPoint& position() const { return m_position; }
    Web::DevicePixelPoint take_position() { return move(m_position); }

private:

    u64 m_page_id;

    Web::DevicePixelPoint m_position;

};

class SetWindowSize final : public IPC::Message {
public:

    SetWindowSize(SetWindowSize const&) = default;
    SetWindowSize(SetWindowSize&&) = default;
    SetWindowSize& operator=(SetWindowSize const&) = default;
    SetWindowSize(u64 page_id, Web::DevicePixelSize size) : m_page_id(move(page_id)), m_size(move(size)) {}

    virtual ~SetWindowSize() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetWindowSize; }
    static i32 static_message_id() { return (int)MessageID::SetWindowSize; }
    virtual const char* message_name() const override { return "WebContentServer::SetWindowSize"; }

    static ErrorOr<NonnullOwnPtr<SetWindowSize>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto size = TRY((decoder.decode<Web::DevicePixelSize>()));

        return make<SetWindowSize>(move(page_id), move(size));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetWindowSize));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_size));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::DevicePixelSize& size() const { return m_size; }
    Web::DevicePixelSize take_size() { return move(m_size); }

private:

    u64 m_page_id;

    Web::DevicePixelSize m_size;

};

class DidUpdateWindowRect final : public IPC::Message {
public:

    DidUpdateWindowRect(DidUpdateWindowRect const&) = default;
    DidUpdateWindowRect(DidUpdateWindowRect&&) = default;
    DidUpdateWindowRect& operator=(DidUpdateWindowRect const&) = default;
    DidUpdateWindowRect(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    DidUpdateWindowRect(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~DidUpdateWindowRect() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::DidUpdateWindowRect; }
    static i32 static_message_id() { return (int)MessageID::DidUpdateWindowRect; }
    virtual const char* message_name() const override { return "WebContentServer::DidUpdateWindowRect"; }

    static ErrorOr<NonnullOwnPtr<DidUpdateWindowRect>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<DidUpdateWindowRect>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::DidUpdateWindowRect));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class GetLocalStorageEntriesResponse final : public IPC::Message {
public:

    GetLocalStorageEntriesResponse(GetLocalStorageEntriesResponse const&) = default;
    GetLocalStorageEntriesResponse(GetLocalStorageEntriesResponse&&) = default;
    GetLocalStorageEntriesResponse& operator=(GetLocalStorageEntriesResponse const&) = default;
    GetLocalStorageEntriesResponse(OrderedHashMap<String, String> entries) : m_entries(move(entries)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, OrderedHashMap<String, String>>)
    GetLocalStorageEntriesResponse(WrappedReturnType&& value)
        : m_entries(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetLocalStorageEntriesResponse() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetLocalStorageEntriesResponse; }
    static i32 static_message_id() { return (int)MessageID::GetLocalStorageEntriesResponse; }
    virtual const char* message_name() const override { return "WebContentServer::GetLocalStorageEntriesResponse"; }

    static ErrorOr<NonnullOwnPtr<GetLocalStorageEntriesResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto entries = TRY((decoder.decode<OrderedHashMap<String, String>>()));

        return make<GetLocalStorageEntriesResponse>(move(entries));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetLocalStorageEntriesResponse));

        TRY(stream.encode(m_entries));

        return buffer;
    }

    const OrderedHashMap<String, String>& entries() const { return m_entries; }
    OrderedHashMap<String, String> take_entries() { return move(m_entries); }

private:

    OrderedHashMap<String, String> m_entries;

};

class GetLocalStorageEntries final : public IPC::Message {
public:

   typedef class GetLocalStorageEntriesResponse ResponseType;

    GetLocalStorageEntries(GetLocalStorageEntries const&) = default;
    GetLocalStorageEntries(GetLocalStorageEntries&&) = default;
    GetLocalStorageEntries& operator=(GetLocalStorageEntries const&) = default;
    GetLocalStorageEntries(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetLocalStorageEntries(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetLocalStorageEntries() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetLocalStorageEntries; }
    static i32 static_message_id() { return (int)MessageID::GetLocalStorageEntries; }
    virtual const char* message_name() const override { return "WebContentServer::GetLocalStorageEntries"; }

    static ErrorOr<NonnullOwnPtr<GetLocalStorageEntries>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetLocalStorageEntries>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetLocalStorageEntries));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class GetSessionStorageEntriesResponse final : public IPC::Message {
public:

    GetSessionStorageEntriesResponse(GetSessionStorageEntriesResponse const&) = default;
    GetSessionStorageEntriesResponse(GetSessionStorageEntriesResponse&&) = default;
    GetSessionStorageEntriesResponse& operator=(GetSessionStorageEntriesResponse const&) = default;
    GetSessionStorageEntriesResponse(OrderedHashMap<String, String> entries) : m_entries(move(entries)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, OrderedHashMap<String, String>>)
    GetSessionStorageEntriesResponse(WrappedReturnType&& value)
        : m_entries(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetSessionStorageEntriesResponse() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetSessionStorageEntriesResponse; }
    static i32 static_message_id() { return (int)MessageID::GetSessionStorageEntriesResponse; }
    virtual const char* message_name() const override { return "WebContentServer::GetSessionStorageEntriesResponse"; }

    static ErrorOr<NonnullOwnPtr<GetSessionStorageEntriesResponse>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto entries = TRY((decoder.decode<OrderedHashMap<String, String>>()));

        return make<GetSessionStorageEntriesResponse>(move(entries));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetSessionStorageEntriesResponse));

        TRY(stream.encode(m_entries));

        return buffer;
    }

    const OrderedHashMap<String, String>& entries() const { return m_entries; }
    OrderedHashMap<String, String> take_entries() { return move(m_entries); }

private:

    OrderedHashMap<String, String> m_entries;

};

class GetSessionStorageEntries final : public IPC::Message {
public:

   typedef class GetSessionStorageEntriesResponse ResponseType;

    GetSessionStorageEntries(GetSessionStorageEntries const&) = default;
    GetSessionStorageEntries(GetSessionStorageEntries&&) = default;
    GetSessionStorageEntries& operator=(GetSessionStorageEntries const&) = default;
    GetSessionStorageEntries(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    GetSessionStorageEntries(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~GetSessionStorageEntries() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::GetSessionStorageEntries; }
    static i32 static_message_id() { return (int)MessageID::GetSessionStorageEntries; }
    virtual const char* message_name() const override { return "WebContentServer::GetSessionStorageEntries"; }

    static ErrorOr<NonnullOwnPtr<GetSessionStorageEntries>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<GetSessionStorageEntries>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::GetSessionStorageEntries));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class HandleFileReturn final : public IPC::Message {
public:

    HandleFileReturn(HandleFileReturn const&) = default;
    HandleFileReturn(HandleFileReturn&&) = default;
    HandleFileReturn& operator=(HandleFileReturn const&) = default;
    HandleFileReturn(u64 page_id, i32 error, Optional<IPC::File> file, i32 request_id) : m_page_id(move(page_id)), m_error(move(error)), m_file(move(file)), m_request_id(move(request_id)) {}

    virtual ~HandleFileReturn() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::HandleFileReturn; }
    static i32 static_message_id() { return (int)MessageID::HandleFileReturn; }
    virtual const char* message_name() const override { return "WebContentServer::HandleFileReturn"; }

    static ErrorOr<NonnullOwnPtr<HandleFileReturn>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto error = TRY((decoder.decode<i32>()));

        auto file = TRY((decoder.decode<Optional<IPC::File>>()));

        auto request_id = TRY((decoder.decode<i32>()));

        return make<HandleFileReturn>(move(page_id), move(error), move(file), move(request_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::HandleFileReturn));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_error));

        TRY(stream.encode(m_file));

        TRY(stream.encode(m_request_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const i32& error() const { return m_error; }
    i32 take_error() { return move(m_error); }

    const Optional<IPC::File>& file() const { return m_file; }
    Optional<IPC::File> take_file() { return move(m_file); }

    const i32& request_id() const { return m_request_id; }
    i32 take_request_id() { return move(m_request_id); }

private:

    u64 m_page_id;

    i32 m_error;

    Optional<IPC::File> m_file;

    i32 m_request_id;

};

class SetSystemVisibilityState final : public IPC::Message {
public:

    SetSystemVisibilityState(SetSystemVisibilityState const&) = default;
    SetSystemVisibilityState(SetSystemVisibilityState&&) = default;
    SetSystemVisibilityState& operator=(SetSystemVisibilityState const&) = default;
    SetSystemVisibilityState(u64 page_id, Web::HTML::VisibilityState visibility_state) : m_page_id(move(page_id)), m_visibility_state(move(visibility_state)) {}

    virtual ~SetSystemVisibilityState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetSystemVisibilityState; }
    static i32 static_message_id() { return (int)MessageID::SetSystemVisibilityState; }
    virtual const char* message_name() const override { return "WebContentServer::SetSystemVisibilityState"; }

    static ErrorOr<NonnullOwnPtr<SetSystemVisibilityState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto visibility_state = TRY((decoder.decode<Web::HTML::VisibilityState>()));

        return make<SetSystemVisibilityState>(move(page_id), move(visibility_state));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetSystemVisibilityState));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_visibility_state));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Web::HTML::VisibilityState& visibility_state() const { return m_visibility_state; }
    Web::HTML::VisibilityState take_visibility_state() { return move(m_visibility_state); }

private:

    u64 m_page_id;

    Web::HTML::VisibilityState m_visibility_state;

};

class AlertClosed final : public IPC::Message {
public:

    AlertClosed(AlertClosed const&) = default;
    AlertClosed(AlertClosed&&) = default;
    AlertClosed& operator=(AlertClosed const&) = default;
    AlertClosed(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    AlertClosed(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~AlertClosed() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::AlertClosed; }
    static i32 static_message_id() { return (int)MessageID::AlertClosed; }
    virtual const char* message_name() const override { return "WebContentServer::AlertClosed"; }

    static ErrorOr<NonnullOwnPtr<AlertClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<AlertClosed>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::AlertClosed));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class ConfirmClosed final : public IPC::Message {
public:

    ConfirmClosed(ConfirmClosed const&) = default;
    ConfirmClosed(ConfirmClosed&&) = default;
    ConfirmClosed& operator=(ConfirmClosed const&) = default;
    ConfirmClosed(u64 page_id, bool accepted) : m_page_id(move(page_id)), m_accepted(move(accepted)) {}

    virtual ~ConfirmClosed() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ConfirmClosed; }
    static i32 static_message_id() { return (int)MessageID::ConfirmClosed; }
    virtual const char* message_name() const override { return "WebContentServer::ConfirmClosed"; }

    static ErrorOr<NonnullOwnPtr<ConfirmClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto accepted = TRY((decoder.decode<bool>()));

        return make<ConfirmClosed>(move(page_id), move(accepted));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ConfirmClosed));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_accepted));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const bool& accepted() const { return m_accepted; }
    bool take_accepted() { return move(m_accepted); }

private:

    u64 m_page_id;

    bool m_accepted;

};

class PromptClosed final : public IPC::Message {
public:

    PromptClosed(PromptClosed const&) = default;
    PromptClosed(PromptClosed&&) = default;
    PromptClosed& operator=(PromptClosed const&) = default;
    PromptClosed(u64 page_id, Optional<String> response) : m_page_id(move(page_id)), m_response(move(response)) {}

    virtual ~PromptClosed() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::PromptClosed; }
    static i32 static_message_id() { return (int)MessageID::PromptClosed; }
    virtual const char* message_name() const override { return "WebContentServer::PromptClosed"; }

    static ErrorOr<NonnullOwnPtr<PromptClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto response = TRY((decoder.decode<Optional<String>>()));

        return make<PromptClosed>(move(page_id), move(response));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::PromptClosed));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_response));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Optional<String>& response() const { return m_response; }
    Optional<String> take_response() { return move(m_response); }

private:

    u64 m_page_id;

    Optional<String> m_response;

};

class ColorPickerUpdate final : public IPC::Message {
public:

    ColorPickerUpdate(ColorPickerUpdate const&) = default;
    ColorPickerUpdate(ColorPickerUpdate&&) = default;
    ColorPickerUpdate& operator=(ColorPickerUpdate const&) = default;
    ColorPickerUpdate(u64 page_id, Optional<Color> picked_color, Web::HTML::ColorPickerUpdateState state) : m_page_id(move(page_id)), m_picked_color(move(picked_color)), m_state(move(state)) {}

    virtual ~ColorPickerUpdate() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ColorPickerUpdate; }
    static i32 static_message_id() { return (int)MessageID::ColorPickerUpdate; }
    virtual const char* message_name() const override { return "WebContentServer::ColorPickerUpdate"; }

    static ErrorOr<NonnullOwnPtr<ColorPickerUpdate>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto picked_color = TRY((decoder.decode<Optional<Color>>()));

        auto state = TRY((decoder.decode<Web::HTML::ColorPickerUpdateState>()));

        return make<ColorPickerUpdate>(move(page_id), move(picked_color), move(state));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ColorPickerUpdate));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_picked_color));

        TRY(stream.encode(m_state));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Optional<Color>& picked_color() const { return m_picked_color; }
    Optional<Color> take_picked_color() { return move(m_picked_color); }

    const Web::HTML::ColorPickerUpdateState& state() const { return m_state; }
    Web::HTML::ColorPickerUpdateState take_state() { return move(m_state); }

private:

    u64 m_page_id;

    Optional<Color> m_picked_color;

    Web::HTML::ColorPickerUpdateState m_state;

};

class FilePickerClosed final : public IPC::Message {
public:

    FilePickerClosed(FilePickerClosed const&) = default;
    FilePickerClosed(FilePickerClosed&&) = default;
    FilePickerClosed& operator=(FilePickerClosed const&) = default;
    FilePickerClosed(u64 page_id, Vector<Web::HTML::SelectedFile> selected_files) : m_page_id(move(page_id)), m_selected_files(move(selected_files)) {}

    virtual ~FilePickerClosed() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::FilePickerClosed; }
    static i32 static_message_id() { return (int)MessageID::FilePickerClosed; }
    virtual const char* message_name() const override { return "WebContentServer::FilePickerClosed"; }

    static ErrorOr<NonnullOwnPtr<FilePickerClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto selected_files = TRY((decoder.decode<Vector<Web::HTML::SelectedFile>>()));

        return make<FilePickerClosed>(move(page_id), move(selected_files));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::FilePickerClosed));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_selected_files));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Vector<Web::HTML::SelectedFile>& selected_files() const { return m_selected_files; }
    Vector<Web::HTML::SelectedFile> take_selected_files() { return move(m_selected_files); }

private:

    u64 m_page_id;

    Vector<Web::HTML::SelectedFile> m_selected_files;

};

class SelectDropdownClosed final : public IPC::Message {
public:

    SelectDropdownClosed(SelectDropdownClosed const&) = default;
    SelectDropdownClosed(SelectDropdownClosed&&) = default;
    SelectDropdownClosed& operator=(SelectDropdownClosed const&) = default;
    SelectDropdownClosed(u64 page_id, Optional<u32> selected_item_id) : m_page_id(move(page_id)), m_selected_item_id(move(selected_item_id)) {}

    virtual ~SelectDropdownClosed() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SelectDropdownClosed; }
    static i32 static_message_id() { return (int)MessageID::SelectDropdownClosed; }
    virtual const char* message_name() const override { return "WebContentServer::SelectDropdownClosed"; }

    static ErrorOr<NonnullOwnPtr<SelectDropdownClosed>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto selected_item_id = TRY((decoder.decode<Optional<u32>>()));

        return make<SelectDropdownClosed>(move(page_id), move(selected_item_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SelectDropdownClosed));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_selected_item_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const Optional<u32>& selected_item_id() const { return m_selected_item_id; }
    Optional<u32> take_selected_item_id() { return move(m_selected_item_id); }

private:

    u64 m_page_id;

    Optional<u32> m_selected_item_id;

};

class ToggleMediaPlayState final : public IPC::Message {
public:

    ToggleMediaPlayState(ToggleMediaPlayState const&) = default;
    ToggleMediaPlayState(ToggleMediaPlayState&&) = default;
    ToggleMediaPlayState& operator=(ToggleMediaPlayState const&) = default;
    ToggleMediaPlayState(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ToggleMediaPlayState(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ToggleMediaPlayState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ToggleMediaPlayState; }
    static i32 static_message_id() { return (int)MessageID::ToggleMediaPlayState; }
    virtual const char* message_name() const override { return "WebContentServer::ToggleMediaPlayState"; }

    static ErrorOr<NonnullOwnPtr<ToggleMediaPlayState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ToggleMediaPlayState>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ToggleMediaPlayState));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class ToggleMediaMuteState final : public IPC::Message {
public:

    ToggleMediaMuteState(ToggleMediaMuteState const&) = default;
    ToggleMediaMuteState(ToggleMediaMuteState&&) = default;
    ToggleMediaMuteState& operator=(ToggleMediaMuteState const&) = default;
    ToggleMediaMuteState(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ToggleMediaMuteState(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ToggleMediaMuteState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ToggleMediaMuteState; }
    static i32 static_message_id() { return (int)MessageID::ToggleMediaMuteState; }
    virtual const char* message_name() const override { return "WebContentServer::ToggleMediaMuteState"; }

    static ErrorOr<NonnullOwnPtr<ToggleMediaMuteState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ToggleMediaMuteState>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ToggleMediaMuteState));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class ToggleMediaLoopState final : public IPC::Message {
public:

    ToggleMediaLoopState(ToggleMediaLoopState const&) = default;
    ToggleMediaLoopState(ToggleMediaLoopState&&) = default;
    ToggleMediaLoopState& operator=(ToggleMediaLoopState const&) = default;
    ToggleMediaLoopState(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ToggleMediaLoopState(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ToggleMediaLoopState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ToggleMediaLoopState; }
    static i32 static_message_id() { return (int)MessageID::ToggleMediaLoopState; }
    virtual const char* message_name() const override { return "WebContentServer::ToggleMediaLoopState"; }

    static ErrorOr<NonnullOwnPtr<ToggleMediaLoopState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ToggleMediaLoopState>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ToggleMediaLoopState));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class ToggleMediaControlsState final : public IPC::Message {
public:

    ToggleMediaControlsState(ToggleMediaControlsState const&) = default;
    ToggleMediaControlsState(ToggleMediaControlsState&&) = default;
    ToggleMediaControlsState& operator=(ToggleMediaControlsState const&) = default;
    ToggleMediaControlsState(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    ToggleMediaControlsState(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~ToggleMediaControlsState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::ToggleMediaControlsState; }
    static i32 static_message_id() { return (int)MessageID::ToggleMediaControlsState; }
    virtual const char* message_name() const override { return "WebContentServer::ToggleMediaControlsState"; }

    static ErrorOr<NonnullOwnPtr<ToggleMediaControlsState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<ToggleMediaControlsState>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::ToggleMediaControlsState));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class TogglePageMuteState final : public IPC::Message {
public:

    TogglePageMuteState(TogglePageMuteState const&) = default;
    TogglePageMuteState(TogglePageMuteState&&) = default;
    TogglePageMuteState& operator=(TogglePageMuteState const&) = default;
    TogglePageMuteState(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    TogglePageMuteState(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~TogglePageMuteState() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::TogglePageMuteState; }
    static i32 static_message_id() { return (int)MessageID::TogglePageMuteState; }
    virtual const char* message_name() const override { return "WebContentServer::TogglePageMuteState"; }

    static ErrorOr<NonnullOwnPtr<TogglePageMuteState>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<TogglePageMuteState>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::TogglePageMuteState));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SetUserStyle final : public IPC::Message {
public:

    SetUserStyle(SetUserStyle const&) = default;
    SetUserStyle(SetUserStyle&&) = default;
    SetUserStyle& operator=(SetUserStyle const&) = default;
    SetUserStyle(u64 page_id, String source) : m_page_id(move(page_id)), m_source(move(source)) {}

    virtual ~SetUserStyle() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SetUserStyle; }
    static i32 static_message_id() { return (int)MessageID::SetUserStyle; }
    virtual const char* message_name() const override { return "WebContentServer::SetUserStyle"; }

    static ErrorOr<NonnullOwnPtr<SetUserStyle>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        auto source = TRY((decoder.decode<String>()));

        return make<SetUserStyle>(move(page_id), move(source));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SetUserStyle));

        TRY(stream.encode(m_page_id));

        TRY(stream.encode(m_source));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

    const String& source() const { return m_source; }
    String take_source() { return move(m_source); }

private:

    u64 m_page_id;

    String m_source;

};

class EnableInspectorPrototype final : public IPC::Message {
public:

    EnableInspectorPrototype(EnableInspectorPrototype const&) = default;
    EnableInspectorPrototype(EnableInspectorPrototype&&) = default;
    EnableInspectorPrototype& operator=(EnableInspectorPrototype const&) = default;
    EnableInspectorPrototype(u64 page_id) : m_page_id(move(page_id)) {}

    template <typename WrappedReturnType>
    requires(!SameAs<WrappedReturnType, u64>)
    EnableInspectorPrototype(WrappedReturnType&& value)
        : m_page_id(forward<WrappedReturnType>(value))
    {
    }

    virtual ~EnableInspectorPrototype() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::EnableInspectorPrototype; }
    static i32 static_message_id() { return (int)MessageID::EnableInspectorPrototype; }
    virtual const char* message_name() const override { return "WebContentServer::EnableInspectorPrototype"; }

    static ErrorOr<NonnullOwnPtr<EnableInspectorPrototype>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        auto page_id = TRY((decoder.decode<u64>()));

        return make<EnableInspectorPrototype>(move(page_id));
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::EnableInspectorPrototype));

        TRY(stream.encode(m_page_id));

        return buffer;
    }

    const u64& page_id() const { return m_page_id; }
    u64 take_page_id() { return move(m_page_id); }

private:

    u64 m_page_id;

};

class SystemTimeZoneChanged final : public IPC::Message {
public:

    SystemTimeZoneChanged(SystemTimeZoneChanged const&) = default;
    SystemTimeZoneChanged(SystemTimeZoneChanged&&) = default;
    SystemTimeZoneChanged& operator=(SystemTimeZoneChanged const&) = default;
    SystemTimeZoneChanged() {}

    virtual ~SystemTimeZoneChanged() override {}

    virtual u32 endpoint_magic() const override { return 4289017466; }
    virtual i32 message_id() const override { return (int)MessageID::SystemTimeZoneChanged; }
    static i32 static_message_id() { return (int)MessageID::SystemTimeZoneChanged; }
    virtual const char* message_name() const override { return "WebContentServer::SystemTimeZoneChanged"; }

    static ErrorOr<NonnullOwnPtr<SystemTimeZoneChanged>> decode(Stream& stream, Queue<IPC::File>& files)
    {
        IPC::Decoder decoder { stream, files };

        return make<SystemTimeZoneChanged>();
    }

    virtual ErrorOr<IPC::MessageBuffer> encode() const override
    {
        IPC::MessageBuffer buffer;
        IPC::Encoder stream(buffer);
        TRY(stream.encode(endpoint_magic()));
        TRY(stream.encode((int)MessageID::SystemTimeZoneChanged));

        return buffer;
    }

private:

};

} // namespace Messages::WebContentServer

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentServerProxy {
public:
    // Used to disambiguate the constructor call.
    struct Tag { };

    WebContentServerProxy(IPC::Connection<LocalEndpoint, PeerEndpoint>& connection, Tag)
        : m_connection(connection)
    { }

    int init_transport(
int peer_pid) {
        return m_connection.template send_sync<Messages::WebContentServer::InitTransport>(peer_pid)->take_peer_pid();
    }

    void async_init_transport(
int peer_pid) {
        MUST(m_connection.post_message(Messages::WebContentServer::InitTransport { peer_pid }));

    }

    IPC::IPCErrorOr<int> try_init_transport(
int peer_pid) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentServer::InitTransport>(peer_pid);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_close_server(
) {
        MUST(m_connection.post_message(Messages::WebContentServer::CloseServer {  }));

    }

    String get_window_handle(
u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentServer::GetWindowHandle>(page_id)->take_handle();
    }

    void async_get_window_handle(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetWindowHandle { page_id }));

    }

    IPC::IPCErrorOr<String> try_get_window_handle(
u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentServer::GetWindowHandle>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_set_window_handle(
u64 page_id, String handle) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetWindowHandle { page_id, move(handle) }));

    }

    void async_connect_to_webdriver(
u64 page_id, ByteString webdriver_ipc_path) {
        MUST(m_connection.post_message(Messages::WebContentServer::ConnectToWebdriver { page_id, move(webdriver_ipc_path) }));

    }

    void async_connect_to_image_decoder(
IPC::File socket_fd) {
        MUST(m_connection.post_message(Messages::WebContentServer::ConnectToImageDecoder { move(socket_fd) }));

    }

    void async_update_system_theme(
u64 page_id, Core::AnonymousBuffer theme_buffer) {
        MUST(m_connection.post_message(Messages::WebContentServer::UpdateSystemTheme { page_id, move(theme_buffer) }));

    }

    void async_update_screen_rects(
u64 page_id, Vector<Web::DevicePixelRect> rects, u32 main_screen_index) {
        MUST(m_connection.post_message(Messages::WebContentServer::UpdateScreenRects { page_id, move(rects), main_screen_index }));

    }

    void async_load_url(
u64 page_id, URL::URL url) {
        MUST(m_connection.post_message(Messages::WebContentServer::LoadUrl { page_id, move(url) }));

    }

    void async_load_html(
u64 page_id, ByteString html) {
        MUST(m_connection.post_message(Messages::WebContentServer::LoadHtml { page_id, move(html) }));

    }

    void async_reload(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::Reload { page_id }));

    }

    void async_traverse_the_history_by_delta(
u64 page_id, i32 delta) {
        MUST(m_connection.post_message(Messages::WebContentServer::TraverseTheHistoryByDelta { page_id, delta }));

    }

    void async_ready_to_paint(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ReadyToPaint { page_id }));

    }

    void async_set_viewport_size(
u64 page_id, Web::DevicePixelSize size) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetViewportSize { page_id, size }));

    }

    void async_key_event(
u64 page_id, Web::KeyEvent event) {
        MUST(m_connection.post_message(Messages::WebContentServer::KeyEvent { page_id, move(event) }));

    }

    void async_mouse_event(
u64 page_id, Web::MouseEvent event) {
        MUST(m_connection.post_message(Messages::WebContentServer::MouseEvent { page_id, move(event) }));

    }

    void async_drag_event(
u64 page_id, Web::DragEvent event) {
        MUST(m_connection.post_message(Messages::WebContentServer::DragEvent { page_id, move(event) }));

    }

    void async_debug_request(
u64 page_id, ByteString request, ByteString argument) {
        MUST(m_connection.post_message(Messages::WebContentServer::DebugRequest { page_id, move(request), move(argument) }));

    }

    void async_get_source(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetSource { page_id }));

    }

    void async_inspect_dom_tree(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::InspectDomTree { page_id }));

    }

    void async_inspect_dom_node(
u64 page_id, Web::UniqueNodeID node_id, Optional<Web::CSS::Selector::PseudoElement::Type> pseudo_element) {
        MUST(m_connection.post_message(Messages::WebContentServer::InspectDomNode { page_id, move(node_id), move(pseudo_element) }));

    }

    void async_inspect_accessibility_tree(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::InspectAccessibilityTree { page_id }));

    }

    void async_get_hovered_node_id(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetHoveredNodeId { page_id }));

    }

    void async_js_console_input(
u64 page_id, ByteString js_source) {
        MUST(m_connection.post_message(Messages::WebContentServer::JsConsoleInput { page_id, move(js_source) }));

    }

    void async_js_console_request_messages(
u64 page_id, i32 start_index) {
        MUST(m_connection.post_message(Messages::WebContentServer::JsConsoleRequestMessages { page_id, start_index }));

    }

    void async_list_style_sheets(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ListStyleSheets { page_id }));

    }

    void async_request_style_sheet_source(
u64 page_id, Web::CSS::StyleSheetIdentifier identifier) {
        MUST(m_connection.post_message(Messages::WebContentServer::RequestStyleSheetSource { page_id, move(identifier) }));

    }

    void async_set_dom_node_text(
u64 page_id, Web::UniqueNodeID node_id, String text) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetDomNodeText { page_id, move(node_id), move(text) }));

    }

    void async_set_dom_node_tag(
u64 page_id, Web::UniqueNodeID node_id, String name) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetDomNodeTag { page_id, move(node_id), move(name) }));

    }

    void async_add_dom_node_attributes(
u64 page_id, Web::UniqueNodeID node_id, Vector<WebView::Attribute> attributes) {
        MUST(m_connection.post_message(Messages::WebContentServer::AddDomNodeAttributes { page_id, move(node_id), move(attributes) }));

    }

    void async_replace_dom_node_attribute(
u64 page_id, Web::UniqueNodeID node_id, String name, Vector<WebView::Attribute> replacement_attributes) {
        MUST(m_connection.post_message(Messages::WebContentServer::ReplaceDomNodeAttribute { page_id, move(node_id), move(name), move(replacement_attributes) }));

    }

    void async_create_child_element(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::CreateChildElement { page_id, move(node_id) }));

    }

    void async_create_child_text_node(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::CreateChildTextNode { page_id, move(node_id) }));

    }

    void async_clone_dom_node(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::CloneDomNode { page_id, move(node_id) }));

    }

    void async_remove_dom_node(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::RemoveDomNode { page_id, move(node_id) }));

    }

    void async_get_dom_node_html(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetDomNodeHtml { page_id, move(node_id) }));

    }

    void async_take_document_screenshot(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::TakeDocumentScreenshot { page_id }));

    }

    void async_take_dom_node_screenshot(
u64 page_id, Web::UniqueNodeID node_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::TakeDomNodeScreenshot { page_id, move(node_id) }));

    }

    void async_request_internal_page_info(
u64 page_id, WebView::PageInfoType type) {
        MUST(m_connection.post_message(Messages::WebContentServer::RequestInternalPageInfo { page_id, type }));

    }

    void async_run_javascript(
u64 page_id, ByteString js_source) {
        MUST(m_connection.post_message(Messages::WebContentServer::RunJavascript { page_id, move(js_source) }));

    }

    ByteString get_selected_text(
u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentServer::GetSelectedText>(page_id)->take_selection();
    }

    void async_get_selected_text(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetSelectedText { page_id }));

    }

    IPC::IPCErrorOr<ByteString> try_get_selected_text(
u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentServer::GetSelectedText>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_select_all(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::SelectAll { page_id }));

    }

    void async_paste(
u64 page_id, String text) {
        MUST(m_connection.post_message(Messages::WebContentServer::Paste { page_id, move(text) }));

    }

    void async_find_in_page(
u64 page_id, String query, AK::CaseSensitivity case_sensitivity) {
        MUST(m_connection.post_message(Messages::WebContentServer::FindInPage { page_id, move(query), case_sensitivity }));

    }

    void async_find_in_page_next_match(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::FindInPageNextMatch { page_id }));

    }

    void async_find_in_page_previous_match(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::FindInPagePreviousMatch { page_id }));

    }

    void async_set_content_filters(
u64 page_id, Vector<String> filters) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetContentFilters { page_id, move(filters) }));

    }

    void async_set_autoplay_allowed_on_all_websites(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetAutoplayAllowedOnAllWebsites { page_id }));

    }

    void async_set_autoplay_allowlist(
u64 page_id, Vector<String> allowlist) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetAutoplayAllowlist { page_id, move(allowlist) }));

    }

    void async_set_proxy_mappings(
u64 page_id, Vector<ByteString> proxies, HashMap<ByteString, size_t> mappings) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetProxyMappings { page_id, move(proxies), move(mappings) }));

    }

    void async_set_preferred_color_scheme(
u64 page_id, Web::CSS::PreferredColorScheme color_scheme) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetPreferredColorScheme { page_id, move(color_scheme) }));

    }

    void async_set_preferred_contrast(
u64 page_id, Web::CSS::PreferredContrast contrast) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetPreferredContrast { page_id, move(contrast) }));

    }

    void async_set_preferred_motion(
u64 page_id, Web::CSS::PreferredMotion motion) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetPreferredMotion { page_id, move(motion) }));

    }

    void async_set_preferred_languages(
u64 page_id, Vector<String> preferred_languages) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetPreferredLanguages { page_id, move(preferred_languages) }));

    }

    void async_set_enable_do_not_track(
u64 page_id, bool enable) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetEnableDoNotTrack { page_id, enable }));

    }

    void async_set_has_focus(
u64 page_id, bool has_focus) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetHasFocus { page_id, has_focus }));

    }

    void async_set_is_scripting_enabled(
u64 page_id, bool is_scripting_enabled) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetIsScriptingEnabled { page_id, is_scripting_enabled }));

    }

    void async_set_device_pixels_per_css_pixel(
u64 page_id, float device_pixels_per_css_pixel) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetDevicePixelsPerCssPixel { page_id, device_pixels_per_css_pixel }));

    }

    void async_set_window_position(
u64 page_id, Web::DevicePixelPoint position) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetWindowPosition { page_id, position }));

    }

    void async_set_window_size(
u64 page_id, Web::DevicePixelSize size) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetWindowSize { page_id, size }));

    }

    void async_did_update_window_rect(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::DidUpdateWindowRect { page_id }));

    }

    OrderedHashMap<String, String> get_local_storage_entries(
u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentServer::GetLocalStorageEntries>(page_id)->take_entries();
    }

    void async_get_local_storage_entries(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetLocalStorageEntries { page_id }));

    }

    IPC::IPCErrorOr<OrderedHashMap<String, String>> try_get_local_storage_entries(
u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentServer::GetLocalStorageEntries>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    OrderedHashMap<String, String> get_session_storage_entries(
u64 page_id) {
        return m_connection.template send_sync<Messages::WebContentServer::GetSessionStorageEntries>(page_id)->take_entries();
    }

    void async_get_session_storage_entries(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::GetSessionStorageEntries { page_id }));

    }

    IPC::IPCErrorOr<OrderedHashMap<String, String>> try_get_session_storage_entries(
u64 page_id) {
        auto result = m_connection.template send_sync_but_allow_failure<Messages::WebContentServer::GetSessionStorageEntries>(page_id);
        if (!result) {
            m_connection.shutdown();
            return IPC::ErrorCode::PeerDisconnected;
        }
        return move(*result);

    }

    void async_handle_file_return(
u64 page_id, i32 error, Optional<IPC::File> file, i32 request_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::HandleFileReturn { page_id, error, move(file), request_id }));

    }

    void async_set_system_visibility_state(
u64 page_id, Web::HTML::VisibilityState visibility_state) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetSystemVisibilityState { page_id, visibility_state }));

    }

    void async_alert_closed(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::AlertClosed { page_id }));

    }

    void async_confirm_closed(
u64 page_id, bool accepted) {
        MUST(m_connection.post_message(Messages::WebContentServer::ConfirmClosed { page_id, accepted }));

    }

    void async_prompt_closed(
u64 page_id, Optional<String> response) {
        MUST(m_connection.post_message(Messages::WebContentServer::PromptClosed { page_id, move(response) }));

    }

    void async_color_picker_update(
u64 page_id, Optional<Color> picked_color, Web::HTML::ColorPickerUpdateState state) {
        MUST(m_connection.post_message(Messages::WebContentServer::ColorPickerUpdate { page_id, move(picked_color), move(state) }));

    }

    void async_file_picker_closed(
u64 page_id, Vector<Web::HTML::SelectedFile> selected_files) {
        MUST(m_connection.post_message(Messages::WebContentServer::FilePickerClosed { page_id, move(selected_files) }));

    }

    void async_select_dropdown_closed(
u64 page_id, Optional<u32> selected_item_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::SelectDropdownClosed { page_id, move(selected_item_id) }));

    }

    void async_toggle_media_play_state(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ToggleMediaPlayState { page_id }));

    }

    void async_toggle_media_mute_state(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ToggleMediaMuteState { page_id }));

    }

    void async_toggle_media_loop_state(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ToggleMediaLoopState { page_id }));

    }

    void async_toggle_media_controls_state(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::ToggleMediaControlsState { page_id }));

    }

    void async_toggle_page_mute_state(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::TogglePageMuteState { page_id }));

    }

    void async_set_user_style(
u64 page_id, String source) {
        MUST(m_connection.post_message(Messages::WebContentServer::SetUserStyle { page_id, move(source) }));

    }

    void async_enable_inspector_prototype(
u64 page_id) {
        MUST(m_connection.post_message(Messages::WebContentServer::EnableInspectorPrototype { page_id }));

    }

    void async_system_time_zone_changed(
) {
        MUST(m_connection.post_message(Messages::WebContentServer::SystemTimeZoneChanged {  }));

    }

private:
    IPC::Connection<LocalEndpoint, PeerEndpoint>& m_connection;
};

template<typename LocalEndpoint, typename PeerEndpoint>
class WebContentServerProxy;
class WebContentServerStub;

class WebContentServerEndpoint {
public:
    template<typename LocalEndpoint>
    using Proxy = WebContentServerProxy<LocalEndpoint, WebContentServerEndpoint>;
    using Stub = WebContentServerStub;

    static u32 static_magic() { return 4289017466; }

    static ErrorOr<NonnullOwnPtr<IPC::Message>> decode_message(ReadonlyBytes buffer, [[maybe_unused]] Queue<IPC::File>& files)
    {
        FixedMemoryStream stream { buffer };
        auto message_endpoint_magic = TRY(stream.read_value<u32>());

        if (message_endpoint_magic != 4289017466) {
            return Error::from_string_literal("Endpoint magic number mismatch, not my message!");
        }

        auto message_id = TRY(stream.read_value<i32>());


        switch (message_id) {

        case (int)Messages::WebContentServer::MessageID::InitTransport:
            return TRY(Messages::WebContentServer::InitTransport::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::InitTransportResponse:
            return TRY(Messages::WebContentServer::InitTransportResponse::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::CloseServer:
            return TRY(Messages::WebContentServer::CloseServer::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetWindowHandle:
            return TRY(Messages::WebContentServer::GetWindowHandle::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetWindowHandleResponse:
            return TRY(Messages::WebContentServer::GetWindowHandleResponse::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetWindowHandle:
            return TRY(Messages::WebContentServer::SetWindowHandle::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ConnectToWebdriver:
            return TRY(Messages::WebContentServer::ConnectToWebdriver::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ConnectToImageDecoder:
            return TRY(Messages::WebContentServer::ConnectToImageDecoder::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::UpdateSystemTheme:
            return TRY(Messages::WebContentServer::UpdateSystemTheme::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::UpdateScreenRects:
            return TRY(Messages::WebContentServer::UpdateScreenRects::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::LoadUrl:
            return TRY(Messages::WebContentServer::LoadUrl::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::LoadHtml:
            return TRY(Messages::WebContentServer::LoadHtml::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::Reload:
            return TRY(Messages::WebContentServer::Reload::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::TraverseTheHistoryByDelta:
            return TRY(Messages::WebContentServer::TraverseTheHistoryByDelta::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ReadyToPaint:
            return TRY(Messages::WebContentServer::ReadyToPaint::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetViewportSize:
            return TRY(Messages::WebContentServer::SetViewportSize::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::KeyEvent:
            return TRY(Messages::WebContentServer::KeyEvent::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::MouseEvent:
            return TRY(Messages::WebContentServer::MouseEvent::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::DragEvent:
            return TRY(Messages::WebContentServer::DragEvent::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::DebugRequest:
            return TRY(Messages::WebContentServer::DebugRequest::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetSource:
            return TRY(Messages::WebContentServer::GetSource::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::InspectDomTree:
            return TRY(Messages::WebContentServer::InspectDomTree::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::InspectDomNode:
            return TRY(Messages::WebContentServer::InspectDomNode::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::InspectAccessibilityTree:
            return TRY(Messages::WebContentServer::InspectAccessibilityTree::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetHoveredNodeId:
            return TRY(Messages::WebContentServer::GetHoveredNodeId::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::JsConsoleInput:
            return TRY(Messages::WebContentServer::JsConsoleInput::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::JsConsoleRequestMessages:
            return TRY(Messages::WebContentServer::JsConsoleRequestMessages::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ListStyleSheets:
            return TRY(Messages::WebContentServer::ListStyleSheets::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::RequestStyleSheetSource:
            return TRY(Messages::WebContentServer::RequestStyleSheetSource::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetDomNodeText:
            return TRY(Messages::WebContentServer::SetDomNodeText::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetDomNodeTag:
            return TRY(Messages::WebContentServer::SetDomNodeTag::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::AddDomNodeAttributes:
            return TRY(Messages::WebContentServer::AddDomNodeAttributes::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ReplaceDomNodeAttribute:
            return TRY(Messages::WebContentServer::ReplaceDomNodeAttribute::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::CreateChildElement:
            return TRY(Messages::WebContentServer::CreateChildElement::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::CreateChildTextNode:
            return TRY(Messages::WebContentServer::CreateChildTextNode::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::CloneDomNode:
            return TRY(Messages::WebContentServer::CloneDomNode::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::RemoveDomNode:
            return TRY(Messages::WebContentServer::RemoveDomNode::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetDomNodeHtml:
            return TRY(Messages::WebContentServer::GetDomNodeHtml::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::TakeDocumentScreenshot:
            return TRY(Messages::WebContentServer::TakeDocumentScreenshot::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::TakeDomNodeScreenshot:
            return TRY(Messages::WebContentServer::TakeDomNodeScreenshot::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::RequestInternalPageInfo:
            return TRY(Messages::WebContentServer::RequestInternalPageInfo::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::RunJavascript:
            return TRY(Messages::WebContentServer::RunJavascript::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetSelectedText:
            return TRY(Messages::WebContentServer::GetSelectedText::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetSelectedTextResponse:
            return TRY(Messages::WebContentServer::GetSelectedTextResponse::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SelectAll:
            return TRY(Messages::WebContentServer::SelectAll::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::Paste:
            return TRY(Messages::WebContentServer::Paste::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::FindInPage:
            return TRY(Messages::WebContentServer::FindInPage::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::FindInPageNextMatch:
            return TRY(Messages::WebContentServer::FindInPageNextMatch::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::FindInPagePreviousMatch:
            return TRY(Messages::WebContentServer::FindInPagePreviousMatch::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetContentFilters:
            return TRY(Messages::WebContentServer::SetContentFilters::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetAutoplayAllowedOnAllWebsites:
            return TRY(Messages::WebContentServer::SetAutoplayAllowedOnAllWebsites::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetAutoplayAllowlist:
            return TRY(Messages::WebContentServer::SetAutoplayAllowlist::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetProxyMappings:
            return TRY(Messages::WebContentServer::SetProxyMappings::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetPreferredColorScheme:
            return TRY(Messages::WebContentServer::SetPreferredColorScheme::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetPreferredContrast:
            return TRY(Messages::WebContentServer::SetPreferredContrast::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetPreferredMotion:
            return TRY(Messages::WebContentServer::SetPreferredMotion::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetPreferredLanguages:
            return TRY(Messages::WebContentServer::SetPreferredLanguages::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetEnableDoNotTrack:
            return TRY(Messages::WebContentServer::SetEnableDoNotTrack::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetHasFocus:
            return TRY(Messages::WebContentServer::SetHasFocus::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetIsScriptingEnabled:
            return TRY(Messages::WebContentServer::SetIsScriptingEnabled::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetDevicePixelsPerCssPixel:
            return TRY(Messages::WebContentServer::SetDevicePixelsPerCssPixel::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetWindowPosition:
            return TRY(Messages::WebContentServer::SetWindowPosition::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetWindowSize:
            return TRY(Messages::WebContentServer::SetWindowSize::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::DidUpdateWindowRect:
            return TRY(Messages::WebContentServer::DidUpdateWindowRect::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetLocalStorageEntries:
            return TRY(Messages::WebContentServer::GetLocalStorageEntries::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetLocalStorageEntriesResponse:
            return TRY(Messages::WebContentServer::GetLocalStorageEntriesResponse::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetSessionStorageEntries:
            return TRY(Messages::WebContentServer::GetSessionStorageEntries::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::GetSessionStorageEntriesResponse:
            return TRY(Messages::WebContentServer::GetSessionStorageEntriesResponse::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::HandleFileReturn:
            return TRY(Messages::WebContentServer::HandleFileReturn::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetSystemVisibilityState:
            return TRY(Messages::WebContentServer::SetSystemVisibilityState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::AlertClosed:
            return TRY(Messages::WebContentServer::AlertClosed::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ConfirmClosed:
            return TRY(Messages::WebContentServer::ConfirmClosed::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::PromptClosed:
            return TRY(Messages::WebContentServer::PromptClosed::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ColorPickerUpdate:
            return TRY(Messages::WebContentServer::ColorPickerUpdate::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::FilePickerClosed:
            return TRY(Messages::WebContentServer::FilePickerClosed::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SelectDropdownClosed:
            return TRY(Messages::WebContentServer::SelectDropdownClosed::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ToggleMediaPlayState:
            return TRY(Messages::WebContentServer::ToggleMediaPlayState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ToggleMediaMuteState:
            return TRY(Messages::WebContentServer::ToggleMediaMuteState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ToggleMediaLoopState:
            return TRY(Messages::WebContentServer::ToggleMediaLoopState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::ToggleMediaControlsState:
            return TRY(Messages::WebContentServer::ToggleMediaControlsState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::TogglePageMuteState:
            return TRY(Messages::WebContentServer::TogglePageMuteState::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SetUserStyle:
            return TRY(Messages::WebContentServer::SetUserStyle::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::EnableInspectorPrototype:
            return TRY(Messages::WebContentServer::EnableInspectorPrototype::decode(stream, files));
        case (int)Messages::WebContentServer::MessageID::SystemTimeZoneChanged:
            return TRY(Messages::WebContentServer::SystemTimeZoneChanged::decode(stream, files));
        default:
            return Error::from_string_literal("Failed to decode WebContentServer message");
        }

        VERIFY_NOT_REACHED();
    }

};

class WebContentServerStub : public IPC::Stub {
public:
    WebContentServerStub() { }
    virtual ~WebContentServerStub() override { }

    virtual u32 magic() const override { return 4289017466; }
    virtual ByteString name() const override { return "WebContentServer"; }

    virtual ErrorOr<OwnPtr<IPC::MessageBuffer>> handle(const IPC::Message& message) override
    {
        switch (message.message_id()) {

        case (int)Messages::WebContentServer::MessageID::InitTransport: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::InitTransport&>(message);
            auto response = init_transport(request.peer_pid());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentServer::MessageID::CloseServer: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::CloseServer&>(message);
            close_server();
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetWindowHandle: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetWindowHandle&>(message);
            auto response = get_window_handle(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentServer::MessageID::SetWindowHandle: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetWindowHandle&>(message);
            set_window_handle(request.page_id(), request.handle());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ConnectToWebdriver: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ConnectToWebdriver&>(message);
            connect_to_webdriver(request.page_id(), request.webdriver_ipc_path());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ConnectToImageDecoder: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ConnectToImageDecoder&>(message);
            connect_to_image_decoder(request.socket_fd());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::UpdateSystemTheme: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::UpdateSystemTheme&>(message);
            update_system_theme(request.page_id(), request.theme_buffer());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::UpdateScreenRects: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::UpdateScreenRects&>(message);
            update_screen_rects(request.page_id(), request.rects(), request.main_screen_index());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::LoadUrl: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::LoadUrl&>(message);
            load_url(request.page_id(), request.url());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::LoadHtml: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::LoadHtml&>(message);
            load_html(request.page_id(), request.html());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::Reload: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::Reload&>(message);
            reload(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::TraverseTheHistoryByDelta: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::TraverseTheHistoryByDelta&>(message);
            traverse_the_history_by_delta(request.page_id(), request.delta());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ReadyToPaint: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ReadyToPaint&>(message);
            ready_to_paint(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetViewportSize: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetViewportSize&>(message);
            set_viewport_size(request.page_id(), request.size());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::KeyEvent: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::KeyEvent&>(message);
            key_event(request.page_id(), request.event());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::MouseEvent: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::MouseEvent&>(message);
            mouse_event(request.page_id(), request.event());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::DragEvent: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::DragEvent&>(message);
            drag_event(request.page_id(), request.event());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::DebugRequest: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::DebugRequest&>(message);
            debug_request(request.page_id(), request.request(), request.argument());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetSource: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetSource&>(message);
            get_source(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::InspectDomTree: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::InspectDomTree&>(message);
            inspect_dom_tree(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::InspectDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::InspectDomNode&>(message);
            inspect_dom_node(request.page_id(), request.node_id(), request.pseudo_element());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::InspectAccessibilityTree: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::InspectAccessibilityTree&>(message);
            inspect_accessibility_tree(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetHoveredNodeId: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetHoveredNodeId&>(message);
            get_hovered_node_id(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::JsConsoleInput: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::JsConsoleInput&>(message);
            js_console_input(request.page_id(), request.js_source());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::JsConsoleRequestMessages: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::JsConsoleRequestMessages&>(message);
            js_console_request_messages(request.page_id(), request.start_index());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ListStyleSheets: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ListStyleSheets&>(message);
            list_style_sheets(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::RequestStyleSheetSource: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::RequestStyleSheetSource&>(message);
            request_style_sheet_source(request.page_id(), request.identifier());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetDomNodeText: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetDomNodeText&>(message);
            set_dom_node_text(request.page_id(), request.node_id(), request.text());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetDomNodeTag: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetDomNodeTag&>(message);
            set_dom_node_tag(request.page_id(), request.node_id(), request.name());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::AddDomNodeAttributes: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::AddDomNodeAttributes&>(message);
            add_dom_node_attributes(request.page_id(), request.node_id(), request.attributes());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ReplaceDomNodeAttribute: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ReplaceDomNodeAttribute&>(message);
            replace_dom_node_attribute(request.page_id(), request.node_id(), request.name(), request.replacement_attributes());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::CreateChildElement: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::CreateChildElement&>(message);
            create_child_element(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::CreateChildTextNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::CreateChildTextNode&>(message);
            create_child_text_node(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::CloneDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::CloneDomNode&>(message);
            clone_dom_node(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::RemoveDomNode: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::RemoveDomNode&>(message);
            remove_dom_node(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetDomNodeHtml: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetDomNodeHtml&>(message);
            get_dom_node_html(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::TakeDocumentScreenshot: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::TakeDocumentScreenshot&>(message);
            take_document_screenshot(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::TakeDomNodeScreenshot: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::TakeDomNodeScreenshot&>(message);
            take_dom_node_screenshot(request.page_id(), request.node_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::RequestInternalPageInfo: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::RequestInternalPageInfo&>(message);
            request_internal_page_info(request.page_id(), request.type());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::RunJavascript: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::RunJavascript&>(message);
            run_javascript(request.page_id(), request.js_source());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetSelectedText: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetSelectedText&>(message);
            auto response = get_selected_text(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentServer::MessageID::SelectAll: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SelectAll&>(message);
            select_all(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::Paste: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::Paste&>(message);
            paste(request.page_id(), request.text());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::FindInPage: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::FindInPage&>(message);
            find_in_page(request.page_id(), request.query(), request.case_sensitivity());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::FindInPageNextMatch: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::FindInPageNextMatch&>(message);
            find_in_page_next_match(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::FindInPagePreviousMatch: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::FindInPagePreviousMatch&>(message);
            find_in_page_previous_match(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetContentFilters: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetContentFilters&>(message);
            set_content_filters(request.page_id(), request.filters());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetAutoplayAllowedOnAllWebsites: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetAutoplayAllowedOnAllWebsites&>(message);
            set_autoplay_allowed_on_all_websites(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetAutoplayAllowlist: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetAutoplayAllowlist&>(message);
            set_autoplay_allowlist(request.page_id(), request.allowlist());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetProxyMappings: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetProxyMappings&>(message);
            set_proxy_mappings(request.page_id(), request.proxies(), request.mappings());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetPreferredColorScheme: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetPreferredColorScheme&>(message);
            set_preferred_color_scheme(request.page_id(), request.color_scheme());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetPreferredContrast: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetPreferredContrast&>(message);
            set_preferred_contrast(request.page_id(), request.contrast());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetPreferredMotion: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetPreferredMotion&>(message);
            set_preferred_motion(request.page_id(), request.motion());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetPreferredLanguages: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetPreferredLanguages&>(message);
            set_preferred_languages(request.page_id(), request.preferred_languages());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetEnableDoNotTrack: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetEnableDoNotTrack&>(message);
            set_enable_do_not_track(request.page_id(), request.enable());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetHasFocus: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetHasFocus&>(message);
            set_has_focus(request.page_id(), request.has_focus());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetIsScriptingEnabled: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetIsScriptingEnabled&>(message);
            set_is_scripting_enabled(request.page_id(), request.is_scripting_enabled());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetDevicePixelsPerCssPixel: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetDevicePixelsPerCssPixel&>(message);
            set_device_pixels_per_css_pixel(request.page_id(), request.device_pixels_per_css_pixel());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetWindowPosition: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetWindowPosition&>(message);
            set_window_position(request.page_id(), request.position());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetWindowSize: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetWindowSize&>(message);
            set_window_size(request.page_id(), request.size());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::DidUpdateWindowRect: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::DidUpdateWindowRect&>(message);
            did_update_window_rect(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::GetLocalStorageEntries: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetLocalStorageEntries&>(message);
            auto response = get_local_storage_entries(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentServer::MessageID::GetSessionStorageEntries: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::GetSessionStorageEntries&>(message);
            auto response = get_session_storage_entries(request.page_id());
            return make<IPC::MessageBuffer>(TRY(response.encode()));

        }

        case (int)Messages::WebContentServer::MessageID::HandleFileReturn: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::HandleFileReturn&>(message);
            handle_file_return(request.page_id(), request.error(), request.file(), request.request_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetSystemVisibilityState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetSystemVisibilityState&>(message);
            set_system_visibility_state(request.page_id(), request.visibility_state());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::AlertClosed: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::AlertClosed&>(message);
            alert_closed(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ConfirmClosed: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ConfirmClosed&>(message);
            confirm_closed(request.page_id(), request.accepted());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::PromptClosed: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::PromptClosed&>(message);
            prompt_closed(request.page_id(), request.response());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ColorPickerUpdate: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ColorPickerUpdate&>(message);
            color_picker_update(request.page_id(), request.picked_color(), request.state());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::FilePickerClosed: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::FilePickerClosed&>(message);
            file_picker_closed(request.page_id(), request.selected_files());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SelectDropdownClosed: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SelectDropdownClosed&>(message);
            select_dropdown_closed(request.page_id(), request.selected_item_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ToggleMediaPlayState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ToggleMediaPlayState&>(message);
            toggle_media_play_state(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ToggleMediaMuteState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ToggleMediaMuteState&>(message);
            toggle_media_mute_state(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ToggleMediaLoopState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ToggleMediaLoopState&>(message);
            toggle_media_loop_state(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::ToggleMediaControlsState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::ToggleMediaControlsState&>(message);
            toggle_media_controls_state(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::TogglePageMuteState: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::TogglePageMuteState&>(message);
            toggle_page_mute_state(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SetUserStyle: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SetUserStyle&>(message);
            set_user_style(request.page_id(), request.source());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::EnableInspectorPrototype: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::EnableInspectorPrototype&>(message);
            enable_inspector_prototype(request.page_id());
            return nullptr;

        }

        case (int)Messages::WebContentServer::MessageID::SystemTimeZoneChanged: {

            [[maybe_unused]] auto& request = static_cast<const Messages::WebContentServer::SystemTimeZoneChanged&>(message);
            system_time_zone_changed();
            return nullptr;

        }

        default:
            return Error::from_string_literal("Unknown message ID for WebContentServer endpoint");
        }
    }

    virtual Messages::WebContentServer::InitTransportResponse init_transport(
[[maybe_unused]] int peer_pid) = 0;

    virtual void close_server(
) = 0;

    virtual Messages::WebContentServer::GetWindowHandleResponse get_window_handle(
[[maybe_unused]] u64 page_id) = 0;

    virtual void set_window_handle(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& handle) = 0;

    virtual void connect_to_webdriver(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& webdriver_ipc_path) = 0;

    virtual void connect_to_image_decoder(
[[maybe_unused]] IPC::File const& socket_fd) = 0;

    virtual void update_system_theme(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Core::AnonymousBuffer const& theme_buffer) = 0;

    virtual void update_screen_rects(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<Web::DevicePixelRect> const& rects, [[maybe_unused]] u32 main_screen_index) = 0;

    virtual void load_url(
[[maybe_unused]] u64 page_id, [[maybe_unused]] URL::URL const& url) = 0;

    virtual void load_html(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& html) = 0;

    virtual void reload(
[[maybe_unused]] u64 page_id) = 0;

    virtual void traverse_the_history_by_delta(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 delta) = 0;

    virtual void ready_to_paint(
[[maybe_unused]] u64 page_id) = 0;

    virtual void set_viewport_size(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::DevicePixelSize size) = 0;

    virtual void key_event(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::KeyEvent const& event) = 0;

    virtual void mouse_event(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::MouseEvent const& event) = 0;

    virtual void drag_event(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::DragEvent const& event) = 0;

    virtual void debug_request(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& request, [[maybe_unused]] ByteString const& argument) = 0;

    virtual void get_source(
[[maybe_unused]] u64 page_id) = 0;

    virtual void inspect_dom_tree(
[[maybe_unused]] u64 page_id) = 0;

    virtual void inspect_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] Optional<Web::CSS::Selector::PseudoElement::Type> const& pseudo_element) = 0;

    virtual void inspect_accessibility_tree(
[[maybe_unused]] u64 page_id) = 0;

    virtual void get_hovered_node_id(
[[maybe_unused]] u64 page_id) = 0;

    virtual void js_console_input(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& js_source) = 0;

    virtual void js_console_request_messages(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 start_index) = 0;

    virtual void list_style_sheets(
[[maybe_unused]] u64 page_id) = 0;

    virtual void request_style_sheet_source(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::StyleSheetIdentifier const& identifier) = 0;

    virtual void set_dom_node_text(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] String const& text) = 0;

    virtual void set_dom_node_tag(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] String const& name) = 0;

    virtual void add_dom_node_attributes(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] Vector<WebView::Attribute> const& attributes) = 0;

    virtual void replace_dom_node_attribute(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id, [[maybe_unused]] String const& name, [[maybe_unused]] Vector<WebView::Attribute> const& replacement_attributes) = 0;

    virtual void create_child_element(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void create_child_text_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void clone_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void remove_dom_node(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void get_dom_node_html(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void take_document_screenshot(
[[maybe_unused]] u64 page_id) = 0;

    virtual void take_dom_node_screenshot(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::UniqueNodeID const& node_id) = 0;

    virtual void request_internal_page_info(
[[maybe_unused]] u64 page_id, [[maybe_unused]] WebView::PageInfoType type) = 0;

    virtual void run_javascript(
[[maybe_unused]] u64 page_id, [[maybe_unused]] ByteString const& js_source) = 0;

    virtual Messages::WebContentServer::GetSelectedTextResponse get_selected_text(
[[maybe_unused]] u64 page_id) = 0;

    virtual void select_all(
[[maybe_unused]] u64 page_id) = 0;

    virtual void paste(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& text) = 0;

    virtual void find_in_page(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& query, [[maybe_unused]] AK::CaseSensitivity case_sensitivity) = 0;

    virtual void find_in_page_next_match(
[[maybe_unused]] u64 page_id) = 0;

    virtual void find_in_page_previous_match(
[[maybe_unused]] u64 page_id) = 0;

    virtual void set_content_filters(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<String> const& filters) = 0;

    virtual void set_autoplay_allowed_on_all_websites(
[[maybe_unused]] u64 page_id) = 0;

    virtual void set_autoplay_allowlist(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<String> const& allowlist) = 0;

    virtual void set_proxy_mappings(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<ByteString> const& proxies, [[maybe_unused]] HashMap<ByteString, size_t> const& mappings) = 0;

    virtual void set_preferred_color_scheme(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::PreferredColorScheme const& color_scheme) = 0;

    virtual void set_preferred_contrast(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::PreferredContrast const& contrast) = 0;

    virtual void set_preferred_motion(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::CSS::PreferredMotion const& motion) = 0;

    virtual void set_preferred_languages(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<String> const& preferred_languages) = 0;

    virtual void set_enable_do_not_track(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool enable) = 0;

    virtual void set_has_focus(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool has_focus) = 0;

    virtual void set_is_scripting_enabled(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool is_scripting_enabled) = 0;

    virtual void set_device_pixels_per_css_pixel(
[[maybe_unused]] u64 page_id, [[maybe_unused]] float device_pixels_per_css_pixel) = 0;

    virtual void set_window_position(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::DevicePixelPoint position) = 0;

    virtual void set_window_size(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::DevicePixelSize size) = 0;

    virtual void did_update_window_rect(
[[maybe_unused]] u64 page_id) = 0;

    virtual Messages::WebContentServer::GetLocalStorageEntriesResponse get_local_storage_entries(
[[maybe_unused]] u64 page_id) = 0;

    virtual Messages::WebContentServer::GetSessionStorageEntriesResponse get_session_storage_entries(
[[maybe_unused]] u64 page_id) = 0;

    virtual void handle_file_return(
[[maybe_unused]] u64 page_id, [[maybe_unused]] i32 error, [[maybe_unused]] Optional<IPC::File> const& file, [[maybe_unused]] i32 request_id) = 0;

    virtual void set_system_visibility_state(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Web::HTML::VisibilityState visibility_state) = 0;

    virtual void alert_closed(
[[maybe_unused]] u64 page_id) = 0;

    virtual void confirm_closed(
[[maybe_unused]] u64 page_id, [[maybe_unused]] bool accepted) = 0;

    virtual void prompt_closed(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Optional<String> const& response) = 0;

    virtual void color_picker_update(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Optional<Color> const& picked_color, [[maybe_unused]] Web::HTML::ColorPickerUpdateState const& state) = 0;

    virtual void file_picker_closed(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Vector<Web::HTML::SelectedFile> const& selected_files) = 0;

    virtual void select_dropdown_closed(
[[maybe_unused]] u64 page_id, [[maybe_unused]] Optional<u32> const& selected_item_id) = 0;

    virtual void toggle_media_play_state(
[[maybe_unused]] u64 page_id) = 0;

    virtual void toggle_media_mute_state(
[[maybe_unused]] u64 page_id) = 0;

    virtual void toggle_media_loop_state(
[[maybe_unused]] u64 page_id) = 0;

    virtual void toggle_media_controls_state(
[[maybe_unused]] u64 page_id) = 0;

    virtual void toggle_page_mute_state(
[[maybe_unused]] u64 page_id) = 0;

    virtual void set_user_style(
[[maybe_unused]] u64 page_id, [[maybe_unused]] String const& source) = 0;

    virtual void enable_inspector_prototype(
[[maybe_unused]] u64 page_id) = 0;

    virtual void system_time_zone_changed(
) = 0;

private:
};

#if defined(AK_COMPILER_CLANG)
#pragma clang diagnostic pop
#endif
